/**
	$Id: XMLTemplate.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.plugins;

import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.List;
import java.util.regex.Pattern;

import net.sf.basedb.core.Item;

import org.jdom.DocType;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

/**
	A XML template that exports data in a simple one-level XML format.
	 
	@author nicklas
	@version 2.2
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class XMLTemplate
	implements ExportTemplate
{
	private final XMLOutputter xmlOut;
	private final Writer exportStream;
	private final List<ExportedProperty> properties;
	private final Item itemType;
	private final String itemIdPrefix;
	private Element itemElement;

	/**
		Create a new template.
		@param exportStream The stream to write to
		@param properties All exported properties
		@param itemType The type of items we are exporting
	*/
	public XMLTemplate(Writer exportStream, List<ExportedProperty> properties, Item itemType)
	{
		xmlOut = new XMLOutputter(Format.getPrettyFormat().setEncoding("UTF-8"));
		this.exportStream = exportStream;
		this.properties = properties;
		this.itemType = itemType;
		this.itemIdPrefix = itemType.name().toLowerCase() + "_";
	}
	
	/**
		@return application/xml
	*/
	@Override
	public String getMimeType()
	{
		return "application/xml";
	}
	
	/**
		@return Always "xml"
		@since 3.0
	*/
	@Override
	public String getDefaultFileExtension()
	{
		return "xml";
	}

	/**
		Write XML start (&lt;?xml ...&gt;) tag and DOCTYPE.
	*/
	@Override
	public void beginDocument()
		throws IOException
	{
		exportStream.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		xmlOut.output(new DocType("export", "export.dtd"), exportStream);
		exportStream.write("\n<export>\n");
	}
	
	/**
		Write property and annotation definitions.
	*/
	@Override
	public void writeHeaders()
		throws IOException
	{
		exportStream.write("<properties>\n");
		for (ExportedProperty ep : properties)
		{
			Element property = new Element(ep.annotationType == null ? "property" : "annotation");
			property.setAttribute("id", ep.name);
			property.setText(ep.title);
			exportStream.write("\t");
			xmlOut.output(property, exportStream);
			exportStream.write("\n");
		}
		exportStream.write("</properties>\n<items>\n");
	}

	/**
		Begin a new item element.
	*/
	@Override
	public void beginItem(int id)
		throws IOException
	{
		itemElement = new Element("item");
		itemElement.setAttribute("id", itemIdPrefix + id);
	}

	/**
		Add annotation values to the current item element.
	*/
	@Override
	@SuppressWarnings("unchecked")
	public void writeAnnotations(ExportedProperty ep, List<?> values, String unit)
		throws IOException
	{
		Element property = new Element("annotation-values");
		property.setAttribute("ref", ep.name);
		if (unit != null) property.setAttribute("unit", unit);
		if (values != null)
		{
			for (Object value : values)
			{
				Element valueElement = new Element("value");
				valueElement.setText(escape(ep.formatter.format(value)));
				property.addContent(valueElement);
			}
		}
		itemElement.addContent(property);
	}
	
	/**
		Add the property value to the current item element.
		@since 2.12
	*/
	@Override
	@SuppressWarnings("unchecked")
	public void writeProperty(ExportedProperty ep, Object data)
	{
		Element property = new Element("property-value");
		property.setAttribute("ref", ep.name);
		property.setText(escape(ep.formatter.format(data)));
		itemElement.addContent(property);
	}
	
	/**
	 	Add collection values to the current item element.
	 */
	@Override
	@SuppressWarnings("unchecked")
	public void writeCollection(ExportedProperty ep, Collection<?> values)
		throws IOException
	{
		
		Element property = new Element("property-value");
		property.setAttribute("ref", ep.name);
		if (values != null)
		{
			for (Object value : values)
			{
				Element valueElement = new Element("value");
				valueElement.setText(escape(ep.formatter.format(value)));
				property.addContent(valueElement);
			}
		}
		itemElement.addContent(property);
	}
	
	/**
		Write the current item element to the export stream.
	*/
	@Override
	public void endItem()
		throws IOException
	{
		xmlOut.output(itemElement, exportStream);
		exportStream.write("\n");
	}
	
	/**
		Close the document tags.
	*/
	@Override
	public void endDocument()
		throws IOException
	{
		exportStream.write("</items>\n</export>");
	}
	
	private static final Pattern NEWLINE = Pattern.compile("\\r?\\n\\r?");
	private String escape(String data)
	{
		if (data == null) 
		{
			return "";
		}
		else
		{
			return NEWLINE.matcher(data).replaceAll("\n");
		}
	}
}
