/*
	$Id: Head.java 6570 2014-10-22 05:55:46Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import net.sf.basedb.clients.web.extensions.ExtensionsControl;
import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Version;
import net.sf.basedb.util.Values;

import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Arrays;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.TagSupport;

/**
	This tag defines the header section of a page. It must be located inside
	a {@link Page &lt;base:page&gt;} ag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:head
      styles=...
      scripts=...
   &gt;
</pre>
	<p>

	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>styles</td>
		<td>-</td>
		<td>no</td>
		<td>
			A comma-separated list of stylesheet filenames that is needed by the page.
			The <code>main.css</code> are <code>menu.css</code> automatically included.
			The stylesheets must be located in the <code>/include/styles</code> folder.
			Do not include the folder names in the stylesheet filename.
		</td>
	</tr>
	<tr>
		<td>scripts</td>
		<td>-</td>
		<td>no</td>
		<td>
			A comma-separated list of javascript filenames that is needed by the page.
			The <code>main.js</code> are <code>menu.js</code> automatically included.
			The scripts must be located in the <code>/include/scripts</code> folder.
			Do not include the folder names in the scripts filename.
		</td>
	</tr>
	</table>


	@author Nicklas
	@version 2.0
*/
public class Head
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = 948719986796860767L;

	// Contain the file name of extension JAR files mapped to versions
	private static final Map<String, String> xtVersions = new HashMap<String, String>();

	// Match files under /base-root/extensions path. group 1 is the JAR file name
	private static final Pattern PATH_MATCH = 
			Pattern.compile("/?[^/]*" + ExtensionsControl.RESOURCES_URL + "/([^/]+)/(.*)");

	/**
		Register the version for an exteniosn JAR file.
		Used for appending a query parameter to all scripts and style sheets
		to prevent that browsers cache the old version.
		@see #getVersionParameter(String)
		@since 3.3.2
	*/
	public static void setExtensionVersion(String jarName, String version)
	{
		synchronized (xtVersions)
		{
			xtVersions.put(jarName, "v="+HTML.urlEncode(version));
		}
	}

	/**
		Get query parameter to the given absolute path (including root directory).
		If the path points to a file in the /base-root/extensions/ directory the version
		of the extension is used, otherwise the BASE version is used. The
		returned parameter is of the form v=<version>.
		@param absolutePath An absolute path with or without the base-root directory
		@since 3.3.2
	*/
	public static final String getVersionParameter(String absolutePath)
	{
		Matcher m = PATH_MATCH.matcher(absolutePath);
		String version = globalVersion;
		if (m.matches())
		{
			String jarFile = m.group(1);
			if (xtVersions.containsKey(jarFile))
			{
				version = xtVersions.get(jarFile);
			}
		}
		return version;
	}

	
	/**
		The parent &lt;base:page&gt; tag.
	*/
	private Page page;
	
	/**
		A comma-separated list of stylesheets filenames.
	*/
	private String styles = null;
	
	/**
		A comma-separated list of JavaScript filenames.
	*/
	private String scripts = null;
	
	private static final String globalVersion = "v=" + Version.getMajor()+"."+Version.getMinor()+"."+Version.getMaintenance();
	private String localVersion;
	
	/*
		Taglib initialization methods
	*/
	public void setStyles(String styles)
	{
		this.styles = styles;
	}
	public String getStyles()
	{
		return styles;
	}
	
	public void setScripts(String scripts)
	{
		this.scripts = scripts;
	}
	public String getScripts()
	{
		return scripts;
	}

	private void appendStyles(StringBuilder sb, String fontStyles, boolean hideLongTexts)
	{
		LinkedHashSet<String> allStyles = new LinkedHashSet<String>();
		allStyles.add("main.css");
		int pageType = page.getTypeCode();
		if (pageType == Page.PAGE_TYPE_DEFAULT) allStyles.add("menu.css");
		if (pageType == Page.PAGE_TYPE_POPUP) allStyles.add("popup.css");
		if (getStyles() != null) 
		{
			allStyles.addAll(Arrays.asList(getStyles().split(",")));
		}
		allStyles.add(fontStyles);
		if (hideLongTexts) allStyles.add("long_texts.css");
		String appRoot = page.getRoot();
		for (String style : allStyles)
		{
			char pp = style.indexOf('?') == -1 ? '?' : '&';
			sb.append("\t<link rel=\"stylesheet\" type=\"text/css\" href=\"");
			if (style.startsWith(appRoot))
			{
				// Full absolute path 
				sb.append(style).append(pp).append(getVersionParameter(style));
			}
			else if (style.startsWith("/"))
			{
				sb.append(appRoot).append(style.substring(1)).append(pp).append(getVersionParameter(style));
			}
			else if (style.startsWith("~"))
			{
				// Relative path is relative the current page (eg. ~bar.css --> bar.css)
				sb.append(style.substring(1)).append(pp).append(localVersion);
			}
			else
			{
				// All other style sheets are in the /include/styles folder
				sb.append(appRoot).append("include/styles/").append(style).append(pp).append(globalVersion);
			}
			sb.append("\">\n");
		}
	}

	private void appendScripts(StringBuilder sb)
	{
		SessionControl sc = page.getSessionControl();
		LinkedHashSet<String> allScripts = new LinkedHashSet<String>();

		int pageType = page.getTypeCode();
		if (pageType != Page.PAGE_TYPE_INCLUDE)
		{
			allScripts.add("main-2.js");
			allScripts.add("main.js");
		}
		if (pageType == Page.PAGE_TYPE_DEFAULT) 
		{
			allScripts.add("menu.js");
		}
		if (pageType == Page.PAGE_TYPE_POPUP) allScripts.add("popup.js");
		if (getScripts() != null) 
		{
			allScripts.addAll(Arrays.asList(getScripts().split(",")));
		}
		
		String appRoot = page.getRoot();
		for (String script : allScripts)
		{
			char pp = script.indexOf('?') == -1 ? '?' : '&';
			sb.append("\t<script type=\"text/javascript\" charset=\"UTF-8\" src=\"");
			if (script.startsWith(appRoot))
			{
				// Full absolute path 
				sb.append(script).append(pp).append(getVersionParameter(script));
			}
			else if (script.startsWith("/"))
			{
				// Absolute path that is relative the BASE root (eg. /foo/bar.js --> /base/foo/bar.js)
				sb.append(appRoot).append(script.substring(1)).append(pp).append(getVersionParameter(script));
			}
			else if (script.startsWith("~"))
			{
				// Relative path is relative the current page (eg. ~bar.js --> bar.js)
				sb.append(script.substring(1)).append(pp).append(localVersion);
			}
			else
			{
				// All other scripts are in the /include/scripts folder
				sb.append(appRoot).append("include/scripts/").append(script).append(pp).append(globalVersion);
			}
			sb.append("\"></script>\n");
		}
	}

	@Override
	public void setPageContext(PageContext pageContext) 
	{
		super.setPageContext(pageContext);
		HttpServletRequest req = (HttpServletRequest)pageContext.getRequest();
		// If we are on /extension subpage, the localVersion is the
		// version for the extension, otherwise we use the global version
		localVersion = getVersionParameter(req.getServletPath());
	}
	
	
	@Override
	public int doStartTag()
		throws JspException
	{
		if (!(getParent() instanceof Page)) throw new JspTagException("Tag <base:head> must be inside a <base:page> tag");
		page = (Page)getParent();
		
		StringBuilder sb = new StringBuilder();
		int returnValue = EVAL_BODY_INCLUDE;
		if (page.getTypeCode() == Page.PAGE_TYPE_INCLUDE) 
		{
			appendScripts(sb);
			returnValue = SKIP_BODY;
		}
		else
		{
			SessionControl sc = page.getSessionControl();
			String longTexts = Values.getString(sc != null ? sc.getUserClientSetting("text.long") : null, "display");
			String fontStyles = Values.getString(sc != null ? sc.getUserClientSetting("appearance.fontsize") : null, "size_m.css");
			
			String favicon = page.getFavicon();
			if (favicon != null && !favicon.startsWith("/"))
			{
				favicon = page.getRoot()+favicon;
			}

			sb.append("<head>\n");
			if (favicon != null)
			{
				sb.append("\t<link id=\"fav-icon\" rel=\"SHORTCUT ICON\" href=\"").append(favicon).append("\">\n");
			}
			appendStyles(sb, fontStyles, !"display".equals(longTexts));
			appendScripts(sb);
			
			String title = page.getTitle();
			if (title == null)
			{
				title = page.getBaseVersion() + " @ " + page.getServerName();
			}
			sb.append("\t<title>").append(title).append("</title>\n");
		}

		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return returnValue;
	}

	@Override
	public int doEndTag() 
		throws JspException
	{
		if (page.getTypeCode() != Page.PAGE_TYPE_INCLUDE)
		{
			StringBuilder sb = new StringBuilder();
			sb.append("</head>\n");
			try
			{
				pageContext.getOut().print(sb.toString());
			}
			catch (Exception ex)
			{
				throw new JspTagException(ex.getMessage());
			}
		}
		return EVAL_PAGE;
	}
}



