/**
	$Id: StylableTag.java 6571 2014-10-22 07:48:36Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.tagext.DynamicAttributes;
import javax.servlet.jsp.tagext.TagSupport;

import net.sf.basedb.clients.web.extensions.DynamicActionAttribute;
import net.sf.basedb.clients.web.extensions.DynamicActionAttributes;


/**
	Generic tag class for tags that want to support 
	id, class and style of the main html object they 
	use. The class can be specified as a main class and
	a subclass which is added instead of replacing the
	main class.

	@author Nicklas
	@since 3.1
	@base.modified $Date: 2014-10-22 09:48:36 +0200 (on, 22 okt 2014) $
*/
public abstract class StylableTag
	extends TagSupport
	implements DynamicAttributes, DynamicActionAttributes
{

	private static final long serialVersionUID = -383421425096524117L;
	private String clazz;
	private String subclass;
	private String style;
	private Map<String, DynamicActionAttribute> dynamicAttributes;
	
	/**
		Create a new stylable tag.
		@param defaultClass The default class to use if the tag doesn't
			override with it's own value
	 */
	protected StylableTag(String defaultClass)
	{
		this.clazz = defaultClass;
	}
	
	/**
		Reset the dynamic attributes to make sure old ones are not
		included in case the tag object is reused.
		@since 3.3
	*/
	@Override
	public void setPageContext(PageContext pageContext) 
	{
		super.setPageContext(pageContext);
		dynamicAttributes = null;
		initDefaultAttributes();
	}

	/**
		Initialize default attributes. Can be called multiple times (once before each tag is used/re-used).
		This default implementation does nothing.
		@since 3.3.2
	*/
	protected void initDefaultAttributes()
	{}
	
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}
	
	public void setSubclass(String subclass)
	{
		this.subclass = subclass;
	}
	public String getSubclass()
	{
		return subclass;
	}
	
	public void setStyle(String style)
	{
		this.style = style;
	}
	public String getStyle()
	{
		return style;
	}

	/*
		From the DynamicAttributes interface
		------------------------------------
	*/
	@Override
	public void setDynamicAttribute(String uri, String name, Object value) 
	{
		if (value == null) return;
		if (dynamicAttributes == null)
		{
			dynamicAttributes = new LinkedHashMap<String, DynamicActionAttribute>();
		}
		dynamicAttributes.put(name, new DynamicActionAttribute(name, value.toString()));
	}
	// -------------------------------------
	
	/*
		From the DynamicActionAttributes interface
		------------------------------------------
	*/
	@Override
	public Iterator<DynamicActionAttribute> getDynamicActionAttributes() 
	{
		return dynamicAttributes == null ? null : dynamicAttributes.values().iterator();
	}
	// ------------------------------------------

	
	/**
		Get the full class (eg. main and subclass) as a single
		string.
		@param extraClasses Additional classes that should also be added
		@return The combined class or null if no class has been set
	*/
	public String getFullClass(String...extraClasses)
	{
		StringBuilder fullClass = new StringBuilder();
		if (getClazz() != null)
		{
			fullClass.append(getClazz());
		}
		if (getSubclass() != null)
		{
			if (fullClass.length() > 0) fullClass.append(" ");
			fullClass.append(getSubclass());
		}
		if (extraClasses != null && extraClasses.length > 0)
		{
			for (String extra: extraClasses)
			{
				if (extra != null)
				{
					if (fullClass.length() > 0) fullClass.append(" ");
					fullClass.append(extra);
				}
			}
		}
		return fullClass.length() == 0 ? null : fullClass.toString();
	}
	
	/**
		Optional prefix to include in the 'id' attribute.
		See {@link #addIdAndStyles(StringBuilder, String...)}.
		@return The default implementation return an empty string
		@since 3.3
	*/
	protected String getIdPrefix()
	{
		return "";
	}
	
	/**
		Get the dynamic attribute for the given name
		@param name The name of the attribute
		@return A DynamicActionAttribute object or null if no attribute with that
			name has been set
	 */
	protected DynamicActionAttribute getDynamicAttribute(String name)
	{
		return dynamicAttributes == null ? null : dynamicAttributes.get(name);
	}
	
	/**
		Add id, class and style attributes the the stringbuilder.
	*/
	protected void addIdAndStyles(StringBuilder sb, String... extraClasses)
	{
		if (getId() != null) sb.append(" id=\"").append(getIdPrefix()).append(getId()).append("\"");
		String fullClass = getFullClass(extraClasses);
		if (fullClass != null) sb.append(" class=\"").append(fullClass).append("\"");
		if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
	}
	
	protected void setMoreDynamicAttributes(Iterator<DynamicActionAttribute> moreAttributes)
	{
		if (moreAttributes == null || !moreAttributes.hasNext()) return;
		if (dynamicAttributes == null)
		{
			dynamicAttributes = new LinkedHashMap<String, DynamicActionAttribute>();
		}
		while (moreAttributes.hasNext())
		{
			DynamicActionAttribute attr = moreAttributes.next();
			dynamicAttributes.put(attr.getName(), attr);
		}
	}
	
	protected void addDynamicAttributes(StringBuilder sb)
	{
		if (dynamicAttributes != null)
		{
			for (DynamicActionAttribute entry : dynamicAttributes.values())
			{
				sb.append(" ").append(entry.getName()).append("=\"").append(entry.getValue()).append("\"");
			}
		}
	}
	
}
