/*
	$Id: Metadata.java 6355 2013-11-26 11:46:42Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.util.ToStringComparator;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.BatchableData;
import net.sf.basedb.core.data.IdentifiableData;
import net.sf.basedb.core.data.ReporterListScoreData;
import net.sf.basedb.core.log.EntityLogOptions;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.HashMap;
import java.util.WeakHashMap;

import org.hibernate.engine.SessionFactoryImplementor;
import org.hibernate.metadata.ClassMetadata;
import org.hibernate.proxy.LazyInitializer;
import org.hibernate.type.CollectionType;
import org.hibernate.type.ComponentType;
import org.hibernate.type.EntityType;
import org.hibernate.type.MapType;
import org.hibernate.EntityMode;
import org.hibernate.TypeHelper;

/**
	Utility methods for retreiving metadata about items.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2013-11-26 12:46:42 +0100 (ti, 26 nov 2013) $
*/
public class Metadata<C>
{

	private static Set<Item> annotatableItems;
	private static Set<Item> shareableItems;
	private static Set<Item> removableItems;
	private static Set<Item> ownableItems;
	private static Set<Item> diskConsumableItems;
	private static Set<Item> fileStoreEnabledItems;
	private static Set<Item> subtypableItems;

	private static boolean isInitialised = false;
	
	static synchronized void init()
	{
		if (isInitialised) return;
		annotatableItems = Collections.unmodifiableSet(getImplementingItems(Annotatable.class));
		shareableItems = Collections.unmodifiableSet(getImplementingItems(Shareable.class));
		removableItems = Collections.unmodifiableSet(getImplementingItems(Removable.class));
		ownableItems = Collections.unmodifiableSet(getImplementingItems(Ownable.class));
		diskConsumableItems = Collections.unmodifiableSet(getImplementingItems(DiskConsumable.class));
		fileStoreEnabledItems = Collections.unmodifiableSet(getImplementingItems(FileStoreEnabled.class));
		subtypableItems = Collections.unmodifiableSet(getImplementingItems(Subtypable.class));
		isInitialised = true;
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		annotatableItems = null;
		shareableItems = null;
		removableItems = null;
		ownableItems = null;
		diskConsumableItems = null;
		fileStoreEnabledItems = null;
		subtypableItems = null;
	}

	/**
		Get all {@link Item}:s implementing/extinding the specified 
		interface/class.
	*/
	private static Set<Item> getImplementingItems(Class<?> clazz)
	{
		Set<Item> items = new TreeSet<Item>(new ToStringComparator<Item>(false));
		for (Item i : Item.values())
		{
			Class<?> itemClass = i.getItemClass();
			if (itemClass != null && clazz.isAssignableFrom(itemClass))
			{
				items.add(i);
			}
		}
		return items;
	}
	
	/**
		Get a set containing all {@link Annotatable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getAnnotatableItems()
	{
		return annotatableItems;
	}
	
	/**
		Get a set containing all {@link Shareable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getShareableItems()
	{
		return shareableItems;
	}

	/**
		Get a set containing all {@link Removable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getRemovableItems()
	{
		return removableItems;
	}
	
	/**
		Get a set containing all {@link Ownable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getOwnableItems()
	{
		return ownableItems;
	}
	
	/**
		Get a set containing all {@link DiskConsumable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getDiskConsumableItems()
	{
		return diskConsumableItems;
	}

	/**
		Get a set containing all {@link FileStoreEnabled} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getFileStoreEnabledItems()
	{
		return fileStoreEnabledItems;
	}
	
	/**
		Get a set containing all {@link Subtypable} item types. 
		The set cannot be modified.
		@since 3.0
	*/
	public static Set<Item> getSubtypableItems()
	{
		return subtypableItems;
	}

	
	/**
		Cache of Metadata instances. Only one instance for each type
		of item exists.
	*/
	private static final Map<String, Metadata<?>> instances = new HashMap<String, Metadata<?>>();
	
	/**
		Get the metadata instance for the specified type of item.
	 	@param <I> 
		@param itemClass The class of the item
	 	@return a new instance of this class. 
	 	@throws BaseException If a new instance could not be created. 
	*/
	@SuppressWarnings({"unchecked"})
	public static <I> Metadata<I> getInstance(Class<I> itemClass)
		throws BaseException
	{
		Item itemType = null;
		if (BasicItem.class.isAssignableFrom(itemClass))
		{
			itemType = Item.fromItemClass((Class<BasicItem>)itemClass);
		}
		else if (BasicData.class.isAssignableFrom(itemClass))
		{
			itemType = Item.fromDataClass((Class<BasicData>)itemClass);
		}
		if (itemType == null)
		{
			throw new BaseException("Class '"+itemClass+"' is not known as a persistent entity");
		}
		return getInstance(itemClass, itemType.getDataClass().getName());
	}

	/**
		Get the metadata instance for the item with the specified 
		entity name. Use this method where one class is mapped to
		several entities, for example RawData.
	 	@param <I> 
		@param clazz The class of the item, should be either a
			BasicItem or BasicData class
		@param entityName The entity name for the item, normally
			this is <code>clazz.getName()</code> but some entities,
			for example <code>RawData</code>, have multiple names
	 	@return a new instance of this class 
	 	@throws BaseException If a new instance could not be created.
	*/
	@SuppressWarnings({"unchecked"})
	public static <I> Metadata<I> getInstance(Class<I> clazz, String entityName)
		throws BaseException
	{
		Metadata<I> md = (Metadata<I>)instances.get(entityName);
		if (md == null)
		{
			synchronized (instances)
			{
				md = new Metadata<I>(clazz, entityName);
				instances.put(entityName, md);
			}
		}
		return md;
	}
	
	private final ClassMetadata classMetadata;
	private final String entityName;
	private final Class<C> clazz;
	private final Map<String, PropertyPath> cachedPaths;
	
	private Metadata(Class<C> clazz, String entityName)
		throws BaseException
	{
		this.entityName = entityName;
		this.clazz = clazz;
		this.classMetadata = HibernateUtil.getClassMetadata(entityName);
		if (classMetadata == null)
		{
			throw new ItemNotFoundException("Metadata[entityName=" + entityName + "]");
		}
		this.cachedPaths = new WeakHashMap<String, PropertyPath>();
	}
	
	/**
		Get a property path that can be used to retrieve a value 
		from an item instead of using the regular <code>get</code>
		methods. The following method calls are equivalent:
		<pre class="code">
// Using getter method
String name = item.getName();

// Using PropertyPath
PropertyPath path = metadata.getPropertyPath("name");
String name = (String)path.getValue(dc, item);
	</pre>
	
		You can even use dot notation to navigate to associated items:
		<pre class="code">
// Using getter methods
String quota = item.getOwner().getQuota().getName();

// Using PropertyPath
PropertyPath path = metadata.getPropertyPath("owner.quota.name");
String quota = (String)path.getValue(dc, item);
</pre>
	
		@param <V> The class of the returned values
		@param path The name of the property to get. Null is not allowed. Use
			. or # as a separator between property and association names
	 	@return A property path object
	 	@throws NullPointerException If the path is null
	 	@throws InvalidDataException If the path is not a valid path
	 	@since 2.15
 	*/
	@SuppressWarnings({ "unchecked" })
	public <V> PropertyPath<C, V> getPropertyPath(String path, boolean ignoreCase)
	{
		if (path == null) throw new NullPointerException("path");
		PropertyPath pp = cachedPaths.get(path);
		if (pp != null) return pp;
		
		String[] pathElements = path.split("\\.|\\#");
		ClassMetadata metadata = classMetadata;
		TypeHelper typeHelper = HibernateUtil.getSessionFactory().getTypeHelper();
		org.hibernate.type.Type nextType = typeHelper.entity(entityName);
		List<PropertyPath> propertyPath = new ArrayList<PropertyPath>();
		
		// Pre-conversion is making sure that we work with data-layer classes downstream
		if (BasicItem.class.isAssignableFrom(clazz))
		{
			// Convert BasicItem to BasicData objects
			propertyPath.add(new BasicItem2DataPropertyPath(clazz, nextType));
		}
		else if (BasicData.class.isAssignableFrom(clazz))
		{
			// Only needed to get nice toString() values
			propertyPath.add(new NoopPropertyPath<C>(clazz, nextType));
		}
		else if (ReporterScore.class.isAssignableFrom(clazz))
		{
			// Convert ReporterScore to ReporterListScoreData objects
			propertyPath.add(new ReporterScore2DataPropertyPath(nextType));
		}
		else 
		{
			throw new AssertionError("Metadata doesn't represent a BasicItem, BasicData or ReporterScore: " + clazz);
		}

		StringBuilder pathSoFar = new StringBuilder();
		pathSoFar.append(clazz.getSimpleName());
		Comparator<String> propertyNames = ignoreCase ? String.CASE_INSENSITIVE_ORDER : 
			new Comparator<String>() 
			{
				@Override
				public int compare(String o1, String o2) 
				{
					return o1.compareTo(o2);
				}
			};
		for (String property : pathElements)
		{
			pathSoFar.append(".").append(property);
			if (nextType.isEntityType())
			{
				metadata = HibernateUtil.getClassMetadata(((EntityType)nextType).getAssociatedEntityName());
				String idName = metadata.getIdentifierPropertyName();
				if (idName != null && propertyNames.compare(property, idName) == 0)
				{
					// 'id' needs special handing. It is not found in metadata.getPropertyNames() array
					propertyPath.add(new IdentifierPropertyPath(nextType));
					nextType = metadata.getIdentifierType();
				}
				else
				{
					int propertyIndex = getIndex(property, metadata.getPropertyNames(), propertyNames);
					if (propertyIndex == -1)
					{
						throw new InvalidDataException("Path does not exist: " + pathSoFar);
					}
					property = metadata.getPropertyNames()[propertyIndex];
					EntityLogOptions logOptions = EntityLogOptions.getEntityLogOptionsForClass(nextType.getReturnedClass());
					if (logOptions != null && logOptions.isPropertyPathProtected(property))
					{
						throw new PermissionDeniedException(pathSoFar.toString());
					}
					nextType = metadata.getPropertyTypes()[propertyIndex];
					propertyPath.add(new EntityPropertyPath(metadata, property, nextType));
				}
			}
			else if (nextType.isComponentType())
			{
				ComponentType ct = (ComponentType)nextType;
				int propertyIndex = getIndex(property, ct.getPropertyNames(), propertyNames);
				if (propertyIndex == -1)
				{
					throw new InvalidDataException("Path does not exist: " + pathSoFar);
				}
				property = ct.getPropertyNames()[propertyIndex];
				nextType = ct.getSubtypes()[propertyIndex];
				propertyPath.add(new ComponentPropertyPath(ct, propertyIndex, nextType));
			}
			else
			{
				throw new InvalidDataException("Invalid property path: " + pathSoFar);
			}
		}

		// Post-conversion
		Class returnClass = nextType.getReturnedClass();
		if (BatchableData.class.isAssignableFrom(returnClass))
		{
			// BatchableData are returned as is, but must be evicted from the session
			propertyPath.add(new EvictPropertyPath(nextType));
		}
		else if (BasicData.class.isAssignableFrom(returnClass))
		{
			// Convert to BasicItem
			propertyPath.add(new BasicData2ItemPropertyPath(Item.fromDataClass(returnClass), nextType));
		}
		else if (Date.class.isAssignableFrom(returnClass))
		{
			// Dates need to be cloned
			propertyPath.add(new DateCopyPropertyPath(nextType));
		}
		else if (nextType.isCollectionType())
		{
			CollectionType cType = (CollectionType)nextType;
			org.hibernate.type.Type elementType = 
				cType.getElementType((SessionFactoryImplementor)HibernateUtil.getSessionFactory());
			if (elementType.isEntityType())
			{
				throw new InvalidDataException("Path ends with entity collection: " + pathSoFar);
			}
			propertyPath.add(nextType instanceof MapType ? 
					new MapPropertyPath(nextType) : new CollectionPropertyPath(nextType));
		}
		else if (!nextType.isMutable())
		{
			// Safe to return the value unprocessed
		}
		else
		{
			throw new AssertionError("Unhandled return type for path " + pathSoFar + ": " + nextType);
		}
		
		pp = new CombinedPropertyPath(propertyPath);
		synchronized (cachedPaths) 
		{
			cachedPaths.put(path, pp);
		}
		return pp;
	}
	
	private int getIndex(String s, String[] list, Comparator<String> check)
	{
		for (int index = 0; index < list.length; ++index)
		{
			if (check.compare(s, list[index]) == 0) return index;
		}
		return -1;
	}
	
	@Override
	public String toString()
	{
		return "Metadata for entity " + entityName;
	}
	
	/**
		Contract for getting the value of a specific property for an item.
		@since 2.15
	*/
	public static interface PropertyPath<C, V>
	{
		/**
			Get the value as defined by the implementation for the given item.
			@param dc A DbControl to use for database access if needed
			@param item The item that owns the value
			@return The value from the item
		*/
		public V getValue(DbControl dc, C item);
		
		/**
			Get type information about the value that is returned by this path.
		*/
		public org.hibernate.type.Type getHibernateType();
		
		/**
			Get the path that this object represents.
			Implementors may return null if the path is an "invisible"
			transition, eg from BasicItem to BasicData or a cloned Date, etc.
		*/
		public String getPath();
	}
	
	/**
		Property path implementation that combines a list of property paths
		to create a long chain. The return value of the first entry is used
		as the argument to the next and so on.
		@since 2.15
	*/
	static class CombinedPropertyPath<C, V>
		implements PropertyPath<C, V>
	{
		private final List<PropertyPath> path;
		public CombinedPropertyPath(List<PropertyPath> path)
		{
			this.path = path;
		}
		@SuppressWarnings("unchecked")
		@Override
		public V getValue(DbControl dc, C item)
		{
			Object value = item;
			for (PropertyPath pp : path)
			{
				value = pp.getValue(dc, value);
				if (value == null) break;
			}
			return (V)value;
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return path.get(path.size()-1).getHibernateType();
		}
		@Override
		public String getPath()
		{
			StringBuilder sb = new StringBuilder();
			for (PropertyPath pp : path)
			{
				String path = pp.getPath();
				if (path != null)
				{
					if (sb.length() > 0) sb.append(".");
					sb.append(path);
				}
			}
			return sb.toString();
		}
		@Override
		public String toString()
		{
			StringBuilder sb = new StringBuilder();
			for (PropertyPath pp : path)
			{
				if (sb.length() > 0) sb.append(".");
				sb.append(pp.toString());
			}
			return sb.toString();
		}
	}
	
	/**
		Property path implementation that converts a BasicItem to a BasicData
		object.
		@since 2.15
	*/
	static class BasicItem2DataPropertyPath<C extends BasicItem<V>, V extends BasicData>
		implements PropertyPath<C, V>
	{
		private final Class<C> clazz;
		private final org.hibernate.type.Type hibernateType;
		BasicItem2DataPropertyPath(Class<C> clazz, org.hibernate.type.Type hibernateType)
		{
			this.clazz = clazz;
			this.hibernateType = hibernateType;
		}
		@Override
		public V getValue(DbControl dc, C item)
		{
			return item.getData();
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return null;
		}
		@Override
		public String toString()
		{
			return clazz.getSimpleName();
		}
	}
	
	/**
		Property path implementation that converts a ReporterScore to a ReporterListScoreData
		object.
		@since 2.15
	*/
	static class ReporterScore2DataPropertyPath
		implements PropertyPath<ReporterScore, ReporterListScoreData>
	{
		private final org.hibernate.type.Type hibernateType;
		ReporterScore2DataPropertyPath(org.hibernate.type.Type hibernateType)
		{
			this.hibernateType = hibernateType;
		}
		@Override
		public ReporterListScoreData getValue(DbControl dc, ReporterScore item)
		{
			return item.getData();
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return null;
		}
		@Override
		public String toString()
		{
			return "ReporterScore";
		}
	}
	
	/**
		Property path implementation that converts a BasicData object
		to a BasicItem.
		@since 2.15
	*/
	static class BasicData2ItemPropertyPath<C extends BasicData, V extends BasicItem<C>>
		implements PropertyPath<C, BasicItem>
	{
		private final Item itemType;
		private final org.hibernate.type.Type hibernateType;
		BasicData2ItemPropertyPath(Item itemType, org.hibernate.type.Type hibernateType)
		{
			this.itemType = itemType;
			this.hibernateType = hibernateType;
		}
		@Override
		public BasicItem getValue(DbControl dc, C item)
		{
			return item == null ? null : dc.getItem(BasicItem.class, item);
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return null;
		}
		@Override
		public String toString()
		{
			return "[" + itemType.toString() + "]";
		}
	}

	/**
		Property path implementation that evicts the selected item from
		the Hibernate first-level cache. Used with BatchableData classes which
		are returned as is to the client.
		@since 2.15
	*/
	static class EvictPropertyPath<C>
		implements PropertyPath<C, C>
	{
		private final org.hibernate.type.Type hibernateType;
		EvictPropertyPath(org.hibernate.type.Type hibernateType)
		{
			this.hibernateType = hibernateType;
		}
		@Override
		public C getValue(DbControl dc, C item)
		{
			dc.getHibernateSession().evict(item);
			return item;
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return null;
		}
		@Override
		public String toString()
		{
			return "[evicted]";
		}		
	}

	/**
		Property path implementation clones a Date.
		@since 2.15
	*/
	static class DateCopyPropertyPath
		implements PropertyPath<Date, Date>
	{
		private final org.hibernate.type.Type hibernateType;
		DateCopyPropertyPath(org.hibernate.type.Type hibernateType)
		{
			this.hibernateType = hibernateType;
		}
		@Override
		public Date getValue(DbControl dc, Date item)
		{
			return DateUtil.copy(item);
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return null;
		}
		@Override
		public String toString()
		{
			return "[cloned]";
		}
	}

	
	/**
		Property path implementation that does nothing.
		@since 2.15
	*/
	static class NoopPropertyPath<C>
		implements PropertyPath<C, C>
	{
		private final org.hibernate.type.Type hibernateType;
		private final Class<C> clazz;
		NoopPropertyPath(Class<C> clazz, org.hibernate.type.Type hibernateType)
		{
			this.clazz = clazz;
			this.hibernateType = hibernateType;
		}
		@Override
		public C getValue(DbControl dc, C item)
		{
			return item;
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return null;
		}
		@Override
		public String toString()
		{
			return clazz.getSimpleName();
		}
	}

	/**
		Property path implementation which gets a value from an entity.
		@since 2.15
	*/
	static class EntityPropertyPath<C, V>
		implements PropertyPath<C, V>
	{
		private final ClassMetadata metadata;
		private final String property;
		private final org.hibernate.type.Type hibernateType;
		EntityPropertyPath(ClassMetadata metadata, String property, org.hibernate.type.Type hibernateType)
		{
			this.metadata = metadata;
			this.property = property;
			this.hibernateType = hibernateType;
		}
		@SuppressWarnings("unchecked")
		@Override
		public V getValue(DbControl dc, C item)
		{
			if (item instanceof org.hibernate.proxy.HibernateProxy)
			{
				// If it is a proxy we must reload it using the current session
				org.hibernate.proxy.HibernateProxy proxy = (org.hibernate.proxy.HibernateProxy)item;
				LazyInitializer initializer = proxy.getHibernateLazyInitializer();
				item = (C)dc.getHibernateSession().load(initializer.getPersistentClass(), initializer.getIdentifier());
			}
			return (V)metadata.getPropertyValue(item, property, EntityMode.POJO);
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return property;
		}
		@Override
		public String toString()
		{
			return property;
		}
	}
	
	/**
		Property path implementation which gets the id from an entity.
		@since 2.15
	*/
	static class IdentifierPropertyPath
		implements PropertyPath<IdentifiableData, Integer>
	{
		private final org.hibernate.type.Type hibernateType;
		IdentifierPropertyPath(org.hibernate.type.Type hibernateType)
		{
			this.hibernateType = hibernateType;
		}
		@Override
		public Integer getValue(DbControl dc, IdentifiableData item)
		{
			return item.getId();
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return "id";
		}
		@Override
		public String toString()
		{
			return "id";
		}
	}

	/**
		Property path implementation which wraps a collection inside an unmodifiable
		collection.
		@since 2.15
	*/
	static class CollectionPropertyPath<C>
		implements PropertyPath<Collection<C>, Collection<C>>
	{
		private final org.hibernate.type.Type hibernateType;
		CollectionPropertyPath(org.hibernate.type.Type hibernateType)
		{
			this.hibernateType = hibernateType;
		}
		@Override
		public Collection<C> getValue(DbControl dc, Collection<C> item)
		{
			return Collections.unmodifiableCollection(item);
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return null;
		}
		@Override
		public String toString()
		{
			return "[unmodifiable]";
		}
	}
	
	/**
		Property path implementation which wraps a map inside an unmodifiable
		map.
		@since 2.15
	*/
	static class MapPropertyPath
		implements PropertyPath<Map, Map>
	{
		private final org.hibernate.type.Type hibernateType;
		MapPropertyPath(org.hibernate.type.Type hibernateType)
		{
			this.hibernateType = hibernateType;
		}
		@SuppressWarnings("unchecked")
		@Override
		public Map getValue(DbControl dc, Map item)
		{
			return Collections.unmodifiableMap(item);
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return null;
		}
		@Override
		public String toString()
		{
			return "[unmodifiable]";
		}
	}
	
	/**
		Property path implementation which gets a value from a component.
		@since 2.15
	*/
	static class ComponentPropertyPath<C, V>
		implements PropertyPath<C, V>
	{
		private ComponentType component;
		private int index;
		private final org.hibernate.type.Type hibernateType;
		ComponentPropertyPath(ComponentType component, int index, org.hibernate.type.Type hibernateType)
		{
			this.component = component;
			this.index = index;
			this.hibernateType = hibernateType;
		}
		@SuppressWarnings("unchecked")
		@Override
		public V getValue(DbControl dc, C item)
		{
			return (V)component.getPropertyValue(item, index, EntityMode.POJO);
		}
		@Override
		public org.hibernate.type.Type getHibernateType()
		{
			return hibernateType;
		}
		@Override
		public String getPath()
		{
			return component.getPropertyNames()[index];
		}
		@Override
		public String toString()
		{
			return component.getPropertyNames()[index];
		}		
	}
	
}
