/*
	$Id: PlatformVariant.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.PlatformVariantData;

/**
	Represents variants of a {@link Platform}.

	@author Nicklas
	@version 2.5
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class PlatformVariant
	extends BasicChildItem<PlatformVariantData>
	implements Nameable, Removable
{
	
	/**
		The external id for the expression-like sequencing data.
		@since 3.0
	*/
	public static final String SEQUENCING_EXPRESSION = "sequencing.expression";

	
	/**
		The type of item represented by this class.
		@see Item#PLATFORM
		@see #getType()
	*/
	public static final Item TYPE = Item.PLATFORMVARIANT;
	
	/**
		This filter will only return items if the logged in user has
		generic read permission to platforms.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.PLATFORM);
	
	
	/**
		Create a new variant of a platform. If the platform is a file-only
		platform the new variant will also be a file-only variant.
		
		@param dc The DbControl to use for database access
		@param platform The platform to create a variant for
		@param externalId The external ID of the variant, it must be 
			unique for all variants
		@return The new PlatformVariant item
	*/
	public static PlatformVariant getNew(DbControl dc, Platform platform, String externalId)
	{
		if (platform.isFileOnly())
		{
			return getNew(dc, platform, externalId, platform.getRawDataType().getChannels());
		}
		else
		{
			return getNew(dc, platform, externalId, platform.getRawDataType());
		}
	}
	
	/**
		Create a new file-only variant of a platform. Data for this variant 
		can't be imported to the database.
		
		@param dc The DbControl to use for database access
		@param platform The platform to create a variant for
		@param externalId The external ID of the variant, it must be 
			unique for all variants
		@param channels Number of channels in this variant
		@return The new variant object
	*/
	public static PlatformVariant getNew(DbControl dc, Platform platform, String externalId, int channels)
	{
		PlatformVariant v = dc.newItem(PlatformVariant.class);
		v.setPlatform(platform);
		v.getData().setFileOnly(true);
		v.getData().setRawDataType("variant." + externalId);
		v.setExternalId(externalId);
		v.setChannels(channels);
		return v;
	}
	
	/**
		Create a new platform variant which may optionally import data 
		into the database.
		
		@param dc The DbControl to use for database access
		@param platform The platform to create a variant for
		@param externalId The external ID of the platform variant, must be 
			unique for all variants (including those for other platforms)
		@param rawDataType A specific raw data type this platform will be locked
			to or null if it is not locked to any specific raw data type
		@return The new platform object
	*/
	public static PlatformVariant getNew(DbControl dc, Platform platform, String externalId, RawDataType rawDataType)
	{
		PlatformVariant v = dc.newItem(PlatformVariant.class);
		v.setPlatform(platform);
		v.getData().setFileOnly(false);
		v.setExternalId(externalId);
		v.setRawDataType(rawDataType);
		return v;
	}	
	
	/**
		Get a <code>PlatformVariant</code> object when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>PlatformVariant</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the items
		@throws BaseException If there is another error
	*/
	public static PlatformVariant getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PlatformVariant p = dc.loadItem(PlatformVariant.class, id);
		if (p == null) throw new ItemNotFoundException("PlatformVariant[id="+id+"]");
		return p;
	}
	
	/**
		Get a <code>PlatformVariant</code> item when you know the external ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param externalId The external ID of the item to load
		@return The <code>PlatformVariant</code> item.
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the item
		@throws BaseException If there is another error
	*/
	public static PlatformVariant getByExternalId(DbControl dc, String externalId)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"GET_PLATFORMVARIANT_FOR_EXTERNAL_ID");
		/*
			SELECT plv
			FROM PlatformVariantData plv
			WHERE plv.externalId = :externalId
		*/
		query.setString("externalId", externalId);
		PlatformVariant v = dc.getItem(PlatformVariant.class, HibernateUtil.loadData(PlatformVariantData.class, query));
		if (v == null) throw new ItemNotFoundException("PlatformVariant[externalId="+externalId+"]");
		return v;
	}	
	
	/**
		Get a query configured to retrieve <code>PlatformVariant</code> items.
		@return An {@link ItemQuery} object
	 	@throws BaseException If the query could not be created. 
	*/
	public static ItemQuery<PlatformVariant> getQuery()
		throws BaseException
	{
		return new ItemQuery<PlatformVariant>(PlatformVariant.class, RUNTIME_FILTER);
	}

	/**
		Creates a new filetype item from the given data.
	
		@param data the data.
	*/
	PlatformVariant(PlatformVariantData data)
	{
		super(data);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
		if (isFileOnly() && isInDatabase()) 
		{
			getRawDataType().setName(name);
		}
	}
	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	@Override
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	@Override
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>any raw bioassays or array design are using this platform variant
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_RAWBIOASSAYS_FOR_VARIANT", "count(*)");
		/*
			SELECT {1}
			FROM RawBioAssayData rba
			WHERE rba.variant = :variant
		*/
		query.setEntity("variant", this.getData());
		long count = HibernateUtil.loadData(Long.class, query);
		if (count == 0)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ARRAYDESIGNS_FOR_VARIANT", "count(*)");
				/*
					SELECT {1}
					FROM ArrayDesignData ad
					WHERE ad.variant = :variant
				*/
			query.setEntity("variant", this.getData());
			count = HibernateUtil.loadData(Long.class, query);			
		}
		return count > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link ArrayDesign}:s and {@link RawBioAssay}:s using
		this platform variant
		<ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_RAWBIOASSAYS_FOR_VARIANT", "rba.id");
			/*
				SELECT {1}
				FROM RawBioAssayData rba
				WHERE rba.variant = :variant
			*/
		query.setEntity("variant", this.getData());
		addUsingItems(using, Item.RAWBIOASSAY, query);
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ARRAYDESIGNS_FOR_VARIANT", "ad.id");
			/*
				SELECT {1}
				FROM ArrayDesignData ad
				WHERE ad.variant = :variant
			*/
		query.setEntity("variant", this.getData());
		addUsingItems(using, Item.ARRAYDESIGN, query);		
		return using;
	}
	/**
		Register file-only platforms as raw data type after insert and unregister
		after delete.
		@since 2.16.2
	*/
	@Override
	void onAfterCommit(Action action)
	{
		super.onAfterCommit(action);
		if (action == Action.CREATE)
		{
			if (isFileOnly()) RawDataTypes.registerFileOnlyRawDataType(new RawDataType(getData()));
		}
		else if (action == Action.DELETE)
		{
			if (isFileOnly()) RawDataTypes.unregisterFileOnlyRawDataType(getRawDataType());
		}
	}
	// -------------------------------------------
	/*
		From the BasicChildItem class
		-------------------------------------------
	*/
	@Override
	Item getParentType()
	{
		return Item.PLATFORM;
	}
	// -------------------------------------------
	
	/**
		Get that platform this is a variant of.
		@return A <code>Platform</code> item
		@throws PermissionDeniedException  If the logged in user doesn't 
			have read permission to the item
	*/
	public Platform getPlatform()
	{
		return getDbControl().getItem(Platform.class, getData().getPlatform());
	}
	
	private void setPlatform(Platform platform)
	{
		if (platform == null) throw new InvalidUseOfNullException("platform");
		getData().setPlatform(platform.getData());
	}
	
	/**
		The maximum length of the external ID that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = PlatformVariantData.MAX_EXTERNAL_ID_LENGTH;
	
	/**
		Get the external id for the platform variant.
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	
	private void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		getData().setExternalId(StringUtil.setNullableString(
			externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}
	
	/**
		Check if the platform variant is a file-only variant or if it allows data
		to be imported into the database.
		@return TRUE if only file data is allowed, FALSE if import to database
			is allowed
	*/
	public boolean isFileOnly()
	{
		return getData().isFileOnly();
	}	
	
	/**
		Set the number of channels for a file-only platform variant.
	*/
	private void setChannels(int channels)
	{
		getData().setChannels(IntegerUtil.checkMin(channels, "channels", 1));
	}
	
	/**
		Set the raw data type. The raw data type must allow
		data to be imported into the database.
		@param rawDataType The raw data type, null is allowed
	*/
	private void setRawDataType(RawDataType rawDataType)
	{
		String rdt = null;
		if (rawDataType != null)
		{
			if (!rawDataType.isStoredInDb())
			{
				throw new InvalidDataException("Raw data type '" + rawDataType.getName() + 
					"' can't store data in the database");
			}
			rdt = rawDataType.getId();
		}
		getData().setRawDataType(rdt);
	}
	
	/**
		Get the raw data type this platform variant is locked to. File-only variants
		are always locked to a raw data type. Variants that support import of
		data to the database may or may not be locked. Note! If the variant is
		a file-only variant that hasn't been committed to the database a 
		new instance of {@link RawDataType} is returned for each call to this
		method.
		@return The raw data type or null if the platform isn't locked
	*/
	public RawDataType getRawDataType()
	{
		return isInDatabase() ?
				RawDataTypes.getRawDataType(getData().getRawDataType()) :
				new RawDataType(getData());
	}
	
	/**
		Get a query returning file types registered with this platform
		variant.
		@param restrict If TRUE the query will only return file types that
			has been directly registered with this variant, if FALSE the
			query will also return file types that have been registered 
			with the platform
		@see PlatformFileType#getQuery(Platform, PlatformVariant, boolean)
	*/
	public ItemQuery<PlatformFileType> getFileTypes(boolean restrict)
	{
		return PlatformFileType.getQuery(getPlatform(), this, restrict);
	}
	

}
