/*
	$Id: PluginRequest.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.JobData;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;

import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;

/**
	Configures or executes a plugin. You will get an instance of
	a subclass to this class when:
	
	<ol>
	<li>You want to configure a plugin: {@link PluginConfiguration#configure()}
	<li>You want to configure a job: {@link Job#configure(GuiContext)}
	<li>You want to execute a job: {@link Job#execute(ProgressReporter, String)}
	</ol>
	
	@see PluginConfigurationRequest
	@see PluginExecutionRequest
	@author Nicklas
	@version 2.0
	@base.modified $Date$
*/
abstract class PluginRequest<P extends Plugin>
{

	private final SessionControl sc;

	private String command;
	
	private final Map<String, List<?>> requestParameters;
	
	private final P plugin;
	
	private final ParameterValuesImpl configurationParameters;
	
	private final PluginConfiguration pluginConfiguration;
	
	private final PluginDefinition pluginDefinition;
	
	private final ParameterValuesImpl jobParameters;
	
	private final Map<String, PluginParameter<?>> parameterInfo;
	
	private final Job job;
	
	PluginRequest(SessionControl sc, P plugin, String command, 
		PluginConfiguration pluginConfiguration, ParameterValuesImpl configurationParameters,
		Job job, ParameterValuesImpl jobParameters, PluginDefinition pluginDefinition)
	{
		this.sc = sc;
		this.plugin = plugin;
		this.command = command;
		this.pluginConfiguration = pluginConfiguration;
		this.pluginDefinition = pluginDefinition;
		this.configurationParameters = configurationParameters;
		this.job = job;
		this.jobParameters = jobParameters;
		this.requestParameters = new HashMap<String, List<?>>();
		this.parameterInfo = new HashMap<String, PluginParameter<?>>();
		plugin.init(sc.getPluginSessionControl(pluginDefinition, job), configurationParameters, jobParameters);
	}

	PluginRequest(PluginRequest<? extends P> copyFrom, String command)
	{
		this.sc = copyFrom.sc;
		this.plugin = copyFrom.plugin;
		this.command = command;
		this.pluginConfiguration = copyFrom.pluginConfiguration;
		this.pluginDefinition = copyFrom.pluginDefinition;
		this.configurationParameters = copyFrom.configurationParameters;
		this.job = copyFrom.job;
		this.jobParameters = copyFrom.jobParameters;
		this.requestParameters = new HashMap<String, List<?>>();
		this.parameterInfo = new HashMap<String, PluginParameter<?>>();
	}
	
	/**
		@return The SessionControl used in this request
		@since 2.11
	*/
	public SessionControl getSessionControl()
	{
		return sc;
	}
	
	/**
		Get the command this request will issue to the plugin when
		{@link #invoke()} is called.
	*/
	public String getCommand()
	{
		return command;
	}

	/**
		Set the value of a request parameter to a single value. Normally a plugin
		validates the value using the information available in the 
		{@link RequestInformation} object.
		
		@param name The name of the parameter
		@param value The value of the parameter
		@see PluginConfigurationRequest#getRequestInformation()
	*/
	public void setParameterValue(String name, Object value)
	{
		requestParameters.put(name, Collections.singletonList(value));
	}

	/**
		Set the value of a request parameter to a list of values. Normally a plugin
		validates the value using the information available in the 
		{@link RequestInformation} object.
		
		@param name The name of the parameter
		@param values The values of the parameter
		@see PluginConfigurationRequest#getRequestInformation()
	*/
	public void setParameterValues(String name, List<?> values)
	{
		requestParameters.put(name, values);
	}

	public Object getCurrentParameterValue(String name)
	{
		List<?> values = getCurrentParameterValues(name);
		return values == null || values.size() == 0 ? null : values.get(0);
	}
	
	/**
		Get the values that are currently stored in the job or plugin configuration
		related to this request. This method will first check the job and then the
		plugin configuration.
		@param name The name of the parameter
		@return The parameter values, or null if no parameter with the given name
			exists
	*/
	public List<?> getCurrentParameterValues(String name)
	{
		List<?> values = null;
		if (jobParameters != null)
		{
			values = jobParameters.getValues(name);
		}
		if (values == null && configurationParameters != null)
		{
			values = configurationParameters.getValues(name);
		}
		return values;
	}
	
	
	/**
		Get the values that are currently stored in the plugin configuration
		related to this request.
		@param name The name of the parameter
		@return The parameter values, or null if no parameter with the given name
			exists or if this request isn't associated with a plugin configuration
	*/
	public List<?> getCurrentConfigurationParameterValues(String name)
	{
		List<?> values = null;
		if (configurationParameters != null)
		{
			values = configurationParameters.getValues(name);
		}
		return values;
	}
	
	/**
		Get the values that are currently stored in the job related to this
		request.
		@param name The name of the parameter
		@return The parameter values, or null if no parameter with the given name
			exists or if this request isn't associated with a job
	*/
	public List<?> getCurrentJobParameterValues(String name)
	{
		List<?> values = null;
		if (jobParameters != null)
		{
			values = jobParameters.getValues(name);
		}
		return values;
	}
	
	/**
		Invoke the plugin and let it do it's work. Note that
		this method doesn't throw any exceptions. If an exception
		occurs during the execution of a plugin, the response
		status will be set to {@link net.sf.basedb.core.plugin.Response.Status#ERROR} and
		the error messages will be available in the 
		{@link PluginResponse#getMessage()} and 
		{@link PluginResponse#getErrorList()} methods.
		
		@return A <code>PluginResponse</code> object
	*/
	public abstract PluginResponse invoke();
	
	/**
		Aborts the execution of the plugin.
	*/
	public void done()
	{
		plugin.done();
	}
	
	/**
		Keep track of all parameters so that we can save label and
		description in the {@link ParameterValuesImpl#saveParameters(Job, Map)}
		method.
		@param parameters
	*/
	void registerParameters(List<PluginParameter<?>> parameters)
	{
		if (parameters != null)
		{
			for (PluginParameter<?> pp : parameters)
			{
				parameterInfo.put(pp.getName(), pp);
			}
		}
	}
	
	/**
		Save all parameters that have got values in this request/response session.
	*/
	void saveParameters(DbControl dc, GuiContext context)
	{
		if (jobParameters != null && !jobParameters.isWriteProtected() && job != null)
		{
			Job theJob = job.isInDatabase() ? Job.getById(dc, job.getId()) : job;
			JobData jobData = theJob.getData();
			if (pluginConfiguration != null)
			{
				jobData.setParameterVersion(pluginConfiguration.getParameterVersion());
			}
			if (context != null)
			{
				jobData.setContextItem(context.getItem().getValue());
				jobData.setContextType(context.getType().getValue());
				jobData.setSubContext(context.getSubContext());
				
			}
			else
			{
				jobData.setContextItem(null);
				jobData.setContextType(null);
				jobData.setSubContext(null);
			}
			jobParameters.saveParameters(theJob, parameterInfo);
		}
		if (configurationParameters != null && !configurationParameters.isWriteProtected() && pluginConfiguration != null)
		{
			PluginConfiguration theConfig = pluginConfiguration.isInDatabase() ? PluginConfiguration.getById(dc, pluginConfiguration.getId()) : pluginConfiguration;
			configurationParameters.saveParameters(theConfig, parameterInfo);
		}
	}

	/**
		Sets the next command in a configuration sequence.
	*/
	void setNextCommand(String command)
	{
		this.command = command;
		requestParameters.clear();
	}
	
	/**
		Get the plugin object.
		@since 2.6 (was not public before)
	*/
	public P getPlugin()
	{
		return plugin;
	}
	
	/**
		Get the job for the request or null if we are not configuring/executing a job.
		@since 2.2
	*/
	Job getJob()
	{
		return job;
	}
	
	/**
		An implementation of the {@link Request} interface 
		that is passed to the plugin.
	*/
	class RequestImpl
		implements Request
	{
		private final boolean allowImmediateExecution;
		RequestImpl(boolean allowImmediateExecution)
		{
			this.allowImmediateExecution = allowImmediateExecution;
		}
		
		/*
			From the Request interface (plugin package)
			---------------------------------------------------
		*/
		@Override
		public String getCommand()
		{
			return command;
		}
		
		@Override
		public Object getParameterValue(String name)
		{
			Object value = null;
			List<?> values = requestParameters.get(name);
			if (values != null && values.size() != 0)
			{
				value = values.get(0);
			}
			return value;
		}
	
		@Override
		public List<?> getParameterValues(String name)
		{
			return requestParameters.get(name);
		}
		
		@Override
		public boolean isAllowedImmediateExecution()
		{
			return allowImmediateExecution;
		}
		// -------------------------------------------
	}
}
