/*
	$Id: PropertyFilter.java 6214 2012-12-13 10:23:32Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.data.UnitData;
import net.sf.basedb.core.query.Annotations;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.util.units.UnitUtil;

/**
	Used to store filter settings in the current context. A filter is
	a simple expression of the form: <code>property operator value</code>,
	ie. <code>name = 'Nicklas'</code> or <code>spots &gt; 10000</code>.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2012-12-13 11:23:32 +0100 (to, 13 dec 2012) $
	@see SessionControl#getCurrentContext(Item)
	@see ItemContext
*/
public class PropertyFilter
{

	/**
		The maximum allowed length of a filter value.
		@since 2.6.2
	*/
	public static final int MAX_VALUE_LENGTH = PropertyFilterData.MAX_VALUE_LENGTH;
	
	/**
		Flag to use for temporary filters.
		@since 2.16
	*/
	public static final boolean TEMPORARY = true;
	
	/**
	 	The string/characters used as a separator in a string with more then one filter values.
	 */
	private static final String value_separator = "|";
	
	private final Pattern ENTITY_COLLECTION = Pattern.compile("\\&([\\w\\.]+)\\((.+)\\)");
	
	private final String property;
	private Type valueType;
	private Operator operator;
	private String value;
	private Unit unit;
	private boolean temporary;
	
	/**
		Create a new filter for a property.
		@param property The name of the property
	*/
	public PropertyFilter(String property)
	{
		this.property = property;
	}

	/**
		Create a new filter for a property.
		@param property The name of the property
		@param operator The operator to use
		@param value The value to check against
		@param valueType The type of the value
	*/
	public PropertyFilter(String property, Operator operator, String value, Type valueType)
	{
		this.property = property;
		this.operator = operator;
		this.value = value;
		this.valueType = valueType;
	}
	
	/**
		Create a new filter for a property.
		@param property The name of the property
		@param operator The operator to use
		@param value The value to check against
		@param valueType The type of the value
		@param temporary A flag indicating if the filter should be temporary or not
		@since 2.16
	*/
	public PropertyFilter(String property, Operator operator, String value, Type valueType, boolean temporary)
	{
		this.property = property;
		this.operator = operator;
		this.value = value;
		this.valueType = valueType;
		this.temporary = temporary;
	}

	
	/**
		Create a new filter from a loaded database object.
	*/
	PropertyFilter(String property, PropertyFilterData filter)
	{
		this.property = property;
		this.valueType = Type.fromValue(filter.getValueType());
		this.operator = Operator.fromValue(filter.getOperator());
		this.value = filter.getValue();
		if (filter.getUnit() != null)
		{
			this.unit = new Unit(filter.getUnit());
			this.unit.getName(); // Forces initialisation if the unit is a proxy
		}
	}
	
	/**
		Create a new database filter object from this filter. 
	*/
	PropertyFilterData getData()
	{
		PropertyFilterData filter = new PropertyFilterData();
		filter.setValueType(valueType.getValue());
		filter.setOperator(operator.getValue());
		filter.setValue(value);
		filter.setUnit(unit == null ? null : unit.getData());
		return filter;
	}
	
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return property + " " + operator + " " + value;
	}
	// -------------------------------------------

	/**
		Set or clear a flag indicating that this filter is temporary filter.
		Temporary filters are not saved between sessions.
		@since 2.16
	*/
	public void setTemporary(boolean temporary)
	{
		this.temporary = temporary;
	}
	
	/**
	 	Is this filter a temporary filter or not.
		@since 2.16
	*/
	public boolean isTemporary()
	{
		return temporary;
	}
	
	/**
		Get the name of the property this filter is used on.
	*/
	public String getProperty()
	{
		return property;
	}
	
	/**
		Get the type of the property.
	*/
	public Type getValueType()
	{
		return valueType;
	}
	/**
		Set the type of the property.
		@param valueType The type
	*/
	public void setValueType(Type valueType)
	{
		this.valueType = valueType;
	}
	
	/**
		Get the operator to use in the filter.
	*/
	public Operator getOperator()
	{
		return operator;
	}
	/**
		Set the operator to use in the filter.
	*/
	public void setOperator(Operator operator)
	{
		this.operator = operator;
	}
	
	/**
		Sets the unit the values for the property are stored in the
		database. When a unit is set, it is possible to have filter values
		with units. The filter values are automatically converted to this
		unit before the filter is applied. For example, if we set the unit to
		'g', someone may filter on '0.5kg', which is converted to 500 in
		the query.
		@param unit The unit, or null to remove the unit from the filter
		@since 2.9
	*/
	public void setUnit(Unit unit)
	{
		this.unit = unit;
	}
	
	/**
		Get the value to use in the filter.
		@return The value as a string
	*/
	public String getValue()
	{
		return value;
	}
	/**
		Set the value to use in the filter.
		@param value The value as a string
	*/
	public void setValue(String value)
	{
		this.value = value;
	}
	
	/**
		Get the value as an object of the correct type. If the 
		filter property is numeric and has a unit, the value is converted
		to the correct	unit automatically. For example, if the value is '5kg'
		and the unit is 'g' this method returns 5000. 
		<p>
		Note! The returned value is cast to the class supported by
		the {@link #getValueType()}. Beware of this when the 
		filter has an integer type and a unit with low precision.
		For example, if the value is '125min' and the unit is
		'hour' this method returns the integer 2. To avoid this
		use {@link #getValueAsObject(Unit, boolean)} with 
		asDouble=true instead.
		
		@return The value as an object
		@throws InvalidDataException If the string value can't be parsed
	*/
	public Object getValueAsObject()
	{
		return getValueAsObject(unit, false);
	}
	
	/**
		Get the value as an object of the proper type converted to the
		specified unit. For example, if the unit is 'g' and the
		value is '5kg' this method returns 5000.
		<p>
		If the asDouble parameter is true, the value is returned as a 
		double, otherwise it is cast to the
		proper class as specified by the {@link #getValueType()}. 
		<p>
		Note! The above only applies for numerical properties that
		has a unit. If not, the value is simply parsed with 
		{@link Type#parseString(String)}.
		
		@param unit A unit, or null to use the default unit
		@param asDouble TRUE to return the value as a double, FALSE
			to let the {@link #getValueType()} decide
		@return The value as an object
		@since 2.9
	*/
	public Object getValueAsObject(Unit unit, boolean asDouble)
	{
		Object oValue = null;
		if (unit == null) unit = this.unit;
		if (unit != null && value != null)
		{
			String[] valueAndUnit = UnitUtil.splitValueWithUnit(value);
			if (valueAndUnit != null)
			{
				boolean newSession = unit.isDetached();
				org.hibernate.Session session = null; 
				try
				{
					session = newSession ? HibernateUtil.newSession() : unit.getDbControl().getHibernateSession();
					Double dbl = convertValueToUnit(session, valueAndUnit[0], valueAndUnit[1], unit);
					oValue = asDouble ? dbl : valueType.convertNumber(dbl);
				}
				finally
				{
					if (newSession && session != null) HibernateUtil.close(session);
				}
			}
		}
		if (oValue == null)
		{
			oValue = valueType.parseString(value);
		}
		return oValue;
	}
	
	/**
	 	Splits the value string at each "|" and creates
	 	a list with objects from the substrings that were found.
	 	@return A list with objects, or null if splitting string of values returns null.
	 	@throws InvalidDataException If any of the string values can't be parsed.
	 	@since 2.5 
	 */
	public List<Object> getValuesAsObjects()
	{
		return getValuesAsObjects(unit, false);
	}
	
	/**
		Same as {@link #getValueAsObject(Unit, boolean)} but the filter string
		is first split at each "|". Each individual string may or may not specify
		a unit. If a unit is specified for some parts but not for all, the nearest
		following unit is used. For example, the string '300g|5|10kg' is the same as
		'300g|5kg|10kg'.'

		@param unit A unit, or null to use the default unit
		@param asDouble TRUE to return the value as a double, FALSE
			to let the {@link #getValueType()} decide
		@return A list with the values
		@since 2.9
	*/
	public List<Object> getValuesAsObjects(Unit unit, boolean asDouble)
	{
		if (value == null) return Collections.emptyList();
		String[] values = value.split("\\"+getValueSeparator(), -1);
		Object[] objects = new Object[values.length];
		
		if (unit == null) unit = this.unit;

		boolean newSession = unit != null && unit.isDetached();
		org.hibernate.Session session = null; 
		try
		{
			if (unit != null)
			{
				session = newSession ? HibernateUtil.newSession() : unit.getDbControl().getHibernateSession();
			}
			
			String lastUnit = null;
			for (int i = values.length - 1; i >= 0; --i)
			{
				String v = values[i];
				if (v.length() == 0) v = null;
				
				Object oValue = null;
				if (unit != null)
				{
					String[] valueAndUnit = UnitUtil.splitValueWithUnit(v);
					if (valueAndUnit != null)
					{
						if (valueAndUnit[1] != null) lastUnit = valueAndUnit[1];
						Double dbl = convertValueToUnit(session, valueAndUnit[0], lastUnit, unit);
						oValue = asDouble ? dbl : valueType.convertNumber(dbl);
					}
				}
				if (oValue == null)
				{
					oValue = valueType.parseString(v);
				}
				objects[i] = oValue;
			}
		}
		finally
		{
			if (newSession && session != null) HibernateUtil.close(session);
		}
		return Arrays.asList(objects);
	}
	
	/**
		Converts a string to a numeric value with the correct unit.
		@param session Hibernate session used to lookup units 
		@param value The string value which should be parsable with
			{@link Double#parseDouble(String)}
		@param unitSymbol The unit of the string value
		@param convertTo The unit to convert the value to
		@return The converted value
	*/
	private Double convertValueToUnit(org.hibernate.Session session, String value, String unitSymbol, Unit convertTo)
	{
		Double result = null;
		if (unitSymbol != null)
		{
			UnitData other = convertTo.findOtherUnit(session, unitSymbol);
			if (other != null)
			{
				result = convertTo.fromOther(Double.parseDouble(value), new Unit(other));
			}
		}
		else
		{
			result = Double.parseDouble(value);
		}
		return result;
	}
	
	/**
		Create a restriction from the filter. There are a few special cases:
		<p>
		If the property starts with a $ it is interpreted as a <code>alias.property</code> 
		value, ignoring the root alias of the query. The alias is removed from the
		property before proceeding.
		<p>
		If the property starts with a # or ## it is a filter on annotations. The rest of
		the property is the ID of the annotation type, and we use a
		{@link AnnotationSimpleRestriction} object for the restriction. If there
		is only one # in the prefix only direct annotations are searched, if there
		are two ##, inherited annotations are also searched.
		<p>
		If the property starts with a £, the rest of the property is the path that
		joins the {@link ReporterData#getReporterListScores()} collection. The filter
		value is the ID of the reporter list.
		<p>
		If the property starts with a @ the property is a collection of values and
		we always use an IN restriction, ignoring the specified operator.
		<p>
		If the property starts with & it is a collection of entities. The property
		must end with a property name from the entity in the collection within paranthesis.
		Eg. &experiments(name). The filter is applied as a subquery.
		<p>
		If the property starts with ! it is another special filter.	There is currently 
		one such filter (since BASE 2.15): !sharedTo.name, which is used to filter 
		shareable items by the name of users/groups/projects it has been shared to. 
		The 'name' part can be replaced with any other property that is common for 
		users, groups and projects (eg. description or id). 
		<p>
		The property may also start with $@.
		@return The restriction
		@since 2.8
	*/
	public Restriction getRestriction(DbControl dc, EntityQuery query)
	{
		
		Restriction restriction = null;
		String property = getProperty();
		String alias = null;
		
		if (property != null && property.startsWith("$"))
		{
			// Property is alias.property or alias only
			int dotIndex = property.indexOf(".");
			if (dotIndex == -1)
			{
				// Alias only
				alias = property.substring(1);
				property = null;
			}
			else if (property != null)
			{
				// Alias.property
				alias = property.substring(1, dotIndex);
				property = property.substring(dotIndex+1);
			}
		}
		
		if (property != null && property.startsWith("#"))
		{
			// Property is id of annotation type: #id
			// If prefixed with two ## inherited annotations should be searched
			boolean searchInherited = property.startsWith("##");
			int annotationTypeId = (Integer)Type.INT.parseString(property.substring(searchInherited ? 2 : 1));
			
			AnnotationType at = AnnotationType.getById(dc, annotationTypeId);
			Unit unit = at.getDefaultUnit();
			
			if (at.getMultiplicity() == 1 && !searchInherited)
			{
				// This is similar to a regular property expression
				Expression propertyExpression = Annotations.expression(alias, at);
				if (getValueType() == Type.TIMESTAMP)
				{
					propertyExpression = Expressions.toDate(propertyExpression);
				}
				if (operator.isListOperator())
				{
					List<Expression> parameters = new ArrayList<Expression>();
					List<Object> objects = getValuesAsObjects(unit, true);
					for (Object obj : objects)
					{
						Expression e = null;
						if (obj != null)
						{
							String parameterName = "p" + System.identityHashCode(obj);
							Type theValueType = obj instanceof Double ? Type.DOUBLE : getValueType();
							if (theValueType == Type.TIMESTAMP) theValueType = Type.DATE;
							e = Expressions.parameter(parameterName, obj, theValueType);
						}
						parameters.add(e);
					}
					restriction = operator.getRestriction(propertyExpression, parameters.toArray(new Expression[parameters.size()]));
				}
				else
				{
					String parameterName = "p" + System.identityHashCode(this);
					Expression parameter = null;
					if (getValue() != null)
					{
						Object value = getValueAsObject(unit, true);
						Type theValueType = value instanceof Double ? Type.DOUBLE : getValueType();
						// We only search on the date part in timestamp values
						if (theValueType == Type.TIMESTAMP) theValueType = Type.DATE;
						parameter = Expressions.parameter(parameterName, value, theValueType);
					}
					restriction = operator.getRestriction(propertyExpression, parameter);
				}
			}
			else
			{
				if (operator.isListOperator())
				{
					List<Object> objects = getValuesAsObjects(unit, true);
					if (operator == Operator.BETWEEN)
					{
						if (objects.size() != 2)
						{
							throw new InvalidDataException("Two values are needed for >< operator");
						}
						restriction = new AnnotationBetweenRestriction(alias, annotationTypeId, getValueType(),
								objects.get(0), objects.get(1), searchInherited);
					}
					else
					{
						restriction = new AnnotationSimpleRestriction(alias,annotationTypeId, getValueType(), 
								operator, objects, searchInherited);
					}
					
				}
				else
				{
					Object value = getValueAsObject(unit, true);
					Operator operator = getOperator();
					if (value == null && (operator == Operator.EQ || operator == Operator.NEQ))
					{
						restriction = new HasAnnotationRestriction(alias, annotationTypeId, 
							operator != Operator.EQ, searchInherited);
					}
					else
					{
						restriction = new AnnotationSimpleRestriction(alias, annotationTypeId, getValueType(), 
								operator, value, searchInherited);
					}
				}
			}
		}
		
		else if (property != null && property.startsWith("£"))
		{
			// Property is reporterListScores collection
			// The value is the ID of the reporter list to
			// filter on
			Integer reporterListId = (Integer)getValueAsObject();
			if (reporterListId != null)
			{
				try
				{
					ReporterList list = ReporterList.getById(dc, reporterListId);
					if (operator == Operator.NEQ)
					{
						restriction = Hql.isNotPartOf(alias, property.substring(1), list);
					}
					else
					{
						restriction = Hql.isPartOf(alias, property.substring(1), list);
					}
				}
				catch (ItemNotFoundException ex)
				{}
			}
		}
		
		else
		{
			String parameterName = "p" + System.identityHashCode(this);
			Expression parameter = null;
			if (getValue() != null && !operator.isListOperator())
			{
				Object value = getValueAsObject(unit, true);
				Type theValueType = value instanceof Double ? Type.DOUBLE : getValueType();
				// We only search on the date part in timestamp values
				if (theValueType == Type.TIMESTAMP) theValueType = Type.DATE;
				parameter = Expressions.parameter(parameterName, value, theValueType);
			}

			if (property != null && property.startsWith("@"))
			{
				// Property is a collection - we only support equal and not-equal tests
				property = property.substring(1);
				Expression elements = Hql.elements(alias, property);
				Restriction isEmpty = Restrictions.eq(Hql.size(alias, property), Expressions.integer(0));
				if (operator.isListOperator())
				{
					// filter to match one of several elements in a collection
					List<Restriction> alternatives = new ArrayList<Restriction>();
					List<Object> objects = getValuesAsObjects(unit, true);
					boolean hasEmptyCheck = false;
					for (Object obj : objects)
					{
						if (obj != null)
						{
							parameterName = "p" + System.identityHashCode(obj);
							Type theValueType = obj instanceof Double ? Type.DOUBLE : getValueType();
							Expression e = Expressions.parameter(parameterName, obj, theValueType);
							alternatives.add(Restrictions.in(e, elements));
						}
						else if (!hasEmptyCheck)
						{
							hasEmptyCheck = true;
							alternatives.add(isEmpty);
						}
					}
					restriction = Restrictions.or(alternatives.toArray(new Restriction[alternatives.size()]));
				}
				else if (parameter != null)
				{
					// filter for a single element in a collection
					restriction = Restrictions.in(parameter, elements);
					if (operator == Operator.NEQ)
					{
						restriction = Restrictions.not(restriction);
					}
				}
				else
				{
					// filter that returns items with empty collections
					restriction = isEmpty;
					if (operator == Operator.NEQ)
					{
						restriction = Restrictions.not(restriction);
					}
				}
			}
			else if (property != null && property.startsWith("&"))
			{
				// Property is an entity collection: collection(filter-property)
				Matcher m = ENTITY_COLLECTION.matcher(property);
				if (!m.matches()) 
				{
					throw new InvalidDataException("Invalid collection filter: " + property);
				}
				String entityCollection = m.group(1);
				String collectionFilter = m.group(2);
				
				/*
					We are going to create a subquery like:
					root-entity = ANY(
						SELECT root-entity FROM root-entity
						LEFT JOIN collection-path cp 
						WHERE cp.filter-property ........
					)
					
					- or -
					
					root-entity <> ALL(
						SELECT root-entity FROM root-entity
						LEFT JOIN collection-path cp 
						WHERE cp.filter-property ........
					)
					
				*/
				Query subquery = new AbstractEntityQuery(query.getItemType(), null, false, null)
				{};
				// LEFT JOIN the parts in the collection path
				String[] cPath = entityCollection.split("\\.");
				String thisAlias = null;
				for (int i = 0; i < cPath.length; ++i)
				{
					String nextAlias = "cp" + i;
					subquery.join(Hql.leftJoin(thisAlias, cPath[i], nextAlias, null, false));
					thisAlias = nextAlias;
				}
				// LEFT JOIN the parts in the filter-property path (except the last one)
				cPath = collectionFilter.split("\\.");
				for (int i = 0; i < cPath.length - 1; ++i)
				{
					String nextAlias = "fp" + i;
					subquery.join(Hql.leftJoin(thisAlias, cPath[i], nextAlias, null, false));
					thisAlias = nextAlias;
				}
				collectionFilter = cPath[cPath.length - 1];
				
				// The following operators should result in "not contains" logic
				Operator subfilterOperator = operator;
				if (operator == Operator.NEQ)
				{
					subfilterOperator = Operator.EQ;
				}
				else if (operator == Operator.NOTIN)
				{
					subfilterOperator = Operator.IN;
				}
				else if (operator == Operator.NOTLIKE)
				{
					subfilterOperator = Operator.LIKE;
				}
				else if (operator == Operator.NOTLIKE_IN)
				{
					subfilterOperator = Operator.LIKE_IN;
				}
				
				// Restrict the subquery
				PropertyFilter pp = new PropertyFilter("$" + thisAlias + "." + collectionFilter, 
						subfilterOperator, getValue(), getValueType());
				subquery.restrict(pp.getRestriction(dc, query));
				
				// Create the main restriction
				if (subfilterOperator != operator)
				{
					// "not contains": root-entity <> ALL(subquery)
					restriction = Restrictions.neq(Hql.alias(query.getRootAlias()), Expressions.all(subquery));
				}
				else
				{
					// "contains": root-entity = ANY(subquery)
					restriction = Restrictions.eq(Hql.alias(query.getRootAlias()), Expressions.any(subquery));
				}
			}
			else if (property != null && property.startsWith("!sharedTo."))
			{
				// A special filter on on shareable items
				String filterProperty = property.substring("!sharedTo.".length());
				String filterValue = getValue();
				Operator filterOp = null;
				if (operator == Operator.EQ || operator == Operator.NEQ)
				{
					filterOp = Operator.EQ;
				}
				else if (operator == Operator.LIKE || operator == Operator.NOTLIKE)
				{
					filterOp = Operator.LIKE;
				}
				else if (operator == Operator.IN ||  operator == Operator.NOTIN)
				{
					filterOp = Operator.IN;
				}
				else if (operator == Operator.LIKE_IN ||  operator == Operator.NOTLIKE_IN)
				{
					filterOp = Operator.LIKE_IN;
				}
				else
				{
					filterOp = operator;
				}
				boolean sharedTo = operator == filterOp;
				
				Restriction users = null;
				Restriction groups = null;
				Restriction projects = null;
				if (filterValue != null)
				{
					users = new PropertyFilter("$usr." + filterProperty, 
							filterOp, filterValue, valueType).getRestriction(dc, query);
					groups = new PropertyFilter("$grp." + filterProperty, 
							filterOp, filterValue, valueType).getRestriction(dc, query);
					projects = new PropertyFilter("$prj." + filterProperty, 
							filterOp, filterValue, valueType).getRestriction(dc, query);
				}
				restriction = Hql.sharedTo(sharedTo, users, groups, projects);
			}
			else
			{
				Expression propertyExpression = Hql.property(alias, property);
				if (getValueType() == Type.TIMESTAMP)
				{
					propertyExpression = Expressions.toDate(propertyExpression);
				}
				if (operator.isListOperator())
				{
					List<Expression> parameters = new ArrayList<Expression>();
					List<Object> objects = getValuesAsObjects(unit, true);
					for (Object obj : objects)
					{
						Expression e = null;
						if (obj != null)
						{
							parameterName = "p" + System.identityHashCode(obj);
							Type theValueType = obj instanceof Double ? Type.DOUBLE : getValueType();
							if (theValueType == Type.TIMESTAMP) theValueType = Type.DATE;
							e = Expressions.parameter(parameterName, obj, theValueType);
						}
						parameters.add(e);
					}
					restriction = operator.getRestriction(propertyExpression, parameters.toArray(new Expression[parameters.size()]));
				}
				else
				{
					restriction = operator.getRestriction(propertyExpression, parameter);
				}
			}
		}
		return restriction;
	}

	public Restriction getDynamicRestriction(DbControl dc)
	{
		Restriction restriction = null;
		String property = getProperty();
		Expression propertyExpression = null;
		if (property.startsWith("£"))
		{
			// Property is reporterListScores collection
			// The value is the ID of the reporter list to
			// filter on
			Integer reporterListId = (Integer)getValueAsObject();
			if (reporterListId != null)
			{
				try
				{
					ReporterList list = ReporterList.getById(dc, reporterListId);
					if (operator == Operator.NEQ)
					{
						restriction = Dynamic.isNotPartOf(list);
					}
					else
					{
						restriction = Dynamic.isPartOf(list);
					}
				}
				catch (ItemNotFoundException ex)
				{}
			}
			return restriction;
		}
		else
		{
			propertyExpression = ItemContext.getDynamicExpression(dc, property);
		}
		String parameterName = "p" + System.identityHashCode(this);
		
		String value = getValue();
		Expression parameter = null;
		List<Expression> parameters = null;
		// Experimental - allows expressions like: >ch1
		if (value != null && (valueType == Type.FLOAT || valueType == Type.DOUBLE || valueType == Type.INT || valueType == Type.LONG))
		{
			if (!getOperator().isListOperator() && value.matches("^ch\\d+$"))
			{
				int channel = Integer.parseInt(value.substring(2));
				parameter = Dynamic.column(VirtualColumn.channelRaw(channel));
			}
			else if (getOperator().isListOperator())
			{
				String[] values = value.split("\\"+getValueSeparator());
				for (String s : values)
				{
					if (s.matches("^ch\\d+$"))
					{
						int channel = Integer.parseInt(s.substring(2));
						if (parameters == null)
						{
							parameters = new ArrayList<Expression>();
						}
						parameters.add(Dynamic.column(VirtualColumn.channelRaw(channel)));
					}
				}
			}			
		}
		if (parameter == null && parameters == null)
		{
			if (getOperator().isListOperator())
			{
				List<Object> objects = getValuesAsObjects(unit, true);				
				parameters = objects == null ? null : new ArrayList<Expression>(objects.size());
				for (Object obj : objects)
				{
					Expression e = null;
					if (obj != null)
					{
						parameterName = "p" + System.identityHashCode(obj);
						e = Expressions.parameter(parameterName, obj);
					}
					parameters.add(e);
				}				
			}
			else
			{
				parameter = value == null ? null : Expressions.parameter(parameterName, getValueAsObject(unit, true));
			}
		}
		if (propertyExpression != null)
		{
			if (parameters != null)
			{
				restriction = operator.getRestriction(propertyExpression, parameters.toArray(new Expression[parameters.size()]));
			}
			else
			{
				restriction = operator.getRestriction(propertyExpression, parameter);
			}
		}
		return restriction;
	}
	
	public String getValueSeparator()
	{
		return value_separator;
	}

}
