/*
	$Id: QueryExecutor.java 6127 2012-09-14 07:28:35Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import net.sf.basedb.core.hibernate.StatelessSessionWrapper;

import org.hibernate.Query;
import org.hibernate.ScrollMode;
import org.hibernate.ScrollableResults;
import org.hibernate.StatelessSession;
import org.hibernate.impl.StatelessSessionImpl;


/**
	Utility class for executing SQL and Hibernate queries. If enabled,
	this class will use a thread pool to execute a query in separate
	threads. The benefits are:
	
	<ul>
	<li>It is possible to limit the number of concurrent queries.
	<li>It is possible to abort a running query by calling Thread.interrupt()
		on the thread that called one of the methods in this class.
	</ul>

	@author Nicklas
	@version 2.6
	@base.modified $Date: 2007-08-17 09:18:29 +0200 (fr, 17 aug 2007) $
*/
public class QueryExecutor 
{

	private static ExecutorService service;

	/**
		This will create a new thread for each query. There is no limit on the
		number of threads.
	*/
	static void init()
	{
		if (service != null) return;
		service = Executors.newCachedThreadPool();
	}
	
	static void unload()
	{
		if (service != null) service.shutdown();
		service = null;
	}
	
	/**
		Execute a query by calling {@link PreparedStatement#executeQuery()}.
		@param ps The prepared statement to execute
		@param sc A session control object that should be kept alive while the
			query is executing, or null
		@return The result from {@link PreparedStatement#executeQuery()}
		@throws InterruptedException If this thread was interrupted
		@throws SQLException If there was an error executing the SQL statement
	*/
	public static ResultSet executeQuery(PreparedStatement ps, SessionControl sc)
		throws InterruptedException, SQLException
	{
		return getFutureResult(new FutureResultSet(ps), new CancelableStatement(ps), sc);
	}
	
	/**
		Execute a query by calling {@link PreparedStatement#executeUpdate()}.
		@param ps The prepared statement to execute
		@param sc A session control object that should be kept alive while the
			query is executing, or null
		@return The result from {@link PreparedStatement#executeUpdate()}
		@throws InterruptedException If this thread was interrupted
	*/
	public static int executeUpdate(PreparedStatement ps, SessionControl sc)
		throws InterruptedException, SQLException
	{
		Integer result = getFutureResult(new FutureUpdate(ps), new CancelableStatement(ps), sc);
		return result == null ? 0 : result;
	}
	
	/**
		Executes a Hibernate query that returns an iterator on a stateless session.
		@param clazz The class of items returned by the iterator
		@param query The query to execute
		@param session The stateless session, this must be the same session
			that was used to create the query
		@param sc A session control object that should be kept alive while the
			query is executing, or null
		@since 2.7
	*/
	public static <T> ScrollIterator<T> loadIterator(Class<T> clazz, Query query, 
		StatelessSession session, SessionControl sc)
		throws InterruptedException
	{
		try
		{
			return getFutureResult(new FutureScrollIterator<T>(clazz, query), 
				new CancelableStatelessSession(session), sc);
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Executes a Hibernate query on a stateless session. If the query 
		returns more than one item, only the first one is returned.

		@param clazz The class of items returned by the iterator
		@param query The query to execute
		@param session The stateless session, this must be the same session
			that was used to create the query
		@param sc A session control object that should be kept alive while the
			query is executing, or null
		@since 2.7
	*/
	public static <T> T loadData(Class<T> clazz, Query query, 
		StatelessSession session, SessionControl sc)
		throws InterruptedException
	{
		try
		{
			return getFutureResult(new FutureItem<T>(clazz, query), 
				new CancelableStatelessSession(session), sc);
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
	}

	
	/**
		Run the callable and wait for the result. Call Thread.interrupt() on this
		thread to abort the query.
		
		@param <T> The type of result that is returned by the callable
		@param callable The procedure to call
		@param cancelable An object that know how to cancel the running query if this 
			thread is interrupted
		@param sc A session control object that should be kept alive while the
			query is executing, or null
		@return The result from the call
		@throws InterruptedException If this thread was interrupted during the query
	*/
	private static <T> T getFutureResult(Callable<T> callable, Cancelable cancelable, 
		SessionControl sc)
		throws InterruptedException, SQLException
	{
		Future<T> future = service.submit(callable);
		T result = null;
		try
		{
			// We wait about half the time of the timeout before updating the session control
			// sessionCacheTimeout() is in  minutes, but our timeout is in seconds
			int timeout = Application.sessionCacheTimeout() * 30;
			if (timeout <= 0) timeout = 30;
			boolean done = false;
			do
			{
				try
				{
					result = future.get(timeout, TimeUnit.SECONDS);
					done = true;
				}
				catch (TimeoutException ex)
				{
					if (sc != null) sc.updateLastAccess();
				}
			} while (!done);
		}
		catch (InterruptedException ex)
		{
			cancelable.cancel();
			throw ex;
		}
		catch (ExecutionException ex)
		{
			if (ex.getCause() instanceof SQLException)
			{
				throw (SQLException)ex.getCause();
			}
			else
			{
				throw new BaseException(ex.getCause());
			}
		}
		return result;
	}
	
	/**
		Interface to be implemented by classes that knows how to cancel
		a query.
		@since 2.7
	*/
	static interface Cancelable
	{
		public abstract void cancel();
	}
	
	/**
		A callable implementation the calls {@link PreparedStatement#executeQuery()}
	*/
	static class FutureResultSet
		implements Callable<ResultSet>
	{
		private PreparedStatement ps;
		FutureResultSet(PreparedStatement ps)
		{
			this.ps = ps;
		}
		
		@Override
		public ResultSet call()
			throws SQLException
		{
			return ps.executeQuery();
		}
	}
	
	/**
		Cancels a statement by calling {@link Statement#cancel()}.
		@since 2.7
	*/
	static class CancelableStatement
		implements Cancelable
	{
		private final Statement s;
		CancelableStatement(Statement s)
		{
			this.s = s;
		}
		
		@Override
		public void cancel()
		{
			try
			{
				if (!s.isClosed()) s.cancel();
			}
			catch (SQLException ex)
			{
				throw new RuntimeException(ex);
			}
		}
	}
	
	/**
		A callable implementation the calls {@link PreparedStatement#executeUpdate()}
	*/
	static class FutureUpdate
		implements Callable<Integer>
	{
		private PreparedStatement ps;
		FutureUpdate(PreparedStatement ps)
		{
			this.ps = ps;
		}
		
		@Override
		public Integer call()
			throws SQLException
		{
			return ps.executeUpdate();
		}
	}
	
	/**
		Calls {@link Query#scroll(ScrollMode)} and returns the result as a
		{@link ScrollIterator}.
		@since 2.7
	*/
	static class FutureScrollIterator<T>
		implements Callable<ScrollIterator<T>>
	{
		private final Query query;
		private final Class<T> clazz;
		FutureScrollIterator(Class<T> clazz, Query query)
		{
			this.clazz = clazz;
			this.query = query;
		}
		
		@Override
		public ScrollIterator<T> call()
		{
			ScrollableResults result = query.scroll(ScrollMode.FORWARD_ONLY);
			return new ScrollIterator<T>(clazz, result); 
		}
	}
	
	static class FutureItem<T>
		implements Callable<T>
	{
		private final Query query;
		private final Class<T> clazz;
		FutureItem(Class<T> clazz, Query query)
		{
			this.clazz = clazz;
			this.query = query;
		}
		
		@Override
		public T call()
		{
			T item = null;
			ScrollableResults result = query.scroll(ScrollMode.FORWARD_ONLY);
			try
			{
				if (result.next())
				{
					item = clazz.cast(result.get(0));
				}
			}
			finally
			{
				if (result != null) result.close();
			}
			return item;
		}
	}
	
	/**
		Cancels the query running in a Hibernate session.
		@since 2.7
	*/
	static class CancelableSession
		implements Cancelable
	{
		private final org.hibernate.Session session;
		CancelableSession(org.hibernate.Session session)
		{
			this.session = session;
		}
		
		@Override
		public void cancel()
		{
			if (session.isOpen()) session.cancelQuery();
		}
	}

	
	/**
		Cancels a query running in a stateless Hibernate session. NOTE! This 
		functionality depends on a cast to {@link StatelessSessionImpl}. It 
		may not be future-proof since it is not part of the public Hibernate 
		API.
		@since 2.7
	*/
	static class CancelableStatelessSession
		implements Cancelable
	{
		private final StatelessSession session;
		CancelableStatelessSession(StatelessSession session)
		{
			this.session = session;
		}
		
		@Override
		public void cancel()
		{
			if (session instanceof StatelessSessionImpl)
			{
				((StatelessSessionImpl)session).getBatcher().cancelLastQuery();
			}
			else if (session instanceof StatelessSessionWrapper)
			{
				((StatelessSessionWrapper)session).cancelQuery();
			}
		}
	}
	
}
