/*
	$Id: RawDataPropertyExpression.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.util.EqualsHelper;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

/**
	An expression representing the property of a
	raw data spot.
	<p>
	Throws an <code>UnsupportedOperationException</code> if 
	{@link Query#getQueryType()} isn't {@link QueryType#SQL}.

	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.Dynamic#rawData(String)
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class RawDataPropertyExpression
	implements Expression
{
	private static Map<String, String> propertyToColumnCache = new HashMap<String, String>();

	private final String property;
	
	public RawDataPropertyExpression(String property)
		throws InvalidDataException
	{
		this.property = property;
	}

	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	@Override
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			RawDataType rdt = null;
			if (query instanceof DynamicQuery)
			{
				rdt = ((DynamicQuery)query).getDataCube().getRawDataType();
			}
			else if (query instanceof DynamicRawDataQuery)
			{
				rdt = ((DynamicRawDataQuery)query).getRawDataType();
			}
			if (rdt == null) throw new InvalidDataException("Could not find raw data type for query: " +query);
			String column = getColumn(rdt, property);
			if (column == null) throw new InvalidDataException("Unknown property '"+property+"' of raw data type '"+rdt.getName()+"'");
			return HibernateUtil.quote(rdt.getRealTable().getAlias()) + "." + HibernateUtil.quote(column); 
		}
		else
		{
			throw new UnsupportedOperationException("'rawDataProperty' operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	@Override
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return "raw." + property;
	}
	/**
		This expression is equal to another RawDataPropertyExpression if they 
		represent the same raw data property.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		RawDataPropertyExpression o = (RawDataPropertyExpression)other;
		return EqualsHelper.equals(property, o.property);
	}
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(property);
	}
	// -------------------------------------------
	
	/**
		Get the database column name for the specified property of the 
		specified raw data type.
		@return The column name, or null if the property isn't found
	*/
	private static String getColumn(RawDataType rdt, String property)
	{
		String column = null;
		String cacheName = rdt.getId() + "." + property;
		
		column = propertyToColumnCache.get(cacheName);
		
		// Check if the property is found in the cache
		if (column == null)
		{
			synchronized (RawDataPropertyExpression.class)
			{
				// Find a property with the specified name from the Hibernate mappings
				org.hibernate.mapping.PersistentClass pc = HibernateUtil.getClassMapping(rdt.getEntityName());
				org.hibernate.mapping.Property prop = null;
				if ("id".equals(property)) 
				{
					// ID is a special property not returned by the getPropertyIterator().
					prop = pc.getIdentifierProperty();
				}
				else
				{
					Iterator propIt = pc.getPropertyIterator();
					while (prop == null && propIt.hasNext())
					{
						org.hibernate.mapping.Property temp = (org.hibernate.mapping.Property)propIt.next();
						if (property.equals(temp.getName()))
						{
							prop = temp;
						}
					}
				}
				if (prop != null)
				{
					// Found a propety, get the name of the column.
					// Note! We expect that each property maps to one column only!
					Iterator colIt = prop.getColumnIterator();
					if (colIt.hasNext())
					{
						org.hibernate.mapping.Column col = (org.hibernate.mapping.Column)colIt.next();
						column = col.getName();
					}
				}
				propertyToColumnCache.put(cacheName, column);
			}
		}
		return column;
	}
}
