/*
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collections;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.data.ItemKeyData;
import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.ExtensionsFilter;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;

/**
	Utility class for the "services" extension point. While part of the web client code
	we need to place it in the core to get access to special SessionControl item
	that can be used by the service extensions to access BASE.

	@author nicklas
	@since 3.3
*/
public final class ServicesUtil 
{

	/**
		Create an extensions invoker for the services extensions. The client
		context is created with special SessionControl that gives the extension
		permission to impersonate another user.
	*/
	public static ExtensionsInvoker<?> useExtensions(ExtensionsFilter filter)
	{
		Registry registry = Application.getExtensionsManager().getRegistry();
		SessionControl sc = new SessionControl(null, "net.sf.basedb.clients.internal", false, Application.getHostName());
		sc.login(new ServicesKeyring());
		ClientContext context = new ClientContext(sc);
		ExtensionsInvoker<?> invoker = registry.useExtensions(context, filter, "net.sf.basedb.clients.web.services");
		return invoker;
	}
	
	
	/**
		Our own keyring containing the only permission: {@link Permission#ACT_AS_ANOTHER_USER}.
		All other permissions are denied.
	*/
	static class ServicesKeyring
		extends Keyring
	{
	
		private final int systemPermissions;
		private final Set<Integer> emptySet;
		private final Map<Integer, Integer> emptyMap;
	
		private ServicesKeyring()
		{
			super(0);
			this.systemPermissions = Permission.grant(Permission.ACT_AS_ANOTHER_USER);
			this.emptySet = Collections.emptySet();
			this.emptyMap = Collections.emptyMap();
		}
		/*
			From the Keyring class
			-------------------------------------------
		*/
		/**
			Never reload since everything is static.
		*/
		@Override
		boolean getReload()
		{
			return false;
		}
		/**
			Return {@link Permission#ACT_AS_ANOTHER_USER} if <code>itemType</code> = 
			{@link Item#SYSTEM}, 0 otherwise. 
		*/
		@Override
		int getRolePermissions(Item itemType)
		{
			if (itemType == Item.SYSTEM)
			{
				return systemPermissions;
			}
			else
			{
				return 0;
			}
		}
		/**
			Return 0.
		*/
		@Override
		int getSharedPermissions(ShareableData sharableData)
		{
			return 0;
		}
		/**
			Return 0.
		*/
		@Override
		int getAllPermissions(Item itemType, UserData owner, ItemKeyData itemKey, ProjectKeyData projectKey)
		{
			return 0;
		}
		/**
			Return empty set.
		*/
		@Override
		Set<Integer> getRoles()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getGroups()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		@Override
		Set<Integer> getUsers()
		{
			return emptySet;
		}
		/**
			Return empty map.
		*/
		@Override
		Map<Integer, Integer> getProjects()
		{
			return emptyMap;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getItemKeys()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getProjectKeys()
		{
			return emptySet;
		}
		// -------------------------------------------
	}
}
