/*
	$Id: SessionControl.java 6448 2014-04-16 11:29:49Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.PermissionTemplateData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.SessionData;
import net.sf.basedb.core.data.ClientData;
import net.sf.basedb.core.data.QuotaData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.data.ItemKeyData;
import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.data.SettingData;
import net.sf.basedb.core.data.UserClientSettingData;
import net.sf.basedb.core.data.UserDefaultSettingData;
import net.sf.basedb.core.data.ClientDefaultSettingData;
import net.sf.basedb.core.data.ContextData;
import net.sf.basedb.core.data.ContextIndex;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.core.authentication.AuthenticatedUser;
import net.sf.basedb.core.authentication.AuthenticationManager;
import net.sf.basedb.core.authentication.AuthenticationMethod;
import net.sf.basedb.core.authentication.ExtraUserInformation;
import net.sf.basedb.core.authentication.LoginRequest;

import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.Locale;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.util.WeakHashMap;
import java.util.Collections;
import java.util.List;


/**
	This class handles user sessions. Each instance is thread-safe and
	can handle a single user at a time. This class is used
	to create {@link DbControl} objects which are needed to access the
	database.

	@author Nicklas
	@version 2.0
	@see Application#newSessionControl(String, String, String)
	@see Application#getSessionControl(String, String)
	@base.modified $Date: 2014-04-16 13:29:49 +0200 (on, 16 apr 2014) $
*/
@SuppressWarnings("deprecation")
public class SessionControl
{

	/**
		Log core events.
	*/
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(SessionControl.class);

	/**
		The id of the object. Used by {@link Application#getSessionControl(String,String)}
	*/
	private final String id;

	/**
		The id of the {@link ClientData}. 0 = no client
	*/
	private final int clientId;
	
	/**
		The external id of the {@link ClientData}. null = no client
	*/
	private final String externalClientId;

	/**
		The remote IP.
	*/
	private final String remoteId;

	/**
		Holds all created <code>DbControl</code> object, so we can close
		them if the user logs out. We use only the keys of the map since there 
		is no <code>WeakHashSet</code> class.
	*/
	private final Map<DbControl,String> dbControlCache;

	/**
		The time of the last access to this object.
	*/
	private long lastAccess;
	
	/**
		If the session control has been closed it can no longer be used.
	*/
	private boolean closed;
	
	/**
		Stores name and value from {@link ClientDefaultSettingData}.
	*/
	private Map<String, SettingInfo> clientDefaultSettings;

	/**
		Login information for the logged in user.
	*/
	private LoginInfo loginInfo;
	
	/**
		Map for storing current contexts.
	*/
	private final Map<ContextKey, ItemContext> currentContexts;
	
	private int pluginId;
	
	private int jobId;
	
	/**
		Create a new session control object using another one as the template.
		Permissions and settings are copied/referenced from the parent session
		control. If the plugin uses permissions those permissions are loaded and
		used for access control.

		@param id The ID if the new session control
		@param parent The parent session control
		@param plugin The plugin that is executing
	*/
	SessionControl(String id, SessionControl parent, PluginDefinition plugin, Job job)
	{
		this.id = id;
		this.clientId = parent.clientId;
		this.externalClientId = parent.externalClientId;
		this.remoteId = parent.remoteId;
		this.dbControlCache = Collections.synchronizedMap(new WeakHashMap<DbControl,String>());
		this.currentContexts = parent.currentContexts;
		if (plugin != null) this.pluginId = plugin.getId();
		if (job != null) this.jobId = job.getId();
		LoginInfo li = new LoginInfo(parent.loginInfo);
		li.keyring = new Keyring(li.keyring, pluginId, plugin != null && plugin.getUsePermissions());
		this.loginInfo = li;
	}
	
	/**
		Creates a new <code>SessionControl</code> object.
		@param id The id used in {@link Application#newSessionControl(String,String,String)}
		@param externalClientId Optional external ID of the client application in use
		@param failIfNoClient If TRUE, an exception is thrown if an external client id
			has been specified, but not found
		@param remoteId For example, the IP-address
	*/
	SessionControl(String id, String externalClientId, boolean failIfNoClient, String remoteId)
		throws BaseException
	{
		this.id = id;
		this.externalClientId = externalClientId;
		this.remoteId = remoteId;
		this.dbControlCache = Collections.synchronizedMap(new WeakHashMap<DbControl,String>());
		this.currentContexts = Collections.synchronizedMap(new HashMap<ContextKey, ItemContext>());
		int tempClientId = 0;
		if (externalClientId != null)
		{
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				
				// Find the internal client id for the given external id
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_CLIENT_FOR_EXTERNAL_ID");
				/*
					SELECT cli
					FROM ClientData cli
					WHERE cli.externalId = :externalId
				*/
				query.setString("externalId", externalClientId);
				ClientData clientData = HibernateUtil.loadData(ClientData.class, query);
				if (clientData != null)
				{
					tempClientId = clientData.getId();
					// load default settings for client application
					clientDefaultSettings = loadClientDefaultSettings(session, tempClientId);
				}
				else if (failIfNoClient)
				{
					throw new ItemNotFoundException("Client[externalId="+externalClientId+"]");
				}
			}
			finally
			{
				if (tx != null) HibernateUtil.commit(tx);
				if (session != null) HibernateUtil.close(session);
			}
		}
		this.clientId = tempClientId;
		updateLastAccess();
	}

	/**
		Get the time this object last was accessed. Used by the cleanup timer
		in {@link Application#cleanSessionControlCache(boolean)}
	 	@return A long object.  
		@since 2.4.5
	*/
	public long getLastAccess()
	{
		return lastAccess;
	}

	/**
		Update the time this object last was accessed. Used by the cleanup timer
		in {@link Application#cleanSessionControlCache(boolean)}
	*/
	public void updateLastAccess()
	{
		lastAccess = System.currentTimeMillis();
	}
	
	/**
		If this session control has been closed or not. A closed session
		control can't be used.
		@return TRUE if the session control is closed
		@since 2.4.5
	*/
	public boolean isClosed()
	{
		return closed;
	}
	
	/**
		Close this session control so it can't be used again.
		@since 2.4.5
	*/
	public void close()
	{
		if (isLoggedIn()) logout();
		closed = true;
	}
	
	
	/**
		Get the id of this <code>SessionControl</code>. Use this value for
		retrieving the object from the {@link Application} cache.
		@return The id
		@see Application#getSessionControl(String,String)
	*/
	public String getId()
	{
		return id;
	}
	
	/**
		Get the id of the <code>Client</code> application in use. Use
		{@link Client#getById(DbControl, int)} to get the {@link Client} object.
	 	@return Client id as an int. 
	*/
	public int getClientId()
	{
		updateLastAccess();
		return clientId;
	}

	/**
		Get the external id of the <code>Client</code> application in use.
	 	@return A java.lang.String object.
		@see #getClientId()
	*/
	public String getExternalClientId()
	{
		updateLastAccess();
		return externalClientId;
	}

	
	/**
		Get the remote id that was passed to the {@link 
		Application#newSessionControl(String,String,String)} method when creating this
		object. This value is checked against the value passed to the {@link
		Application#getSessionControl(String,String)} method.
	*/
	String getRemoteId()
	{
		return remoteId;
	}
	
	/**
		Create a new <code>DbControl</code> object for this session.
		@return A <code>DbControl</code> object
		@throws BaseException If there is an error
	*/
	public DbControl newDbControl()
		throws BaseException
	{
		updateLastAccess();
		DbControl dbControl = new DbControl(this);
		dbControlCache.put(dbControl, null);
		return dbControl;
	}
	
	void login(Keyring keyring)
	{
		updateLastAccess();
		LoginInfo li = new LoginInfo();
		li.keyring = keyring;
		li.sessionId = 0;
		li.userId = 0;
		li.userLogin = null;
		loginInfo = li;
	}
	
	/**
		@deprecated In 3.3, use {@link #login(LoginRequest)} instead
	*/
	@Deprecated
	public synchronized void login(String login, String password, String comment)
	{
		LoginRequest loginRequest = new LoginRequest();
		loginRequest.setLogin(login);
		loginRequest.setPassword(password);
		loginRequest.setComment(comment);
		login(loginRequest);
	}

	/**
		Log in to BASE. The method checks that the given login is valid, 
		the password is correct and that the user has USE permission for
		the client application if one has been specified.

		@param loginRequest Request containing login/password and other information
		@throws ItemNotFoundException If a user with the specified username is not found
		@throws InvalidPasswordException If the specified password is incorrect
		@throws PermissionDeniedException If the user doesn't have 
			{@link Permission#USE} permission for the current client 
			application or if a user is already logged in
		@throws BaseException If there is any other error
		@see #logout()
		@see #isLoggedIn()
		@see #getLoggedInUserId()
		@since 3.3 
	*/
	public synchronized void login(LoginRequest loginRequest)
	{
		updateLastAccess();
		// Are we already logged in?
		if (isLoggedIn())
		{
			throw new AlreadyLoggedInException(loginInfo.userLogin);
		}
		if (loginRequest == null) throw new InvalidUseOfNullException("loginRequest");
		if (loginRequest.getLogin() == null) throw new InvalidUseOfNullException("loginRequest.login");

		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);

			// Try external authentication at first, except...
			// ...root user should always use internal verification
			UserData root = HibernateUtil.loadData(session, UserData.class, SystemItems.getId(User.ROOT));
			assert root != null : "root == null";
			
			AuthenticatedUser authUser = null;
			if (!loginRequest.getLogin().equals(root.getLogin()))
			{
				authUser = verifyUserExternal(session, loginRequest);
			}
			
			if (authUser == null)
			{
				// If no user was found, use internal authentication
				authUser = verifyUserInternal(session, loginRequest);
			}
			
			UserData user = HibernateUtil.loadData(session, UserData.class, authUser.getInternalId());
			
			LoginInfo li = createLoginInfo(session, user, loginRequest.getComment(), false, authUser.getAuthenticationMethod());
			HibernateUtil.commit(tx);
			currentContexts.clear();
			loginInfo = li;
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
	}
	

	/**
		Verify the user with internal authentication. If the LoginRequest
		has a user id this is used first, otherwise the login is used. 
	*/
	AuthenticatedUser verifyUserInternal(org.hibernate.Session session, LoginRequest loginRequest)
	{
		String login = loginRequest.getLogin();
		String password = loginRequest.getPassword();
		int userId = loginRequest.getUserId();
		
		UserData userData = null;
		if (userId > 0)
		{
			userData = HibernateUtil.loadData(session, UserData.class, userId);
			if (userData == null)
			{
				throw new ItemNotFoundException("The user with id '" + userId + "' is not known to BASE.");
			}
			login = userData.getLogin();
		}
		else if (login != null)
		{
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_USER_FOR_LOGIN");
			/*
				SELECT usr
				FROM UserData usr
				WHERE usr.login = :login
			*/
			query.setString("login", login);
			userData = HibernateUtil.loadData(UserData.class, query);
			if (userData == null)
			{
				throw new ItemNotFoundException("The user with login '" + login + "' is not known to BASE.");
			}
		}
		if (userData == null)
		{
			throw new ItemNotFoundException("User");
		}
		
		if (userData.isRemoved())
		{
			throw new PermissionDeniedException("The user with login '" + login + "' is flagged for deletion.");	
		}
		if (userData.isDisabled())
		{
			throw new PermissionDeniedException("The user with login '" + login + "' is disabled.");	
		}
		Date expirationDate = userData.getExpirationDate();
		if ((expirationDate != null) && expirationDate.before(new Date()))
		{
			throw new AccountExpiredException(login, expirationDate);
		}
		
		// Check the password
		String cryptedPassword = userData.getPassword().getCryptedPassword();
		if (cryptedPassword == null || !User.checkPassword(password, cryptedPassword))
		{
			throw new InvalidPasswordException("The password for login '" + login + "' is not correct.");
		}
		
		return new AuthenticatedUser(AuthenticationMethod.INTERNAL, userData);
	}
	
	/**
		Verify the user with external authentication.
	*/
	@SuppressWarnings("unchecked")
	private AuthenticatedUser verifyUserExternal(org.hibernate.Session session, LoginRequest loginRequest)
	{
		AuthenticationContext context = new AuthenticationContext(this, session, loginRequest);
		
		ExtensionsManager xtManager = Application.getExtensionsManager();
		Registry registry = xtManager.getRegistry();
		ExtensionsInvoker<AuthenticationManager> invoker = (ExtensionsInvoker<AuthenticationManager>)registry.useExtensions(context, xtManager.getSettings(), "net.sf.basedb.core.authentication-manager");
		
		AuthenticatedUser authUser = null;
		try
		{
			// Load all installed authentication managers and iterate until 
			// someone returns an info object or throws an exception
			for (AuthenticationManager auth : invoker)
			{
				authUser = auth.authenticate();
				if (authUser != null)
				{
					// Found a valid login
					break;
				}
			}
		}
		catch (net.sf.basedb.core.authentication.LoginException ex)
		{
			throw new LoginException(ex.getMessage(), ex);
		}
		catch (net.sf.basedb.core.authentication.UnknownLoginException ex)
		{
			throw new ItemNotFoundException("The user with login '" + loginRequest.getLogin() + "' is not known to BASE.", ex);
		}
		catch (net.sf.basedb.core.authentication.InvalidPasswordException ex)
		{
			throw new InvalidPasswordException("The password for login '" + loginRequest.getLogin() + "' is not correct.", ex);
		}
		catch (net.sf.basedb.core.authentication.AuthenticationException ex)
		{
			if (!Config.getBoolean("auth.cachepasswords"))
			{
				throw new BaseException(ex);
			}
		}
		
		// Return null to force internal verification
		if (authUser == null) return null;

		
		// Load user information, priority order is 'internal id', 'external id', 'login'
		UserData userData = null;
		if (authUser.getInternalId() > 0)
		{
			userData = HibernateUtil.loadData(session, UserData.class, authUser.getInternalId());
		}
		if (userData == null && authUser.getExternalId() != null)
		{
			userData = context.getUserByExternalId(authUser.getExternalId());
		}
		if (userData == null && authUser.getLogin() != null)
		{
			userData = context.getUserByLogin(authUser.getLogin());
		}
			
		if (userData == null) 
		{
			// Create new user
			userData = new UserData();
			userData.setExternalId(authUser.getExternalId());
			userData.setLogin(authUser.getLogin());
			userData.setName(authUser.getLogin());
			userData.setQuota(HibernateUtil.loadData(session, QuotaData.class, SystemItems.getId(Quota.DEFAULT)));
			User.addDefultRolesAndGroups(session, userData);
		}
		
		// Cache the password if allowed by config
		if (Config.getBoolean("auth.cachepasswords"))
		{
			userData.getPassword().setCryptedPassword(User.encryptPassword(loginRequest.getPassword()));
			int daysToCache = Config.getInt("auth.daystocache", 0);
			userData.setExpirationDate(daysToCache > 0 ? new Date(System.currentTimeMillis()+daysToCache*24L*3600L*1000L) : null);
		}
		
		// Synchronize extra information if supported and allowed by config
		if (authUser.hasExtraInformation() && (Config.getBoolean("auth.synchronize") || userData.getId() == 0))
		{
			ExtraUserInformation extraInfo = authUser.getExtraInformation();
			if (extraInfo.getName() != null)
			{
				userData.setName(extraInfo.getName());
			}
			userData.setOrganisation(extraInfo.getOrganisation());
			userData.setAddress(extraInfo.getAddress());
			userData.setPhone(extraInfo.getPhone());
			userData.setFax(extraInfo.getFax());
			userData.setEmail(extraInfo.getEmail());
			userData.setUrl(extraInfo.getUrl());
			userData.setDescription(extraInfo.getDescription());
			
			List<ExtendedProperty> properties = ExtendedProperties.getProperties("UserData");
			if (properties != null)
			{
				for (ExtendedProperty ep : properties)
				{
					Object value = extraInfo.getExtended(ep.getName());
					if (value == null || ep.getType().isCorrectType(value))
					{
						userData.setExtended(ep.getName(), value);
					}
				}
			}
		}
		if (userData.getId() == 0)
		{
			HibernateUtil.saveData(session, userData);	
		}
		HibernateUtil.flush(session);
		return new AuthenticatedUser(authUser.getAuthenticationMethod(), userData);
	}

	/**
		Log in as another user or create a clone of the currently logged in user's session.
		If this call is successful, you will get a new <code>SessionControl</code> object which 
		is equivalent to a <code>SessionControl</code> where that user logged in by normal means. 
		This method requires that the logged in user has {@link Permission#ACT_AS_ANOTHER_USER} 
		permission or that the userId is the same as the currently logged in user.

		@param userId The id of the user to login as
		@param comment A comment that will be placed in the {@link Session} information
		@return A new <code>SessionControl</code> object
		@throws ItemNotFoundException If no user with the specified id exists
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#ACT_AS_ANOTHER_USER} permission and the userId
			is a different user from the currently logged in user
		@throws BaseException If there is another error
	*/
	public SessionControl impersonateLogin(int userId, String comment)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (!hasSystemPermission(Permission.ACT_AS_ANOTHER_USER) && userId != getLoggedInUserId())
		{
			throw new PermissionDeniedException("No permission to act as another user.");
		}

		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			// Load user data
			UserData userData = HibernateUtil.loadData(session, UserData.class, userId);
			LoginInfo li = createLoginInfo(session, userData, comment, true, getAuthenticationMethod());
			HibernateUtil.commit(tx);
			SessionControl impersonated = Application.newSessionControl(getExternalClientId(), getRemoteId(), null);
			impersonated.loginInfo = li;
			return impersonated;
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Log in as the owner of the specified item.
	 	@param item Login as the owner of this Item. Null is not allowed 
	 	@param comment A comment that will be placed in the {@link Session} information
	 	@return A {@link net.sf.basedb.core.SessionControl} for the created session.
		@see #impersonateLogin(int, String)
	*/
	public SessionControl impersonateLogin(Ownable item, String comment)
	{
		UserData owner = ((OwnableData)((BasicItem)item).getData()).getOwner();
		return impersonateLogin(owner.getId(), comment);
	}
	
	/**
		Create a LoginInfo object and load all information that it needs.
	*/
	private LoginInfo createLoginInfo(org.hibernate.Session session, UserData userData, String comment, boolean impersonated, AuthenticationMethod authenticationMethod)
		throws BaseException
	{
		// Load the Keyring
		LoginInfo li = new LoginInfo();
		li.impersonated = impersonated;
		li.keyring = new Keyring(userData.getId());
		
		// If a client has been specified - check for USE permission
		ClientData clientData = null;
		if (getClientId() != 0) 
		{
			clientData = HibernateUtil.loadData(session, ClientData.class, getClientId());
			if (!impersonated)
			{
				int permissions = li.keyring.getAllPermissions(
					Item.CLIENT, clientData.getOwner(), clientData.getItemKey(), clientData.getProjectKey()
					);
				if (!Permission.hasPermission(permissions, Permission.USE))
				{
					throw new PermissionDeniedException(Permission.USE, "Client[externalId="+clientData.getExternalId()+"]");
				}
			}
		}

		// Load settings
		if (getClientId() != 0)
		{
			li.userClientSettings = loadUserClientSettings(session, userData.getId(), clientId);
		}
		li.userDefaultSettings = loadUserDefaultSettings(session, userData.getId());

		// Create a new session information entry
		SessionData sessionData = new SessionData();
		sessionData.setUser(userData);
		sessionData.setClient(clientData);
		sessionData.setLoginTime(new Date());
		sessionData.setLoginComment(comment);
		sessionData.setRemoteId(getRemoteId());
		sessionData.setImpersonated(impersonated);
		HibernateUtil.saveData(session, sessionData);

		li.sessionId = sessionData.getId();
		li.userId = userData.getId();
		li.userLogin = userData.getLogin();
		li.authenticationMethod = authenticationMethod;
		li.sessionSettings = Collections.synchronizedMap(new HashMap<String,Object>());
		return li;
	}
	
	/**
		Log out of BASE. If the logged in user was impersonated the system will 
		revert to the original user. If no user is logged in this method does
		nothing. If there are any {@link DbControl} objects still open, they
		will be closed. Changes made to items managed by those <code>DbControl</code>
		objects will be lost unless the items are connected to a new <code>DbControl</code> 
		object.

		@throws BaseException If there is an error
		@see #login(String, String, String)
		@see #isLoggedIn()
		@see #impersonateLogin(int, String)
		@see #isImpersonated()
	*/
	public synchronized void logout()
		throws BaseException
	{
		updateLastAccess();
		if (!isLoggedIn()) return;
		
		clearDbControlCache();

		if (loginInfo.userId != 0)
		{
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				// Load and save settings
				saveAllUserSettings(session);
	
				// Load session information and set logout time
				SessionData sessionData = HibernateUtil.loadData(session, SessionData.class, loginInfo.sessionId);
				sessionData.setLogoutTime(new Date());
				HibernateUtil.commit(tx);
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}
		currentContexts.clear();
		loginInfo = null;
	}
	
	/**
		Checks if a user is logged in or not.
		@return TRUE if a user is logged in, FALSE otherwise.
		@see #login(String, String, String)
		@see #logout()
		@see #getLoggedInUserId()
	*/
	public boolean isLoggedIn()
	{
		return loginInfo != null;
	}

	/**
		Get the id of the logged in user, or 0 if no user is logged in.
		Use {@link User#getById(DbControl, int)} to get the {@link User} object.
	 	@return User id as an int. 
		@see #login(String, String, String)
		@see #logout()
		@see #isLoggedIn()
	*/
	public int getLoggedInUserId()
	{
		updateLastAccess();
		return loginInfo == null ? 0 : loginInfo.userId;
	}
	
	/**
		Get the method that was used to authenticate the currently
		logged in user. Internal authentication return
		{@link AuthenticationMethod#INTERNAL}.
		@return An AuthenticationMethod object or null if no user is logged in
		@since 3.3
	*/
	public AuthenticationMethod getAuthenticationMethod()
	{
		return loginInfo == null ? null : loginInfo.authenticationMethod;
	}

	/**
		Get the id of the current user session, or 0 if no user is logged in.
		Use {@link Session#getById(DbControl, int)} to the {@link Session}
		object.
		@return The session id, or 0 if no user is logged in
		@since 2.13
	*/
	public int getCurrentSessionId()
	{
		updateLastAccess();
		return loginInfo == null ? 0 : loginInfo.sessionId;
	}
	
	/**
		Get the id of the plug-in that is currently running with this
		session control.
		@return The plug-in id, or 0 if not running by a plug-in
		@since 2.13
	*/
	public int getPluginId()
	{
		updateLastAccess();
		return pluginId;
	}
	
	/**
		Get the id of the job that is currently running with this
		session control.
		@return The job id, or 0 if not running by a job
		@since 2.13
	*/
	public int getJobId()
	{
		updateLastAccess();
		return jobId;
	}
	
	/**
		Check if the logged in user was impersonated by another user.
		@return TRUE or FALSE
		@see #impersonateLogin(int, String)
	*/
	public boolean isImpersonated()
	{
		updateLastAccess();
		return loginInfo != null && loginInfo.impersonated;
	}

	/**
		Tell the core to reload the logged in users permission the next time
		the client application needs to check the permissions.
	*/
	public void reloadPermissions()
	{
		updateLastAccess();
		if (loginInfo != null) loginInfo.keyring.setReload(true);
	}
	
	/**
		Reloads user- and client-specific settings. 
		
		@param onlyClientDefaultSettings Setting this to TRUE will only reload
			{@link ClientDefaultSetting}:s
		@param saveCurrentUserSettings Setting this to TRUE will save the current user-specific 
			settings before reloading them. This parameter is ignored if <code>onlyClientDefaultSettings</code>
			is true. The settings are never changed if the logged in user is a multi-user account.
		@throws BaseException If the settings cannot be reloaded
	*/
	public void reloadSettings(boolean onlyClientDefaultSettings, boolean saveCurrentUserSettings)
		throws BaseException
	{
		updateLastAccess();
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			if (getClientId() != 0)
			{
				clientDefaultSettings = loadClientDefaultSettings(session, getClientId());
			}
			if (!onlyClientDefaultSettings && isLoggedIn())
			{
				if (saveCurrentUserSettings)
				{
					saveAllUserSettings(session);
				}
				if (getClientId() != 0)
				{
					loginInfo.userClientSettings = loadUserClientSettings(session, getLoggedInUserId(), getClientId());
				}
				loginInfo.userDefaultSettings = loadUserDefaultSettings(session, getLoggedInUserId());
			}
		}
		finally
		{
			if (tx != null) HibernateUtil.commit(tx);
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Get the id of the active project, or 0 if no project is active.
		Use {@link Project#getById(DbControl, int)} to get the {@link Project} object.
	 	@return An int. 
		@see #setActiveProject(Project)
	*/
	public int getActiveProjectId()
	{
		updateLastAccess();
		return loginInfo == null ? 0 : loginInfo.activeProjectId;
	}

	/**
		Set the active project. When a project is active queries will by 
		default only load items shared to the that project, and new {@link Shareable}
		items will automatically be shared to the active project with delete permission.
		The active project doesn't affect already existing items.

		@param project The project to make active
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#USE} permission for the project
		@throws BaseException If there is another error
		@see #getActiveProjectId()
	*/
	public synchronized void setActiveProject(Project project)
		throws PermissionDeniedException, BaseException
	{
		updateLastAccess();
		if (loginInfo != null) 
		{
			if (project == null)
			{
				loginInfo.keyring.setActiveProject(null);
				loginInfo.activeProjectId = 0;
				loginInfo.projectKeyId = 0;
				loginInfo.itemKeyId = 0;
			}
			else
			{
				project.checkPermission(Permission.READ);
				loginInfo.keyring.setActiveProject(project.getData());
				loginInfo.activeProjectId = project.getId();
				if (project.hasPermission(Permission.USE))
				{
					PermissionTemplateData template = project.getData().getPermissionTemplate();
					if (template != null)
					{
						org.hibernate.Session session = null;
						org.hibernate.Transaction tx = null;
						try
						{
							if (project.isDetached())
							{
								// Make sure that we can load the template's item and project keys
								session = HibernateUtil.newSession();
								tx = HibernateUtil.newTransaction(session);
								template = HibernateUtil.loadData(session, PermissionTemplateData.class, template.getId());
							}
							ProjectKeyData pk = template.getProjectKey();
							ItemKeyData ik = template.getItemKey();
							loginInfo.projectKeyId = pk == null ? 0 : pk.getId();
							loginInfo.itemKeyId = ik == null ? 0 : ik.getId();
						}
						finally
						{
							if (tx != null) HibernateUtil.commit(tx);
							if (session != null) HibernateUtil.close(session);
						}
					}
					else
					{
						ProjectPermissions pp = new ProjectPermissions();
						pp.setPermissions(project, project.getAutoPermission());
						loginInfo.projectKeyId = ProjectKey.getNewOrExistingId(null, pp.getProjectPermissions());
						loginInfo.itemKeyId = 0;
					}
				}
				else
				{
					loginInfo.projectKeyId = 0;
					loginInfo.itemKeyId = 0;
				}
			}
		}
	}
	
	/**
		Get the permissions that the logged in user has in the currently active project.
		@return A Set with the permissions
		@since 3.1
	*/
	public Set<Permission> getActiveProjectPermissions()
	{
		return Permission.fromInt(loginInfo == null ? 0 : loginInfo.keyring.getMaxProjectPermission());
	}
	

	
	/**
		Create a new session control for executing a plugin.
		@param plugin The plugin to execute
		@param job The job that executes the plug-in, or null
		@return A plugin session control
		@since 2.15 (was package private before that)
	*/
	public PluginSessionControl getPluginSessionControl(PluginDefinition plugin, Job job)
	{
		return Application.newPluginSessionControl(this, plugin, job);
	}
	
	/**
		Create a new session control for executing a job. The new session
		control will inherit all settings from this session control.
		@since 3.3
	*/
	public PluginSessionControl getJobSessionControl(Job job)
	{
		return Application.newPluginSessionControl(this, null, job);
	}
	
	/**
		Get the id of the {@link ProjectKey} that can be used to share an item to
		the active project with full permission, or 0 if no project is active or
		the logged in user only has read permission to the active project.
		Use {@link ProjectKey#getById(DbControl, int)} to get the {@link ProjectKey} object.
		@see #getActiveProjectId()
		@see #setActiveProject(Project)
	*/
	public int getProjectKeyId()
	{
		return loginInfo == null ? 0 : loginInfo.projectKeyId;
	}
	
	/**
		Get the id of the {@link ItemKey} that new items should be shared to
		according to the settings in the currently active project. If no
		project is active or if the projects doesn't share to any item
		key, 0 is returned.
		Use {@link ItemKey#getById(DbControl, int)} to get the {@link ItemKey} object.
		@see #getActiveProjectId()
		@see #setActiveProject(Project)
		@see #getProjectKeyId()
		@since 2.16
	*/
	public int getItemKeyId()
	{
		return loginInfo == null ? 0 : loginInfo.itemKeyId;
	}
	
	/**
		Check if the logged in user has the specified system permission.
		@param permission One of the system permission constants defined
			by the {@link Permission} enumeration
		@return TRUE if the logged in user has the permission, FALSE otherwise
	*/
	public boolean hasSystemPermission(Permission permission)
	{
		return Permission.hasPermission(getRolePermissions(Item.SYSTEM), permission);
	}

	/**
		Check if the logged in user has the specfied permission for
		all item of the specified type. This method checks the permissions
		given to the roles the user is a member of and, if a plugin is active,
		the permissions given to the plugin.
		
		@param permission One of the permissions constants defined
			by the {@link Permission} enumeration
		@param itemType One of the item constants defined by the
			{@link Item} enumeration
		@return TRUE if the logged in user has the permission, FALSE otherwise
	*/
	public boolean hasPermission(Permission permission, Item itemType)
	{
		int permissions = getRolePermissions(itemType);
		if (Permission.hasPermission(permissions, Permission.DENIED))
		{
			return false;
		}
		else
		{
			PluginPermission pp = getPluginPermissions(itemType);
			permissions = (permissions & ~pp.denied) | pp.granted;
		}
		return Permission.hasPermission(permissions, permission);
	}

	/**
		Check if the logged in user has been denied the specified permission
		for items of the specified type. This method checks the permissions
		given to the roles the user is a member of and, if a plugin is active,
		the permissions given to the plugin. Note that this method is not the
		inverse of the {@link #hasPermission(Permission, Item)} method. Both 
		methods may return false in which case the system also checks ownership
		and sharing information to determine if access is granted. If this method
		return TRUE the <code>hasPermission</code> method always return false.
		
		@param permission One of the permissions constants defined
			by the {@link Permission} enumeration
		@param itemType One of the item constants defined by the
			{@link Item} enumeration
		@return TRUE if the logged in user is denied the permission, FALSE otherwise
	*/
	public boolean isDenied(Permission permission, Item itemType)
	{
		int permissions = getRolePermissions(itemType);
		if (Permission.hasPermission(permissions, Permission.DENIED))
		{
			return true;
		}
		else
		{
			PluginPermission pp = getPluginPermissions(itemType);
			return Permission.isDeniedPermission(pp.denied & ~pp.granted, permission);
		}
	}
	
	/**
		Get all role-based permissions for the specified type of items.
	*/
	int getRolePermissions(Item itemType)
	{
		return loginInfo == null ? 0 : loginInfo.keyring.getRolePermissions(itemType);
	}

	/**
		Get all plugin-specific permissions for the specified item type.
	*/
	PluginPermission getPluginPermissions(Item itemType)
	{
		return loginInfo == null ? 
			PluginPermission.NO_PLUGIN : loginInfo.keyring.getPluginPermissions(itemType);
	}
	
	/**
		Get the shared permissions for a {@link ShareableData} object.
	*/
	int getSharedPermissions(ShareableData shareableData)
	{
		return loginInfo == null ? 0 : loginInfo.keyring.getSharedPermissions(shareableData);
	}
	
	/**
		Get the combined permissions for an item of the specified type, owned by the specified
		user and shared with the specified item and project keys.
	*/
	int getAllPermissions(Item itemType, UserData owner, ItemKeyData itemKey, ProjectKeyData projectKey)
	{
		return loginInfo == null ? 0 : loginInfo.keyring.getAllPermissions(itemType, owner, itemKey, projectKey);
	}
	
	/**
		Same as {@link #getAllPermissions(Item, UserData, ItemKeyData, ProjectKeyData)}
	*/
	int getAllPermissions(SharedData sharedData)
	{
		return getAllPermissions(Item.fromDataObject(sharedData), sharedData.getOwner(), sharedData.getItemKey(), sharedData.getProjectKey());
	}
	
	/**
		Check if the logged in user is a member, directly or indirectly, of the specified group.
		@param group The group to check
		@return TRUE if the user is a member, FALSE otherwise
	*/
	public boolean isMemberOf(Group group)
	{
		return loginInfo == null ? false : loginInfo.keyring.getGroups(false).contains(group.getId());
	}
	
	/**
		Check if the logged in user is a member of the specified role.
		@param role The role to check
		@return TRUE if the user is a member, FALSE otherwise
	*/
	public boolean isMemberOf(Role role)
	{
		return loginInfo == null ? false : loginInfo.keyring.getRoles().contains(role.getId());
	}
	
	/**
		Check if the logged in user is a member, directly or indirectly, of the specified project.
		@param project The project to check
		@return TRUE if the user is a member, FALSE otherwise
	*/
	public boolean isMemberOf(Project project)
	{
		return loginInfo == null ? false : loginInfo.keyring.getProjects().keySet().contains(project.getId());
	}
	
	int getProjectPermission(Project project)
	{
		return loginInfo == null ? 0 : IntegerUtil.getInt(loginInfo.keyring.getProjects().get(project.getId()));
	}
	
	/**
		Check if the logged in user is a friend of of the specified user.
		A users is a friend if the logged in user is a member, directly or indirectly, 
		of a group where the other user is direct member.
		@param user The user to check
		@return TRUE if the user is a friend, FALSE otherwise
	*/
	public boolean isFriendOf(User user)
	{
		return loginInfo == null ? false : loginInfo.keyring.getUsers().contains(user.getId());
	}

	/**
		Get the id of all roles where the logged in user is a member.
		@return A <code>Set</code> containing role id:s
	*/
	public Set<Integer> getRoles()
	{
		return loginInfo == null ? null : loginInfo.keyring.getRoles();
	}
	
	/**
		Get the id of all groups where the logged in user is a member, directly or indirectly.
		@return A <code>Set</code> containing group id:s
	*/
	public Set<Integer> getGroups()
	{
		return loginInfo == null ? null : loginInfo.keyring.getGroups(false);
	}

	/**
		Get the id of all friends to logged in user. A users is a friend if the logged in 
		user is a member, directly or indirectly, of a group where the other user is direct member.
		@return A <code>Set</code> containing user id:s
	*/
	public Set<Integer> getFriends()
	{
		return loginInfo == null ? null : loginInfo.keyring.getUsers();
	}

	/**
		Get the id of all projects where the logged in user is a member, directly or indirectly.
		@return A <code>Set</code> containing project id:s
	*/
	public Set<Integer> getProjects()
	{
		return loginInfo == null ? null : loginInfo.keyring.getProjects().keySet();
	}

	/**
		Get the id of all item keys where the logged in user has read permission.
		@return A <code>Set</code> containing item key id:s
	*/
	public Set<Integer> getItemKeys()
	{
		return getItemKeys(Permission.READ);
	}
	
	/**
		Get the id of all item keys where the logged in user has the
		requested permission.
	 	@param permission Requested permissions. 
		@return A <code>Set</code> containing item key id:s
	*/
	public Set<Integer> getItemKeys(Permission permission)
	{
		return loginInfo == null ? null : loginInfo.keyring.getItemKeys(permission);
	}

	/**
		Get the id of all project keys where the logged in user has read permission.
		@return A <code>Set</code> containing project key id:s
	*/
	public Set<Integer> getProjectKeys()
	{
		return getProjectKeys(Permission.READ);
	}
	
	/**
		Get the id of all project keys where the logged in user has the
		requested permission.
	 	@param permission The requested permissions 
		@return A <code>Set</code> containing project key id:s
	*/
	public Set<Integer> getProjectKeys(Permission permission)
	{
		return loginInfo == null ? null : loginInfo.keyring.getProjectKeys(permission);
	}
	
	/**
		Same as: <code>getCurrentContext(item, "", null)</code>.
		@see #getCurrentContext(Item, String, ItemContext)
	*/
	public ItemContext getCurrentContext(Item item)
	{
		return getCurrentContext(item, "", null);
	}
	
	/**
		Same as: <code>getCurrentContext(item, subContext, null)</code>.
		@see #getCurrentContext(Item, String, ItemContext)
	*/
	public ItemContext getCurrentContext(Item item, String subContext)
	{
		return getCurrentContext(item, subContext, null);
	}
	
	/**
		Same as: <code>getCurrentContext(item, "", defaultContext)</code>.
	 	@param item Type of item for which current context should be returned.  
	 	@param defaultContext A default context to copy values from if a new context
			must be created
	 	@return A {@link net.sf.basedb.core.ItemContext} object.
		@see #getCurrentContext(Item, String, ItemContext)
	*/
	public ItemContext getCurrentContext(Item item, ItemContext defaultContext)
	{
		return getCurrentContext(item, "", defaultContext);
	}
	
	/**
		Get the current context. If no current context exists for the item/subcontext
		and a user is logged in, the default context will be loaded from the database. 
		If it doesn't exist in the database or no user is logged in a new current context 
		will be created. If a default context is specified the settings will be copied to
		the new context. The current context is automatically saved in the database when 
		the user logs out. 
		
		@param item The type of item to get the context for
	 	@param subContext Current sub context.
		@param defaultContext A default context to copy values from if a new context
			must be created.
	 	@return A {@link net.sf.basedb.core.ItemContext} object.
		@throws InvalidDataException If the item is null
		@throws BaseException If there is another error
		@see #getCurrentContext(Item)
		@see #loadContextInternal(Item, String, String)
		@see #saveCurrentContextAs(Item, String, boolean, boolean)
	*/
	public ItemContext getCurrentContext(Item item, String subContext, ItemContext defaultContext)
	{
		if (item == null) throw new InvalidUseOfNullException("item");
		ContextKey contextKey = new ContextKey(item, subContext);
		ItemContext current = currentContexts.get(contextKey);
		boolean forceCopy = false;
		if (current == null)
		{
			// Load from db if it doesn't exist, and a user is logged in
			if (isLoggedIn() && getClientId() != 0)
			{
				current = loadContextInternal(item, subContext, ItemContext.DEFAULT_NAME);
			}
			if (current == null)
			{
				// Create new context
				current = new ItemContext(item, subContext, ItemContext.DEFAULT_NAME);
				forceCopy = true;
			}
			currentContexts.put(contextKey, current);
		}
		if (defaultContext != null) current.copyFrom(defaultContext, forceCopy);
		return current;
	}
	
	/**
		Load a context and make it the current context. All information
		in the current context is lost. If a saved context with the specified ID
		doesn't exists, or if the logged in user does't have permission to read it,
		null is returned and the current context is left unchanged.
		
		@param contextId The ID of the context to load
		@return An <code>ItemContext</code> object or null
		@throws NotLoggedInException If no user is logged in
		@throws BaseException If there is another error
		@see #saveCurrentContextAs(Item, String, boolean, boolean)
		@see #getContextNames(Item)
	*/
	public ItemContext loadContext(int contextId)
		throws NotLoggedInException, BaseException
	{
		updateLastAccess();
		if (!isLoggedIn())
		{
			throw new NotLoggedInException();
		}
		ItemContext context = loadContextInternal(contextId);
		if (context != null) 
		{
			ContextKey contextKey = new ContextKey(context.getItemType(), context.getSubContext());
			currentContexts.put(contextKey, context);
		}
		return context;
	}
	
	/**
		Load a context without making it the current context. If a saved context 
		with the specified ID doesn't exists, or if the logged in user does't 
		have permission to read it, null is returned.

		@param contextId The ID of the context to load
		@return An <code>ItemContext</code> object or null
		@throws NotLoggedInException If no user is logged in
		@throws BaseException If there is another error
		@see #loadContext(int)
		@see #getContextNames(Item)
	*/
	public ItemContext getContext(int contextId)
	{
		updateLastAccess();
		if (!isLoggedIn())
		{
			throw new NotLoggedInException();
		}
		ItemContext context = loadContextInternal(contextId);
		return context;
	}
	
	/**
		Same as: <code>saveCurrentContextAs(item, "", name, overwrite, isPublic)</code>.
		@see #saveCurrentContextAs(Item, String, String, boolean, boolean)
	*/
	public int saveCurrentContextAs(Item item, String name, boolean overwrite, boolean isPublic)
		throws PermissionDeniedException, InvalidDataException, ItemAlreadyExistsException, BaseException
	{
		return saveCurrentContextAs(item, "", name, overwrite, isPublic);
	}
	
	/**
		Save the current context as a named context. Named contexts are 
		stored in the database and can be loaded at any time with the
		{@link #loadContext(int)} method. If no current context
		exists, this method does nothing.
		@param item The type of item to save the current context for
		@param subContext The name of the sub-context
		@param name The name of the context
		@param overwrite If TRUE, an existing context with the same name is 
			overwritten, otherwise an exception is thrown
		@param isPublic If the saved context should be visible to everyone
			or not (requires {@link Permission#SHARE_TO_EVERYONE} permission)
	 	@return Saved context's id as an int. 
		@throws PermissionDeniedException If no user is logged in,
			or if the logged in user is a multi-user account or if the <code>isPublic</code>
			parameter is true but the logged in user doesn't have <code>SHARE_TO_EVERYONE</code>
			permission
		@throws InvalidDataException If the item or name parameter is null
		@throws ItemAlreadyExistsException If another context with the 
			same name exists and the overwrite parameter is FALSE
		@throws BaseException If there is another error
	*/
	public int saveCurrentContextAs(Item item, String subContext, String name, boolean overwrite, boolean isPublic)
		throws PermissionDeniedException, InvalidDataException, ItemAlreadyExistsException, BaseException
	{
		updateLastAccess();
		if (!isLoggedIn())
		{
			throw new NotLoggedInException();
		}
		else if (getClientId() == 0)
		{
			throw new PermissionDeniedException("Must use a registered client application to save contexts");
		}
		else if (isPublic && !hasSystemPermission(Permission.SHARE_TO_EVERYONE))
		{
			throw new PermissionDeniedException(Permission.SHARE_TO_EVERYONE);
		}
		if (item == null) throw new InvalidUseOfNullException("item");
		if (name == null) throw new InvalidUseOfNullException("name");
		
		ContextKey contextKey = new ContextKey(item, subContext);
		ItemContext current = currentContexts.get(contextKey);
		int id = 0;
		if (current != null)
		{
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				UserData userData = HibernateUtil.loadData(session, UserData.class, loginInfo.userId);
				if (userData.isMultiuserAccount())
				{
					throw new PermissionDeniedException(Permission.WRITE, "ItemContext[item="+item+"; name="+name+"]");
				}
				ClientData clientData = HibernateUtil.loadData(session, ClientData.class, getClientId());
				ContextIndex index = new ContextIndex(clientData, item.getValue(), subContext, name);
				if (userData.getContexts().containsKey(index))
				{
					if (!overwrite)
					{
						throw new ItemAlreadyExistsException("ItemContext[item="+item+
							"; subContext="+subContext+"; name="+name+"]");
					}
					HibernateUtil.deleteData(session, userData.getContexts().get(index));
				}
				ContextData context = current.getData(name, userData, clientData, isPublic);
				HibernateUtil.saveData(session, context);
				HibernateUtil.commit(tx);
				id = context.getId();
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}
		return id;
	}

	/**
		Delete a context. If a saved context with the specified id doesn't exists, 
		this method does nothing.
		@param contextId The id of the context
		@throws PermissionDeniedException If no user is logged in,
			or if he logged in user is a multi-user account or isn't the owner of the context
		@throws BaseException If there is another error
		@see #getContextNames(Item, String)
	*/
	public void deleteContext(int contextId)
		throws PermissionDeniedException, BaseException
	{
		updateLastAccess();
		if (!isLoggedIn())
		{
			throw new NotLoggedInException();
		}
		else if (getClientId() == 0)
		{
			throw new PermissionDeniedException("Must use a registered client application to save contexts");
		}

		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			UserData userData = HibernateUtil.loadData(session, UserData.class, loginInfo.userId);
			if (userData.isMultiuserAccount())
			{
				throw new PermissionDeniedException(Permission.DELETE, "ItemContext[id="+contextId+"]");
			}
			ContextData context = HibernateUtil.loadData(session, ContextData.class, contextId);
			if (context != null)
			{
				if (context.getUser().getId() != userData.getId())
				{
					throw new PermissionDeniedException(Permission.DELETE, "ItemContext[id="+contextId+"]");
				}
				HibernateUtil.deleteData(session, context);
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
	}

	/**
		Same as: <code>getContextNames(item, "")</code>
		@see #getContextNames(Item, String)
	*/
	public Enumeration<Integer, String> getContextNames(Item item)
		throws InvalidDataException, BaseException
	{
		return getContextNames(item, "");
	}
	
	/**
		Get a list with the names of all named contexts for a given item and
		subcontext.
	 	@param item Item for which named contexts should be returned 
	 	@param subContext Certain sub context to return contexts for.
		@return A enumeration containing the id and names, the list will be empty if no user
			is logged in or if no names contexts exists for the specified item type/subcontext
		@throws InvalidDataException If the item parameter is null
		@throws BaseException If there is another error
	*/
	public Enumeration<Integer, String> getContextNames(Item item, String subContext)
		throws InvalidDataException, BaseException
	{
		updateLastAccess();
		Enumeration<Integer, String> names = new Enumeration<Integer, String>();
		if (!isLoggedIn())
		{
			return names;
		}
		if (item == null) throw new InvalidUseOfNullException("item");
		
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "LOAD_CONTEXT_NAMES");
			/*
				SELECT ctx.id, ctx.name
				FROM ContextData ctx
				WHERE (ctx.user = :user OR ctx.public = true) 
				AND ctx.client = :client AND ctx.itemType = :itemType
				AND ctx.subContext = :subContext
				ORDER BY ctx.name
			*/
			query.setInteger("user", loginInfo.userId);
			query.setInteger("client", getClientId());
			query.setInteger("itemType", item.getValue());
			query.setString("subContext", subContext);
			for (Object ctx : HibernateUtil.loadList(Object.class, query, null))
			{
				Object[] o = (Object[])ctx;
				Integer id = (Integer)o[0];
				String name = (String)o[1];
				if (!ItemContext.DEFAULT_NAME.equals(name))
				{
					names.add(id, name);
				}
			}
			HibernateUtil.commit(tx);
			return names;
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Load a context with the specified ID. If no context with that ID exists or
		if the context belongs to a different client application or user and not is 
		public null is returned.
	*/
	private ItemContext loadContextInternal(int contextId)
		throws BaseException
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		ItemContext context = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			ContextData contextData = HibernateUtil.loadData(session, ContextData.class, contextId);
			if (contextData != null && contextData.getClient().getId() == getClientId() && 
				(contextData.getUser().getId() == getLoggedInUserId() || contextData.isPublic()))
			{
				context = new ItemContext(contextData);
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return context;
	}
	
	/**
		Load a named context. Returns null if no context with the specified name
		exists for the specified item type/subcontext.
	*/
	private ItemContext loadContextInternal(Item item, String subContext, String name)
		throws BaseException
	{
		if (getClientId() == 0) return null;
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		ItemContext context = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "LOAD_CONTEXT_BY_NAME");
			/*
				SELECT ctx
				FROM ContextData ctx
				WHERE ctx.user.id = :user
				AND ctx.client.id = :client
				AND ctx.itemType = :itemType
				AND ctx.subContext  = :subContext
				AND ctx.name = :name
			*/
			query.setInteger("user", loginInfo.userId);
			query.setInteger("client", getClientId());
			query.setInteger("itemType", item.getValue());
			query.setString("subContext", subContext);
			query.setString("name", name);
			
			ContextData cd = HibernateUtil.loadData(ContextData.class, query);
			if (cd != null)
			{
				context = new ItemContext(cd);
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return context;
	}
	
	private void saveAllCurrentContexts(org.hibernate.Session session, UserData userData, ClientData clientData)
	{
		for (ItemContext current : currentContexts.values())
		{
			ContextIndex index = new ContextIndex(clientData, current.getItemType().getValue(), current.getSubContext(), ItemContext.DEFAULT_NAME);
			ContextData old = userData.getContexts().get(index);
			if (old != null) session.delete(old);
			session.save(current.getData(ItemContext.DEFAULT_NAME, userData, clientData, false));
		}
	}
	
	/**
		Get a list with the current default contexts. The list may include, 
		only in-memory contexts, in-database contexts, or both. If both
		options are enabled and an in-memory context also has a database
		representation, both are included in the list. Eg. the list
		may include more than one context were {@link ItemContext#getContextId()}
		returns the same value.
		
		@param inMemory If TRUE, the in-memory contexts are included in the list
		@param inDatabase If TRUE, the in-database contexts are included in the list
		@return A list with ItemContext objects
		@since 2.15
	*/
	public List<ItemContext> getAllCurrentContexts(boolean inMemory, boolean inDatabase)
	{
		List<ItemContext> all = new ArrayList<ItemContext>();
		if (inMemory)
		{
			all.addAll(currentContexts.values());
		}
		if (inDatabase)
		{
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			ItemContext context = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
					"LOAD_USER_CONTEXTS");
					/*
						SELECT ctx 
						FROM ContextData ctx
						WHERE ctx.user.id = :user
						AND ctx.client.id = :client
						AND ctx.name = :name
					*/
				query.setInteger("user", loginInfo.userId);
				query.setInteger("client", getClientId());
				query.setString("name", ItemContext.DEFAULT_NAME);
				List<ContextData> contexts = HibernateUtil.loadList(ContextData.class, query, this);
				for (ContextData ctx : contexts)
				{
					all.add(new ItemContext(ctx));
				}
				HibernateUtil.commit(tx);
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}
		return all;
	}
	
	/**
		Delete all default contexts for the logged in user.
		
		@param itemType The item type to delete the contexts for, or null
			to delete all contexts
		@param inMemory TRUE to delete in-memory contexts
		@param inDatabase TRUE to delete in-database contexts
		@return The number of contexts deleted
		@since 2.8
	*/
	public int deleteCurrentContexts(Item itemType, boolean inMemory, boolean inDatabase)
	{
		int numDeleted = 0;
		if (inMemory) 
		{
			if (itemType == null)
			{
				numDeleted += currentContexts.size();
				currentContexts.clear();
			}
			else
			{
				Iterator<ItemContext> ccit = currentContexts.values().iterator();
				while (ccit.hasNext())
				{
					ItemContext cc = ccit.next();
					if (cc.getItemType() == itemType) 
					{
						ccit.remove();
						numDeleted++;
					}
				}
			}
		}
		
		if (inDatabase)
		{
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			ItemContext context = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
					"LOAD_USER_CONTEXTS");
					/*
						SELECT ctx 
						FROM ContextData ctx
						WHERE ctx.user.id = :user
						AND ctx.client.id = :client
						AND ctx.name = :name
					*/
				query.setInteger("user", loginInfo.userId);
				query.setInteger("client", getClientId());
				query.setString("name", ItemContext.DEFAULT_NAME);
				List<ContextData> contexts = HibernateUtil.loadList(ContextData.class, query, this);
				for (ContextData ctx : contexts)
				{
					if (itemType == null || itemType.getValue() == ctx.getItemType())
					{
						HibernateUtil.deleteData(session, ctx);
						numDeleted++;
					}
				}
				HibernateUtil.commit(tx);
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}
		return numDeleted;
	}
	
	/**
		Get the locale that is curently assigned to this session control. 
		@return A Locale object
		@since 2.16
	*/
	public Locale getLocale()
	{
		return Config.getLocale();
	}
	
	/**
		Get the value of a session setting with the specified name.
		@param name The name of the setting
		@return A object with the value of the setting, or null if
			no setting is found
	*/
	public Object getSessionSetting(String name)
	{
		return loginInfo == null ? null : loginInfo.sessionSettings.get(name);
	}

	/**
		Set the value of a session setting.

		@param name The name of the setting
		@param value The new value of the setting, or null to remove the
			setting
		@return The old value of the setting, or null if it did not exist
	*/
	public Object setSessionSetting(String name, Object value)
	{
		if (loginInfo == null || loginInfo.sessionSettings == null) return null;
		if (value == null)
		{
			return loginInfo.sessionSettings.remove(name);
		}
		else
		{
			return loginInfo.sessionSettings.put(name, value);
		}
	}

	/**
		Get the value of the {@link UserClientSetting} with the
		specified name. If no user is logged in null is returned.

		@param name The name of the setting
		@return A string with the value of the setting, or null if
			no setting is found
	*/
	public String getUserClientSetting(String name)
	{
		if (loginInfo == null || loginInfo.userClientSettings == null) return null;
		SettingInfo si = loginInfo.userClientSettings.get(name);
		return si == null ? null : si.value;
	}
	
	/**
		Set the value of a {@link UserClientSetting}.

		@param name The name of the setting
		@param value The new value of the setting, or null to remove the
			setting
		@return The old value of the setting, or null if it did not exist
			or no user is logged in
	*/
	public String setUserClientSetting(String name, String value)
	{
		if (loginInfo == null || loginInfo.userClientSettings == null) return null;
		SettingInfo si = loginInfo.userClientSettings.put(name, new SettingInfo(value, true));
		return si == null ? null : si.value;
	}

	/**
		Get the value of the {@link UserDefaultSetting} with the
		specified name.
		@param name The name of the setting
		@return A string with the value of the setting, or null if
			no setting is found
	*/
	public String getUserDefaultSetting(String name)
	{
		if (loginInfo == null || loginInfo.userDefaultSettings == null) return null;
		SettingInfo si = loginInfo.userDefaultSettings.get(name);
		return si == null ? null : si.value;
	}

	/**
		Set the value of a {@link UserDefaultSetting}.

		@param name The name of the setting
		@param value The new value of the setting, or null to remove the
			setting
		@return The old value of the setting, or null if it did not exist
			or no user is logged in
	*/
	public String setUserDefaultSetting(String name, String value)
	{
		if (loginInfo == null || loginInfo.userDefaultSettings == null) return null;
		SettingInfo si = loginInfo.userDefaultSettings.put(name, new SettingInfo(value, true));
		return si == null ? null : si.value;
	}

	/**
		Get the value of the {@link ClientDefaultSetting} with the
		specified name.
		@param name The name of the setting
		@return A string with the value of the setting, or null if
			no setting is found
	*/
	public String getClientDefaultSetting(String name)
	{
		if (clientDefaultSettings == null) return null;
		SettingInfo si = clientDefaultSettings.get(name);
		return si == null ? null : si.value;
	}

	/**
		Copy settings from a List to a Map. The key is the name of the setting 
		({@link SettingData#getName()}) and the value is a {@link SettingInfo}
		object.
	*/
	private Map<String, SettingInfo> listToMap(List<? extends SettingData> settings)
	{
		Map<String, SettingInfo> map = Collections.synchronizedMap(new HashMap<String, SettingInfo>(settings.size()));
		for (SettingData s : settings)
		{
			map.put(s.getName(), new SettingInfo(s.getValue(), false));
		}
		return map;
	}

	/**
		Load client applications default settings.
	*/
	private Map<String, SettingInfo> loadClientDefaultSettings(org.hibernate.Session session, int clientId)
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "LOAD_CLIENT_DEFAULT_SETTINGS");
		/*
			SELECT s 
			FROM ClientDefaultSettingData s
			WHERE s.client = :client
		*/
		query.setInteger("client", clientId);
		return listToMap(HibernateUtil.loadList(ClientDefaultSettingData.class, query, null));
	}

	/**
		Load users default settings.
	*/
	private Map<String, SettingInfo> loadUserDefaultSettings(org.hibernate.Session session, int userId)
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "LOAD_USER_DEFAULT_SETTINGS");
		/*
			SELECT s 
			FROM UserDefaultSettingData s
			WHERE s.user = :user
		*/
		query.setInteger("user", userId);
		return listToMap(HibernateUtil.loadList(UserDefaultSettingData.class, query, null));
	}

	/**
		Load users client-specific settings.
	*/
	private Map<String, SettingInfo> loadUserClientSettings(org.hibernate.Session session, int userId, int clientId)
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "LOAD_USER_CLIENT_SETTINGS");
		/*
			SELECT s 
			FROM UserClientSettingData s
			WHERE s.user = :user AND s.client = client
		*/
		query.setInteger("user", userId);
		query.setInteger("client", clientId);
		return listToMap(HibernateUtil.loadList(UserClientSettingData.class, query, null));
	}
	
	/**
		Saves all user settings.
	*/
	private void saveAllUserSettings(org.hibernate.Session session)
	{
		UserData userData = HibernateUtil.loadData(session, UserData.class, loginInfo.userId);
		if (!userData.isMultiuserAccount())
		{
			if (clientId != 0)
			{
				ClientData clientData = HibernateUtil.loadData(session, ClientData.class, clientId);
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "LOAD_USER_CLIENT_SETTINGS");
				query.setEntity("user", userData);
				query.setEntity("client", clientData);
				saveSettings(session, query, loginInfo.userClientSettings, UserClientSettingData.class, userData, clientData);
				saveAllCurrentContexts(session, userData, clientData);
			}
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "LOAD_USER_DEFAULT_SETTINGS");
			query.setEntity("user", userData);
			saveSettings(session, query, loginInfo.userDefaultSettings, UserDefaultSettingData.class, userData, null);
		}
	}

	
	/**
		Save settings when a user logs out. This method will delete settings that have been
		deleted, update changed settings and create new settings.
	
		@param session The Hibernate session
		@param query A Hibernate query to load the settings already in the database
		@param settings The settings to save
		@param dataClass The <code>Class</code> of the settings, should be one of
			{@link UserDefaultSettingData} or {@link UserClientSettingData}
		@param userData The {@link UserData} object for the logged in user
		@param clientData The {@link ClientData} object for the client application
	*/
	private <I extends SettingData>
	void saveSettings(org.hibernate.Session session, org.hibernate.Query query, Map<String, SettingInfo> settings, Class<I> dataClass, UserData userData, ClientData clientData)
		throws BaseException
	{
		// Load the existing settings from the database
		List<I> dbSettings = HibernateUtil.loadList(dataClass, query, null);
		// Iterator the existing settings and check what has happened to them
		for (SettingData s : dbSettings)
		{
			// Get the info, use remove so we know which ones have been processed
			SettingInfo si = settings.remove(s.getName());
			// Check if the setting has been modified
			if (si != null && si.modified)
			{
				// A null value --> delete the setting
				if (si.value == null)
				{
					HibernateUtil.deleteData(session, s);
				}
				else
				{
					s.setValue(si.value);
				}
			}
		}
		// The settings map now contain only those setting not found in the database
		try
		{
			for (Map.Entry<String, SettingInfo> entry : settings.entrySet())
			{
				SettingInfo si = entry.getValue();
				// If the setting has been modified, it should be created
				if (si != null && si.value != null && si.modified)
				{
					I s = dataClass.newInstance();
					s.setName(entry.getKey());
					s.setValue(si.value);
					if (s instanceof UserDefaultSettingData)
					{
						((UserDefaultSettingData)s).setUser(userData);
					}
					if (s instanceof UserClientSettingData)
					{
						((UserClientSettingData)s).setUser(userData);
						((UserClientSettingData)s).setClient(clientData);
					}
					HibernateUtil.saveData(session, s);
				}
			}
		}
		catch (InstantiationException ex)
		{
			throw new BaseException(ex);
		}
		catch (IllegalAccessException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	protected void clearDbControlCache()
	{
		// Close any open DbControl objects obtained during this login
		synchronized (dbControlCache)
		{
			for (DbControl dc : dbControlCache.keySet())
			{
				dc.close();
			}
			dbControlCache.clear();
		}
	}
	
	protected void cleanUp()
	{
		if (!isClosed() && isLoggedIn()) logout();
	}
	
	/**
		Clean up if a bad client application forgets to logout.
	*/
	@Override
	protected void finalize()
		throws Throwable
	{
		try
		{
			cleanUp();
		}
		catch (Throwable t)
		{
			log.warn("Exception during SessionControl.finalize()", t);
		}
		super.finalize();
	}
	
	/**
		Internal class to hold information about the logged in user.
	*/
	private static class LoginInfo
	{
		/**
			If the logged in user was logged in with the {@link SessionControl#impersonateLogin(int, String)}
		*/
		private boolean impersonated;
	
		/**
			The id of the {@link SessionData}.
		*/
		private int sessionId;

		/**
			The id of the {@link UserData} object for the logged in user.
		*/
		private int userId;
		
		/**
			The login property of the logged in user.
		*/
		private String userLogin;
		
		/**
			The authentication method used to allow the user to
			login.
		*/
		private AuthenticationMethod authenticationMethod;
		
		/**
			The id of the {@link ProjectData} object of the active project.
		*/
		private int activeProjectId;

		/**
			The id of the {@link ProjectKeyData} object that new items should 
			be shared to.
		*/
		private int projectKeyId;
		
		/**
			The id of the {@link ItemKeyData} object that new items should be
			shared to.
		*/
		private int itemKeyId;
		
		/**
			The {@link Keyring} of the logged in user.
		*/
		private Keyring keyring;
		
		/**
			Stores name and value from {@link UserClientSettingData}.
		*/
		private Map<String,SettingInfo> userClientSettings;

		/**
			Stores name and value from {@link UserDefaultSettingData}.
		*/
		private Map<String,SettingInfo> userDefaultSettings;

		/**
			Stores name and value of session settings.
		*/
		private Map<String,Object> sessionSettings;

		private LoginInfo()
		{}
		
		private LoginInfo(LoginInfo parent)
		{
			this.impersonated = parent.impersonated;
			this.sessionId = parent.sessionId;
			this.userId = parent.userId;
			this.userLogin = parent.userLogin;
			this.activeProjectId = parent.activeProjectId;
			this.projectKeyId = parent.projectKeyId;
			this.keyring = parent.keyring;
			this.userClientSettings = parent.userClientSettings;
			this.userDefaultSettings = parent.userDefaultSettings;
			this.sessionSettings = parent.sessionSettings;
		}
		
	}
	
	/**
		Internal class to hold information about a setting
	*/
	private static class SettingInfo
	{
		/**
			The value of the setting, null indicates removal
		*/
		private final String value;
		/**
			If the setting has been modified while the user was logged in or not
		*/
		private final boolean modified;
		
		private SettingInfo(String value, boolean modified)
		{
			this.value = value;
			this.modified = modified;
		}
		
		@Override
		public String toString()
		{
			return "value=" + value + "; modified=" + modified;
		}
	}
	
}
