/*
	$Id: VirtualDb.java 6444 2014-04-09 12:21:20Z nicklas $
	
	Copyright (C) 2005 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.VirtualDbData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.hibernate.ExecuteUpdateWork;
import net.sf.basedb.util.filter.Filter;

import java.sql.SQLException;

/**
	A virtual database represents the set of tables in the dynamic database used
	by a single experiment to store it's analysed data.
	<p>
	This class is not intended to be exposed to client applications. It is only
	used internally to handle the creation and deletion of the tables in the
	dynamic database.
	<p>
	The names of the dynamic tables depends on the ID of this object. The name
	is generated by the {@link VirtualTable} enumeration.
	
	@author Nicklas
	@version 2.0
	@see VirtualTable
	@see VirtualColumn
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (on, 09 apr 2014) $
*/
public class VirtualDb
	extends ChildItem<VirtualDbData>
{
	/**
		Log all SQL statements.
	*/
	private static final org.slf4j.Logger logSql = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.batcher.sql");
	
	/**
		So we don't always have to call logSql.debug()
	*/
	private static final boolean debugSqlEnabled = logSql.isDebugEnabled();

	/**
		The type of item represented by this class.
		@see Item#VIRTUALDB
		@see #getType()
	*/
	public static final Item TYPE = Item.VIRTUALDB;

	/**
		Create a new <code>DynamicTables</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@return The new <code>DynamicTables</code> item
		@throws BaseException If there is an error
	*/
	static VirtualDb getNew(DbControl dc)
		throws BaseException
	{
		VirtualDb vdb = dc.newItem(VirtualDb.class);
		return vdb;
	}

	/**
		Get a <code>VirtualDb</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>VirtualDb</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	static VirtualDb getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		VirtualDb vdb = dc.loadItem(VirtualDb.class, id);
		if (vdb == null) throw new ItemNotFoundException("VirtualDb[id=" + id + "]");
		return vdb;
	}


	/**
	 	Creates a new experiment item from the given data.
		@param data the data
	*/
	VirtualDb(VirtualDbData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		Remove all dynamic tables if action == Action.DELETE
	*/
	@Override
	void onAfterCommit(Transactional.Action action)
	{
		super.onAfterCommit(action);
		if (action == Transactional.Action.DELETE)
		{
			dropAllTables();
		}
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getExperiment();
	}
	/**
		USE permission is enough to be able to manage virtual db.
	*/
	@Override
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.USE;
	}
	// -------------------------------------------

	/**
		Get the raw data type the experiment uses for the raw data.
		@return A <code>RawDataType</code> object
	*/
	public RawDataType getRawDataType()
	{
		return RawDataTypes.getSafeRawDataType(getData().getExperiment().getRawDataType());
	}

	/**
		Get the experiment this set of tables stores data for.
		@return An <code>Experiment</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Experiment getExperiment()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Experiment.class, getData().getExperiment());
	}
	
	/**
		Get the reporter clone template that is active in the current virtual database.
		If no cloned reporter information exists, this method return null.
		@return A <code>ReporterCloneTemplate</code> (locked) or null
		@since 3.1
	*/
	public ReporterCloneTemplate getReporterCloneTemplate()
	{
		return getDbControl().getItem(ReporterCloneTemplate.class, getData().getReporterCloneTemplate());
	}
	
	/**
		Check if this virtual db is storing cloned reporter information or not.
		@since 3.1
	*/
	public boolean hasClonedReporters()
	{
		return getData().getReporterCloneTemplate() != null;
	}
	
	/**
		Checks if this virtual db has cloned the given reporter property or not.
		@param propertyName The name of a reporter property
		@since 3.1
	*/
	public boolean hasClonedReporterProperty(String propertyName)
	{
		return hasClonedReporters() && getData().getReporterCloneTemplate().getProperties().containsKey(propertyName);
	}
	
	/**
		Return a filter that can be used to check that a query is only
		using cloned reporter properties for sorting, filtering, etc.
		The filter will reject reporter properties (starting with '@')
		that have not been cloned.
		
		@return A filter or null if this virtual db doesn't have cloned reporters
		@since 3.1
		@see ItemContext#setPropertyInspector(Filter)
	 */
	public Filter<String> getQueryPropertyInspector()
	{
		Filter<String> filter = null;
		if (hasClonedReporters())
		{
			filter = new Filter<String>()
			{
				@Override
				public boolean evaluate(String property)
				{
					if (!property.startsWith("@")) return true;
					return hasClonedReporterProperty(property.substring(1));
				}
			};
		}
		return filter;
	}
	
	/**
		Get the number of {@link DataCube}:s in this virtual database.
		Each new data cube increases this value with one.
	*/
	public short getCubes()
	{
		return getData().getCubes();
	}

	/**
		Generate the next available cube number. This method will open a
		second transaction and update the {@link VirtualDb#getCubes()}
		property by one.
		@return The next cube number
	*/
	short getNextCubeNo()
		throws BaseException
	{
		short cubeNo = 0;
		if (!isInDatabase())
		{
			cubeNo = (short)(getCubes() + 1);
		}
		else
		{
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				
				// Increase the number of cubes
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "UPDATE_DATACUBENO_FOR_VIRTUALDB");
				/*
					UPDATE VirtualDbData vdb
					SET vdb.cubes = vdb.cubes + 1
					WHERE vdb.id = :virtualDb
				*/
				query.setInteger("virtualDb", this.getId());
				HibernateUtil.executeUpdate(query);
				
				// Get the new value
				query = HibernateUtil.getPredefinedQuery(session, "GET_DATACUBENO_FOR_VIRTUALDB");
				/*
					SELECT vdb.cubes
					FROM VirtualDbData vdb
					WHERE vdb.id = :virtualDd
				*/
				query.setInteger("virtualDb", this.getId());
				cubeNo = HibernateUtil.loadData(Short.class, query);
				HibernateUtil.commit(tx);
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}

		// Update data object and return
		getData().setCubes(cubeNo);
		return cubeNo;
	}
	
	// DataCubeMapping class does not exsist, but will in the near future.
	/**
		Get the number of DataCubeMapping:s in this virtual database.
		Each new mapping increases this value with one.
	*/
	public short getMappings()
	{
		return getData().getMappings();
	}

	/**
		Create a new data cube in this virtual database. The number of
		the new data cube will be one more than the current value of
		{@link #getCubes()}.
	*/
	DataCube newDataCube()
		throws PermissionDeniedException, BaseException
	{
		return DataCube.getNew(getDbControl(), this);
	}

	/**
		Create the real database tables for this virtual database. The table
		is created by first building a Hibernate mapping structure and then
		using Hibernate to generate the SQL that is needed.

		@param tables The tables to create
		@see HibernateUtil#createVirtualTable(VirtualDb, VirtualTable)
	*/
	void createTables(VirtualTable... tables)
		throws BaseException
	{
		for (VirtualTable table : tables)
		{
			HibernateUtil.createVirtualTable(this, table);
		}
	}
	
	/**
		Check if a real database table for this virtual database exists.
		@param table The table to check
	 	@return TRUE if database table exists, FALSE otherwise 
	 	@throws BaseException If there is some kind of error.
		@see HibernateUtil#virtualTableExists(VirtualDb, VirtualTable)
	*/
	public boolean hasTable(VirtualTable table)
		throws BaseException
	{
		return HibernateUtil.virtualTableExists(this, table);
	}
	
	/**
		Drop all real database tables in this virtual database. The tables
		are deleted by first building a Hibernate mapping structure and then
		using Hibernate to generate the SQL that is needed.
		@see HibernateUtil#dropVirtualTable(VirtualDb, VirtualTable)
	*/
	private void dropAllTables()
		throws BaseException
	{
		for (VirtualTable table : VirtualTable.values())
		{
			HibernateUtil.dropVirtualTable(this, table);
		}
	}
	
	/**
		Delete all analysed data for the specified data cube.

		@param cubeNo The number of the cube
		@return The number of deleted rows
		@see DataCube#onBeforeCommit(Transactional.Action)
	*/
	int deleteAllCubeData(short cubeNo)
		throws BaseException
	{
		int rowsDeleted = 0;
		for (VirtualTable table : VirtualTable.values())
		{
			if (HibernateUtil.virtualTableExists(this, table))
			{
				String tableName = table.getQualifiedTableName(this);
				String sql = "DELETE FROM " + tableName + " WHERE "+
					HibernateUtil.quote("cube")+" = " + cubeNo;
				try
				{
					if (debugSqlEnabled) logSql.debug(sql);
					rowsDeleted += HibernateUtil.doJdbcWork(getDbControl().getHibernateSession(),
							new ExecuteUpdateWork(getSessionControl(), sql));
				}
				catch (SQLException ex)
				{
					logSql.error(sql, ex);
					throw new BaseException(ex);
				}
			}
		}
		return rowsDeleted;
	}
	
	/**
		Delete all analysed data for the specified layer in the specified data cube.

		@param cubeNo The number of the cube
		@param layerNo The number of the layer
		@return The number of deleted rows
		@see DataCubeLayer#onBeforeCommit(Transactional.Action)
	*/
	int deleteAllCubeLayerData(short cubeNo, short layerNo)
		throws BaseException
	{
		int rowsDeleted = 0;
		if (HibernateUtil.virtualTableExists(this, VirtualTable.SPOT))
		{
			String tableName = VirtualTable.SPOT.getQualifiedTableName(this);
			String sql = "DELETE FROM " + tableName + " WHERE "+
				HibernateUtil.quote("cube")+" = " + cubeNo + " AND "+
				HibernateUtil.quote("layer")+" = " + layerNo;
			try
			{
				if (debugSqlEnabled) logSql.debug(sql);
				rowsDeleted += HibernateUtil.doJdbcWork(getDbControl().getHibernateSession(),
						new ExecuteUpdateWork(getSessionControl(), sql));
			}
			catch (SQLException ex)
			{
				logSql.error(sql, ex);
				throw new BaseException(ex);
			}
		}
		return rowsDeleted;
	}

	/**
		Delete all analysed data for the specified filter in the specified data cube.

		@param cubeNo The number of the cube
		@param filterNo The number of the filter
		@return The number of deleted rows
		@see DataCubeFilter#onBeforeCommit(Transactional.Action)
	*/
	int deleteAllCubeFilterData(short cubeNo, short filterNo)
		throws BaseException
	{
		int rowsDeleted = 0;
		if (HibernateUtil.virtualTableExists(this, VirtualTable.FILTER))
		{
			String tableName = VirtualTable.FILTER.getQualifiedTableName(this);
			String sql = "DELETE FROM " + tableName + " WHERE "+
				HibernateUtil.quote("cube")+" = " + cubeNo + " AND "+
				HibernateUtil.quote("filter")+" = " + filterNo;
			
			try
			{
				if (debugSqlEnabled) logSql.debug(sql);
				rowsDeleted += HibernateUtil.doJdbcWork(getDbControl().getHibernateSession(), 
						new ExecuteUpdateWork(getSessionControl(), sql));
			}
			catch (SQLException ex)
			{
				logSql.error(sql, ex);
				throw new BaseException(ex);
			}
		}
		return rowsDeleted;
	}

	/**
		Delete all extra value data for the specified extra value type in the
		specified data cube.

		@param cubeNo The number of the cube
		@param extraNo The number of the extra value type
		@return The number of deleted rows
		@see DataCubeExtraValue#onBeforeCommit(Transactional.Action)
	*/
	int deleteAllCubeExtraValueData(short cubeNo, short extraNo, VirtualTable extraTable)
		throws BaseException
	{
		int rowsDeleted = 0;
		if (HibernateUtil.virtualTableExists(this, extraTable))
		{
			String tableName = extraTable.getQualifiedTableName(this);
			final String sql = "DELETE FROM " + tableName + " WHERE "+
				HibernateUtil.quote("cube")+" = " + cubeNo + " AND " +
				HibernateUtil.quote("extra")+" = " + extraNo;
			
			try
			{
				if (debugSqlEnabled) logSql.debug(sql);
				rowsDeleted += HibernateUtil.doJdbcWork(getDbControl().getHibernateSession(),
						new ExecuteUpdateWork(getSessionControl(), sql)); 
			}
			catch (SQLException ex)
			{
				logSql.error(sql, ex);
				throw new BaseException(ex);
			}
		}
		return rowsDeleted;
	}
}
