/**
	$Id: ItemSubtypeData.java 6214 2012-12-13 10:23:32Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
	Item subtypes can be used to classify items into sub-categories.
	Subtypes can be defined for all item types that implement the
	{@link SubtypableData} interface. For a given item type the name
	of the subtype should be unique.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2012-12-13 11:23:32 +0100 (to, 13 dec 2012) $
	@hibernate.class table="`ItemSubtypes`" lazy="false"
	@see <a href="../../../../../../html/developer/api/data_api.platforms.html">Developer documentation: Experimental platforms and item subtypes</a>
*/
public class ItemSubtypeData
	extends BasicData
	implements NameableData, RemovableData, RegisteredData, SystemData
{

	public ItemSubtypeData()
	{}
	
	/*
		From the NameableData interface
		-------------------------------------------
	*/
	private String name;
	/**
		Get the name of the subtype which must be unique for subtypes related to a given item type.
		@hibernate.property type="string"
		@hibernate.column name="`name`" length="255" not-null="true" index="name_idx" unique-key="uniquesubtype"
	*/
	@Override
	public String getName()
	{
		return name;
	}
	@Override
	public void setName(String name)
	{
		this.name = name;
	}
	private String description;
	@Override
	public String getDescription()
	{
		return description;
	}
	@Override
	public void setDescription(String description)
	{
		this.description = description;
	}
	// -------------------------------------------
	/*
		From the RemovableData interface
		-------------------------------------------
	*/
	private boolean removed;
	@Override
	public boolean isRemoved()
	{
		return removed;
	}
	@Override
	public void setRemoved(boolean removed)
	{
		this.removed = removed;
	}
	// -------------------------------------------

	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	@Override
	public Date getEntryDate()
	{
		return entryDate;
	}
	public void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------

	/*
		From the SystemData interface
		-------------------------------------------
	*/
	private String systemId;
	/**
		Override the property definition since we may have to update the system id
		during installation.
		@hibernate.property column="`system_id`" type="string" length="255" not-null="false" update="true"
	*/
	@Override
	public String getSystemId()
	{
		return systemId;
	}
	public void setSystemId(String systemId)
	{
		this.systemId = systemId;
	}
	// -------------------------------------------
	
	private int itemType;
	/**
		Get the item type this subtype can be applied to.
		Can't be modified after the subtype has been created.
		@hibernate.property update="false" 
		@hibernate.column name="`item_type`" type="int" not-null="true" unique-key="uniquesubtype"
	*/
	public int getItemType()
	{
		return itemType;
	}
	public void setItemType(int itemType)
	{
		this.itemType = itemType;
	}

	private boolean pushAnnotations;
	/**
		If this flag is set, the annotations on items of this subtype should be 
		enabled for pushing to the parent item.
		
		NOTE! This property is mapped in hibernate-properties-ItemSubtypeData.xml since
		'default' is not supported by XDoclet
		//hibernate.property type="boolean"
		//hibernate.column name="`push_annotations`" not-null="true" default="false"

		@since 3.1
	*/
	public boolean getPushAnnotations()
	{
		return pushAnnotations;
	}
	public void setPushAnnotations(boolean pushAnnotations)
	{
		this.pushAnnotations = pushAnnotations;
	}
	
	private Map<Integer, ItemSubtypeData> relatedSubtypes;
	/**
		@hibernate.map table="`RelatedSubtypes`" lazy="true"
		@hibernate.collection-key column="`itemsubtype_id`"
		@hibernate.collection-index column="`item_type`" type="int" not-null="true"
		@hibernate.collection-many-to-many column="`related_subtype_id`" class="net.sf.basedb.core.data.ItemSubtypeData"
	*/
	public Map<Integer, ItemSubtypeData> getRelatedSubtypes()
	{
		if (relatedSubtypes == null)
		{
			relatedSubtypes = new HashMap<Integer, ItemSubtypeData>();
		}
		return relatedSubtypes;
	}
	void setRelatedSubtypes(Map<Integer, ItemSubtypeData> relatedSubtypes)
	{
		this.relatedSubtypes = relatedSubtypes;
	}
	
	private Set<ItemSubtypeData> parents;
	/**
		This is the inverse end. See {@link #getRelatedSubtypes()}.
		@hibernate.set table="`RelatedSubtypes`" lazy="true"
		@hibernate.collection-key column="`related_subtype_id`"
		@hibernate.collection-many-to-many column="`itemsubtype_id`" class="net.sf.basedb.core.data.ItemSubtypeData"
	*/
	Set<ItemSubtypeData> getParents()
	{
		return parents;
	}
	void setParents(Set<ItemSubtypeData> parents)
	{
		this.parents = parents;
	}
	
	private Map<DataFileTypeData, ItemSubtypeFileTypeData> fileTypes;
	/**
		The file types that can be used on items of this subtype.
		This is the inverse end. 
		@see ItemSubtypeFileTypeData#getItemSubtype()
		@hibernate.map lazy="true" cascade="delete" inverse="true"
		@hibernate.collection-key column="`subtype_id`"
		@hibernate.index-many-to-many column="`datafiletype_id`" class="net.sf.basedb.core.data.DataFileTypeData"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.ItemSubtypeFileTypeData"
	*/
	public Map<DataFileTypeData, ItemSubtypeFileTypeData> getFileTypes()
	{
		if (fileTypes == null)
		{
			fileTypes = new HashMap<DataFileTypeData, ItemSubtypeFileTypeData>();
		}
		return fileTypes;
	}
	void setFileTypes(Map<DataFileTypeData, ItemSubtypeFileTypeData> fileTypes)
	{
		this.fileTypes = fileTypes;
	}

	
}
