/*
	$Id: ParameterValueData.java 6354 2013-11-25 13:38:13Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.Type;

/**
	The base class for the different types of parameter values.
	@author Nicklas, Samuel
	@version 2.0
	@see <a href="../../../../../../html/developer/api/data_api.parameters.html">Developer documentation: Parameters</a>
	@hibernate.class table="`ParameterValues`" discriminator-value="-1" lazy="true" batch-size="10"
	@hibernate.discriminator column="`discriminator`" type="int"
	@base.modified $Date: 2013-11-25 14:38:13 +0100 (må, 25 nov 2013) $
*/
public abstract class ParameterValueData<T>
	extends BasicData
{
	// Keep the old values for logging purposes
	private transient List<T> oldValues;
	
	ParameterValueData()
	{}
	
	@SafeVarargs
	ParameterValueData(T... values)
	{
		replaceValues(Arrays.asList(values));
	}

	/**
		The maximum length of the label of the parameter that can be
		stored in the database.
		@see #setLabel(String)
	*/
	public static final int MAX_LABEL_LENGTH = 255;
	private String label;
	/**
		Get the label of the parameter.
		@hibernate.property column="`label`" type="string" size="255" not-null="false"
	*/
	public String getLabel()
	{
		return label;
	}
	public void setLabel(String label)
	{
		this.label = label;
	}
	
	/**
		The maximum length of the description of the parameter that can be
		stored in the database.
		@see #setDescription(String)
	*/
	public static final int MAX_DESCRIPTION_LENGTH = 65535;
	private String description;
	/**
		Get the description of the parameter.
		@hibernate.property column="`description`" type="text" not-null="false"
	*/
	public String getDescription()
	{
		return description;
	}
	public void setDescription(String description)
	{
		this.description = description;
	}
	
	/**
		Get the type of parameter values that are store in this parameter.
		The value should correspond to one of the {@link Type#getValue()}
		values, or 0 if the parameter has no corresponding type.
		@return An integer value found in the {@link Type} enumeration
			or 0
		@since 2.16
	*/
	public abstract int getValueType();
	
	/**
		Get values of this parameter. Hibernate mapped in each subclass 
		to a bag containing all values for the parameter.
		@return Collection of values.
	*/
	public abstract List<T> getValues();
	
	/**
		Set values of this parameter.
		<p>
		NOTE! Used by Hibernate to initialize the object. Do not 
		call {@link AnnotationData#setLastUpdate(java.util.Date)}!
		@param values Collection of values.
	*/
	abstract void setValues(List<T> values);
	
	/**
		Replace the values in the database with the new
		values in the specified list. It is expected that
		the list contains only objects of the correct type.
		<p>
		NOTE! When this method is called to set annotation values
		for an {@link AnnotationData}, don't forget to also update
		the last update timestamp: {@link AnnotationData#setLastUpdate(java.util.Date)}.
		and to invalidate the annotation set snapshot: {@link AnnotationSet#setSnapshotInvalid()}
	*/
	@SuppressWarnings("unchecked")
	public void replaceValues(List<?> values)
	{
		List<T> current = getValues();
		if (current.size() > 0 && oldValues == null) 
		{
			oldValues = new ArrayList<T>(current);
		}
		current.clear();
		if (values != null) current.addAll((List<T>) values);
	}
	
	/**
		Replace the current list with a single new value.
		It is expected that the value is of the correct type.
		<p>
		NOTE! When this method is called to set annotation values
		for an {@link AnnotationData}, don't forget to also update
		the last update timestamp: {@link AnnotationData#setLastUpdate(java.util.Date)}.
		and to invalidate the annotation set snapshot: {@link AnnotationSet#setSnapshotInvalid()}
	*/
	@SuppressWarnings("unchecked")
	public void setSingleValue(Object value)
	{
		List<T> current = getValues();
		if (current.size() > 0 && oldValues == null) 
		{
			oldValues = new ArrayList<T>(current);
		}
		current.clear();
		current.add((T) value);
	}
	
	/**
		Get the old values that has been replaced by new values in this session.
		@return A list or null if the values has not been replaced
		@since 3.3
	*/
	public List<T> getOldValues()
	{
		return oldValues;
	}
}
