/**
	$Id: AbstractDbEngine.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006, 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.dbengine;

import java.util.regex.Pattern;

import net.sf.basedb.core.hibernate.SchemaGenerator;

import org.hibernate.dialect.Dialect;

/**
	An abstract superclass with default implementations for most {@link DbEngine}
	methods. Subclasses only needs to override methods which behave differently.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public abstract class AbstractDbEngine
	implements DbEngine
{

	/**
		A regexp checking for invalid characters.
	*/
	private static final Pattern valid = Pattern.compile("[a-zA-Z_][a-zA-Z0-9_]*");
	
	/**
		Create AbstractDbEngine.
	*/
	public AbstractDbEngine()
	{}
	
	/*
		From the DbEngine interface
		-------------------------------------------
	*/
	/**
		Return the SQL unmodified.
		@since 2.6
	*/
	@Override
	public String makeSafeCreateTable(String sql, String catalog, String schema, String table)
	{
		return sql;
	}
	/**
		Most databases doesn't. Any more than Oracle?
	*/
	@Override
	public boolean useThetaJoin()
	{
		return false;
	}

	/**
		Returns FALSE. It should always work.
	*/
	@Override
	public boolean supportsColumnAliasInWhere()
	{
		return false;
	}

	/**
		Returns FALSE. It should always work.
	*/
	@Override
	public boolean supportColumnAliasInOrderBy()
	{
		return false;
	}

	/**
		Returns FALSE. It should always work.
	*/
	@Override
	public boolean supportColumnAliasInGroupBy()
	{
		return false;
	}

	/**
		Returns FALSE. It should always work.
	*/
	@Override
	public boolean supportColumnAliasInHaving()
	{
		return false;
	}
	/**
		Returns FALSE.
	*/
	@Override
	public boolean checkForInvalidNumberOperation()
	{
		return false;
	}
	/**
		Returns FALSE.
	*/
	@Override
	public boolean selectOrderByColumnsIfDistinct()
	{
		return false;
	}

	/**
		Return <code>LN(&lt;value&gt;)</code>.
	*/
	@Override
	public String ln(String value)
	{
		return "LN("+ value + ")";
	}
	/**
		Return <code>ABS(&lt;value&gt;)</code>.
	*/
	@Override
	public String abs(String value)
	{
		return "ABS("+ value + ")";
	}
	/**
		Return <code>EXP(&lt;value&gt;)</code>.
	*/
	@Override
	public String exp(String value)
	{
		return "EXP("+ value + ")";
	}
	/**
		Return <code>POWER(&lt;base&gt;, &lt;exponent&gt;)</code>.
	*/
	@Override
	public String power(String base, String exponent)
	{
		return "POWER("+ base + "," + exponent + ")";
	}
	/**
		Return <code>&lt;value&gt; = &lt;regexp&gt;</code>
		@since 2.8
	*/
	@Override
	public String rlike(String value, String regexp)
	{
		return value + "=" + regexp;
	}
	
	/**
		Checks that the name only contains the following characters: a-zA-Z0-9_
		It must also start with a letter or underscore.
		@since 2.4
	*/
	@Override
	public boolean isValidTableName(String tableName)
	{
		return isValidName(tableName);
	}
	
	/**
		Checks that the name only contains the following characters: a-zA-Z0-9_
		It must also start with a letter or underscore.
	*/
	@Override
	public boolean isValidColumnName(String columnName)
	{
		return isValidName(columnName);
	}
	
	/**
		Return null.
		@since 2.9
	 */
	@Override
	public String getCaseSensitiveVarchar(int length)
	{
		return null;
	}
	
	/**
		@return ANSI SQL: cast(value as date)
	*/
	@Override
	public String castToDate(String value)
	{
		return "cast(" + value + " as date)";
	}
	
	/**
		Remove "not null" constraints from "alter table ... add column"
		statements that doesn't supply a default value for the new column
		when updating.
		@since 3.1
	*/
	@Override
	public String inspectSchemaGenerationSQL(String sql, Dialect dialect,
			SchemaGenerator.Mode mode)
	{
		String addColumn = dialect.getAddColumnString();
		if (mode == SchemaGenerator.Mode.UPDATE && 
				sql.contains("alter table") && sql.contains(addColumn) && 
				sql.contains("not null") && !sql.contains("default"))
		{
			sql = sql.replace("not null", "");
		}
		return sql;
	}
	/**
		Put quotes (") around the name.
	*/
	@Override
	public String getQuotedName(String name)
	{
		return "\"" + name + "\"";
	}
	/**
		@return Always FALSE, but may be overridden by subclasses
	 */
	@Override
	public boolean useSavePointToContinueTransactionFromSqlFailure() 
	{
		return false;
	}
	/**
		@return null, since this depends on capabilities of underlying database
	*/
	@Override
	public String getApproximateRowCountSql(String catalog, String schema, String table)
	{
		return null;
	}
	// -------------------------------------------

	/**
		Check that the name only contains a-zA-Z0-9_ and starts with
		a letter or underscore.
		@since 2.4
	*/
	protected boolean isValidName(String name)
	{
		return name == null ? false : valid.matcher(name).matches();
	}
	
	
}
