/**
	$Id: ResultSetWork.java 5881 2011-11-21 10:54:26Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;


import net.sf.basedb.core.QueryExecutor;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.signal.SignalException;


/**
	A work implementation that executes an SQL query on the connection,
	and returns a value that is picked by a subclass from the resulting
	{@link ResultSet}. This class needs to be subclassed and implementors
	should provide an implementation of the {@link #getResult(ResultSet)}
	method. By overridding the {@link #setParameters(PreparedStatement)}
	method it is possible to set parameters on the prepared statement before 
	the SQL is executed.

	@author nicklas
	@version 2.9
	@base.modified $Date: 2011-11-21 11:54:26 +0100 (må, 21 nov 2011) $
*/
public abstract class ResultSetWork<R>
	implements JdbcWork<R>
{

	private final SessionControl sc;
	private final String sql;
	private R result;
	
	/**
		Create a new work item.
		@param sc The session control to keep alive during the execution
			of the sql, or null
		@param sql The sql to execute
	*/
	public ResultSetWork(SessionControl sc, String sql)
	{
		this.sc = sc;
		this.sql = sql;
	}
	
	/**
		From the Work interface
		-----------------------
	*/
	@Override
	public void execute(Connection connection) 
		throws SQLException
	{
		PreparedStatement ps = null;
		ResultSet rs = null;
		try
		{
			ps = connection.prepareStatement(sql);
			setParameters(ps);
			rs = QueryExecutor.executeQuery(ps, sc);
			result = getResult(rs);
		}
		catch (InterruptedException ex)
		{
			throw new SignalException("Aborted by user", ex);
		}
		finally
		{
			try
			{
				if (rs != null) rs.close();
				if (ps != null) ps.close();
			}
			catch (Throwable t)
			{}
		}
	}
	// --------------------------
	
	/*
		From the JdbcWork interface
		---------------------------
	*/
	/**
		@return The result from the {@link #getResult(ResultSet)} method.
	*/
	@Override
	public R getResult()
	{
		return result;
	}
	// ----------------------------

	/**
		The that is executed by this work.
		@since 3.1
	*/
	public String getSql()
	{
		return sql;
	}
	
	/**
		This method should be overridden by subclasses that needs to
		set parameters on the statement that is going to be executed.
		The default implementation does nothing. This method is called
		after the statement has been prepared with 
		{@link Connection#prepareStatement(String)} and before it is executed
		with {@link PreparedStatement#executeUpdate()}.
		@param ps The prepared statement
	*/
	protected void setParameters(PreparedStatement ps)
		throws SQLException
	{}

	/**
		This method needs to be overridden by subclasses to extract the 
		result from the {@link ResultSet}. The result set will be closed
		after the call to this method.
	
		@param rs The result set to get the result from
	*/
	protected abstract R getResult(ResultSet rs)
		throws SQLException;
	
}
