/**
	$Id: SchemaGenerator.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.dbengine.DbEngine;

import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.Dialect;
import org.hibernate.jdbc.Work;
import org.hibernate.tool.hbm2ddl.DatabaseMetadata;

/**
	Replaces the Hibernate SchemaExport and SchemaUpdate tools. We need
	this to be able to intercept and modify some of the SQL statments that
	Hibernate generates since they don't always work as expected.
	In particular, since Hibernate 3.5 they generate "not null" constraints for
	columns that are added to existing tables and we don't want that.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (on, 09 apr 2014) $
*/
public class SchemaGenerator
	implements Work
{
	
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(SchemaGenerator.class);
	
	private final Configuration cfg;
	private final Dialect dialect;
	private final DbEngine dbEngine;
	private final Mode mode;
	private final ProgressReporter progress;
	private List<String> ignoredSql;
	
	/**
		Create a new schema generator. 
		
		@param cfg The current configuration
		@param dialect The dialect of the database we are connecting to
		@param dbEngine The DbEngine for the database we are connecting to
		@param mode The installation mode
		@param progress An optional progress reporter
	*/
	public SchemaGenerator(Configuration cfg, Dialect dialect, DbEngine dbEngine, Mode mode, ProgressReporter progress)
	{
		this.cfg = cfg;
		this.dialect = dialect;
		this.dbEngine = dbEngine;
		this.mode = mode;
		this.progress = progress;
	}

	/*
		From the Work interface
		-----------------------
	*/
	@Override
	public void execute(Connection connection) 
		throws SQLException
	{
		log.info("Starting schema generator");
		log.info("Dialect: " + dialect);
		log.info("DbEngine: " + dbEngine);
		log.info("Mode: " + mode);
		
		this.ignoredSql = new ArrayList<String>();
		
		Statement stmt = connection.createStatement();
		try
		{
			List<String> allSql = new ArrayList<String>();
			if (mode == Mode.UPDATE)
			{
				log.info("Fetching database metadata");
				DatabaseMetadata meta = new DatabaseMetadata(connection, dialect);
				log.info("Generating schema update script");
				allSql.addAll(Arrays.asList(cfg.generateSchemaUpdateScript(dialect, meta)));
			}
			else if (mode == Mode.INSTALL)
			{
				log.info("Generating schema creation script");
				allSql.addAll(Arrays.asList(cfg.generateSchemaCreationScript(dialect)));
			}
			else
			{
				throw new UnsupportedOperationException(mode.name());
			}
			
			int numStatements = allSql.size();
			int numDone = 0;
			int interval = Math.max(1, numStatements / 30);
			for (String sql : allSql)
			{
				String actualSql = dbEngine.inspectSchemaGenerationSQL(sql, dialect, mode);
				if (actualSql != null)
				{
					if (!actualSql.equals(sql))
					{
						log.debug("Original SQL: " + sql);
					}
					log.debug("Executing: " + actualSql);
					stmt.executeUpdate(actualSql);
					//System.out.println(actualSql);
					if (progress != null)
					{
						if (numDone % interval == 0)
						{
							progress.display((100 * numDone) / numStatements, "--" + StringUtil.trimString(actualSql, 55));
						}
						else
						{
							progress.append(".");
						}
					}
				}
				else
				{
					ignoredSql.add(sql);
					log.debug("Ignoring: " + sql);
				}
				numDone++;
			}
		}
		catch (SQLException ex)
		{
			log.error("Schema update failed", ex);
			throw ex;
		}
		finally
		{
			if (stmt != null) 
			{
				try
				{
					stmt.close();
				}
				catch (SQLException ex)
				{}
			}
		}
		log.info("Schema update complete" );
		
	}

	/**
		Get a list with the SQL statements that was ignored in the last run.
		@return A list or null if the generator hasn't been executed
		@since 3.1
	*/
	public List<String> getIgnoredSql()
	{
		return ignoredSql;
	}
	
	/**
		The installation mode.
		@since 3.1
	*/
	public static enum Mode
	{
		/**
			Creating a fresh installation.
		*/
		INSTALL(true, false),
		/**
			Updating an existing installation.
		*/
		UPDATE(false, true);
		
		private final boolean requireEmptyDatabase;
		private final boolean verifySchemaVersion;
		private Mode(boolean requireEmptyDatabase, boolean verifySchemaVersion)
		{
			this.requireEmptyDatabase = requireEmptyDatabase;
			this.verifySchemaVersion = verifySchemaVersion;
		}
		
		/**
			Does this installation mode require an empty database or not?
		*/
		public boolean requireEmptyDatabase()
		{
			return requireEmptyDatabase;
		}
		
		/**
			Does this installation mode require a check against the schema version
			of the intalled database or not?
		*/
		public boolean verifySchemaVersion()
		{
			return verifySchemaVersion;
		}
	}
	
}
