/**
	$Id: AnnotationLogger.java 6510 2014-08-11 07:29:17Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log.db;

import org.hibernate.Hibernate;
import org.hibernate.Query;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.LogControl;
import net.sf.basedb.core.PredefinedQuery;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.AnnotationTypeData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.data.LoggableData;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.log.EntityDetails;
import net.sf.basedb.core.log.EntityLogger;
import net.sf.basedb.core.log.LogManager;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.formatter.Formatter;

/**
	Logger for annotations that logs changes as
	updates to the parent item the annotations belong to.
	
	@author Nicklas
	@version 2.13
	@base.modified $Date: 2014-08-11 09:29:17 +0200 (må, 11 aug 2014) $
*/
public class AnnotationLogger
	implements EntityLogger
{

	private final DbLogManagerFactory factory;

	public AnnotationLogger(DbLogManagerFactory factory)
	{
		this.factory = factory;
	}
	
	/*
		From the EntityLogger interface
		-------------------------------
	*/
	@SuppressWarnings("unchecked")
	@Override
	public void logChanges(LogManager logManager, EntityDetails details)
	{
		LogControl logControl = logManager.getLogControl();
		AnnotationData annotation = (AnnotationData)details.getEntity();
		
		// Get the name of the annotation type
		String annotationType = null;
		Type valueType = null;
		boolean disableLogOfValues = false;
		AnnotationTypeData at = annotation.getAnnotationType();
		if (!Hibernate.isInitialized(at))
		{
			Query q = logControl.createHqlQuery(
					PredefinedQuery.getQueryString("DBLOG_GET_ANNOTATIONTYPE_INFO"));
			q.setInteger("annotationTypeId", at.getId());
			Object[] tmp = (Object[])q.uniqueResult();
			annotationType = (String)tmp[0];
			valueType = Type.fromValue((Integer)tmp[1]);
			disableLogOfValues = (Boolean)tmp[2];
		}
		else
		{
			annotationType = at.getName();
			valueType = Type.fromValue(at.getValueType());
			disableLogOfValues = at.getDisableLogOfValues();
		}
		
		// Get the type of the parent item
		AnnotationSetData as = annotation.getAnnotationSet();
		Integer parentType = null;
		if (!Hibernate.isInitialized(as))
		{
			Query q = logControl.createHqlQuery(
				PredefinedQuery.getQueryString("DBLOG_GET_ANNOTATION_ITEMTYPE"));
			q.setInteger("annotationSetId", as.getId());
			parentType = (Integer)q.uniqueResult();			
		}
		else
		{
			parentType = as.getItemType();
		}
		
		// May happen in case the annotation set has been deleted in this transaction
		if (parentType == null) return;
		
		Class parentClass = Item.fromValue(parentType).getDataClass();
		// If the parent item isn't implementing the LoggableData interface
		// we don't log the changes
		if (!LoggableData.class.isAssignableFrom(parentClass)) return;
		
		// Get the id of the parent item
		Query q = logControl.createHqlQuery(
			PredefinedQuery.getQueryString("DBLOG_GET_ITEMID_WITH_ANNOTATION", parentClass.getName()));
		q.setInteger("annotationSetId", as.getId());
		Integer parentId = (Integer)q.uniqueResult();
		
		// May happen in case the parent item has been deleted in this transaction
		if (parentId == null) return;
		
		ChangeType changeType = details.getChangeType();
		if (changeType == ChangeType.CREATE) 
		{
			changeType = ChangeType.ADD;
		}
		else if (changeType == ChangeType.DELETE)
		{
			changeType = ChangeType.REMOVE;
		}
		
		ChangeHistoryDetailData change = new ChangeHistoryDetailData();
		change.setChangeType(changeType.getValue());
		change.setItemId(parentId);
		change.setItemType(parentType);
		change.setChangeInfo("Annotation["+annotationType+"]");
		if (factory.isLoggingOldPropertyValues() && !disableLogOfValues)
		{
			Formatter f = null;
			if (valueType == Type.DATE)
			{
				f = EntityDetails.DATE_FORMAT;
			}
			else if (valueType == Type.TIMESTAMP)
			{
				f = EntityDetails.DATETIME_FORMAT;
			}
			String newValues = Values.getString(annotation.getValues().getValues(), ", ", true, f);
			String oldValues = Values.getString(annotation.getValues().getOldValues(), ", ", true, f);
			change.setNewValue(newValues.length() == 0 ? null : newValues);
			change.setOldValue(oldValues.length() == 0 ? null : oldValues);
		}
		((DbLogManager)logManager).logChangeDetails(change, details);
	}
	// -------------------------------
}
