/**
	$Id: FileSetMemberLogger.java 6510 2014-08-11 07:29:17Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log.db;

import java.util.List;

import org.hibernate.Hibernate;
import org.hibernate.Query;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.LogControl;
import net.sf.basedb.core.PredefinedQuery;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.data.DataFileTypeData;
import net.sf.basedb.core.data.FileSetData;
import net.sf.basedb.core.data.FileSetMemberData;
import net.sf.basedb.core.data.LoggableData;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.log.EntityDetails;
import net.sf.basedb.core.log.EntityLogger;
import net.sf.basedb.core.log.LogManager;

/**
	Logger for file set members that logs changes as
	updates to the parent item the file belong to.
	
	@author Nicklas
	@version 2.13
	@base.modified $Date: 2014-08-11 09:29:17 +0200 (må, 11 aug 2014) $
*/
public class FileSetMemberLogger
	implements EntityLogger
{

	private final DbLogManagerFactory factory;

	public FileSetMemberLogger(DbLogManagerFactory factory)
	{
		this.factory = factory;
	}
	
	/*
		From the EntityLogger interface
		-------------------------------
	*/
	@Override
	public void logChanges(LogManager logManager, EntityDetails details)
	{
		LogControl logControl = logManager.getLogControl();
		FileSetMemberData member = (FileSetMemberData)details.getEntity();
		
		// Get the name of the data file type
		String fileType = null;
		DataFileTypeData ft = member.getDataFileType();
		if (!Hibernate.isInitialized(ft))
		{
			Query q = logControl.createHqlQuery(
					PredefinedQuery.getQueryString("DBLOG_GET_NAME_FOR_ITEM", "DataFileTypeData"));
			q.setInteger("id", ft.getId());
			fileType = (String)q.uniqueResult();
		}
		else
		{
			fileType = ft.getName();
		}
		
		// Get the type of the parent item
		FileSetData fs = member.getFileSet();
		Integer parentType = null;
		if (!Hibernate.isInitialized(fs))
		{
			Query q = logControl.createHqlQuery(
				PredefinedQuery.getQueryString("DBLOG_GET_FILESET_ITEMTYPE"));
			q.setInteger("fileSetId", fs.getId());
			parentType = (Integer)q.uniqueResult();			
		}
		else
		{
			parentType = fs.getItemType();
		}
		
		// May happen in case the file set has been deleted in this transaction
		if (parentType == null) return;
		
		Class parentClass = Item.fromValue(parentType).getDataClass();
		// If the parent item isn't implementing the LoggableData interface
		// we don't log the changes
		if (!LoggableData.class.isAssignableFrom(parentClass)) return;
		
		// Get the id of the parent item
		Query q = logControl.createHqlQuery(
			PredefinedQuery.getQueryString("DBLOG_GET_ITEMID_WITH_FILESET", parentClass.getName()));
		q.setInteger("fileSetId", fs.getId());
		Integer parentId = (Integer)q.uniqueResult();
		
		// May happen in case the parent item has been deleted in this transaction
		if (parentId == null) return;
		
		ChangeType changeType = details.getChangeType();
		DbLogManager dbLogManager = ((DbLogManager)logManager);
		if (changeType == ChangeType.UPDATE)
		{
			// Delegate to EntityDetails
			List<ChangeHistoryDetailData> changes = details.toChangeDetails(Item.fromValue(parentType), parentId, "File["+fileType+"]", changeType, factory.isLoggingDetailedProperties(), factory.isLoggingOldPropertyValues());
			dbLogManager.logChangeDetails(changes, details);
		}
		else
		{
			if (changeType == ChangeType.CREATE) 
			{
				changeType = ChangeType.ADD;
			}
			else if (changeType == ChangeType.DELETE)
			{
				changeType = ChangeType.REMOVE;
			}

			// Create a single entry with the file name
			ChangeHistoryDetailData change = new ChangeHistoryDetailData();
			change.setChangeType(changeType.getValue());
			change.setItemId(parentId);
			change.setItemType(parentType);
			change.setChangeInfo("File[" + fileType+"]");
			if (factory.isLoggingOldPropertyValues())
			{
				String fileName = member.getFile().getName();
				if (changeType == ChangeType.ADD)
				{
					change.setNewValue(fileName);
				}
				else
				{
					change.setOldValue(fileName);
				}
			}
			dbLogManager.logChangeDetails(change, details);
		}
	}
	// -------------------------------
}
