/**
	$Id: ProcessResults.java 6501 2014-06-30 09:13:13Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.manager;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.WeakHashMap;

/**
	Class for collecting results from a scan for new/deleted/updated
	extensions.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2014-06-30 11:13:13 +0200 (må, 30 jun 2014) $
*/
public class ProcessResults
{

	private final long startTime;

	private Map<ExtensionsFile, FileResults> fileResults;
	private Map<ExtensionsFile, FileResults> oldFileResults;
	
	private boolean hasError;
	private int numErrorFiles;
	private long endTime;
	private String summary;

	/**
		Create a new process results object.
	*/
	public ProcessResults()
	{
		this(null);
	}
	
	/**
		Create a new process results object, keeping information about
		previous results.
		@since 3.2.4
	*/
	public ProcessResults(ProcessResults oldResults)
	{
		this.fileResults = new TreeMap<ExtensionsFile, FileResults>();
		this.oldFileResults = new WeakHashMap<ExtensionsFile, FileResults>();
		if (oldResults != null)
		{
			oldFileResults.putAll(oldResults.oldFileResults);
		}
		this.hasError = false;
		this.numErrorFiles = 0;
		this.startTime = System.currentTimeMillis();
	}

	/**
		Get the start time of the scan as a millisecond value.
		@see System#currentTimeMillis()
	*/
	public long getStartTime()
	{
		return startTime;
	}
	
	/**
		Get the end time of the scan as a millisecond value.
		@see System#currentTimeMillis()
	*/
	public long getEndTime()
	{
		return endTime;
	}
	
	/**
		Set the end time of the scan.
	*/
	public void setEnded()
	{
		this.endTime = System.currentTimeMillis();
	}

	
	/**
		Get a summary of the scan results as a textual description.
	*/
	public String getSummary()
	{
		return summary;
	}
	
	/**
		Set a summary of the results.
	*/
	public void setSummary(String summary)
	{
		this.summary = summary;
	}
	
	/**
		Check if the scan was successful or not. For more detailed
		information use {@link #getResults(ExtensionsFile)}.
		@return TRUE if there was any problems, FALSE if everything was 
			successful
	*/
	public boolean hasError()
	{
		return hasError;
	}	

	/**
		The number of extension files that could not be completely loaded
		because of an error.
	*/
	public int getNumErrorFiles()
	{
		return numErrorFiles;
	}

	public List<FileResults> getAllResults()
	{
		return new ArrayList<FileResults>(fileResults.values());
	}
	
	/**
		Get the results for a specific extensions file. If
		no current results exists, old results are searched.
		@param xtFile The extensions file to get the results for
		@return A {@link FileResults} object or null if the
			extensions file is not found
	*/
	public FileResults getResults(ExtensionsFile xtFile)
	{
		FileResults results = fileResults.get(xtFile);
		if (results == null) results = oldFileResults.get(xtFile);
		return results;
	}

	/**
		Forget old processing results for the given file.
		Should be called when an extension has been uninstalled,
	 	@param xtFile The extensions file to forget
	 	@since 3.2.4
	*/
	public void forgetOldResults(ExtensionsFile xtFile)
	{
		oldFileResults.remove(xtFile);
	}
	
	/**
		Sets a short overall status message for an entire file. Example:
		"Installed", "Failed", etc.
		@param xtFile The file to set the status on
		@param status The status message
		@see #addMessage(ExtensionsFile, String)
		@see #addErrorMessage(ExtensionsFile, String)
	*/
	public void setStatus(ExtensionsFile xtFile, String status)
	{
		FileResults m = fileResults.get(xtFile);
		if (m == null) 
		{
			m = new FileResults(xtFile, oldFileResults.get(xtFile));
			fileResults.put(xtFile, m);
			oldFileResults.put(xtFile, m);
		}
		m.status = status;
	}
	
	/**
		Adds a detailed message indicating a successful operation 
		to a file.
		@param xtFile The file to add the message to
		@param message The message
		@see #addErrorMessage(ExtensionsFile, String)
	*/
	public void addMessage(ExtensionsFile xtFile, String message)
	{
		addMessage(xtFile, message, false);
	}

	/**
		Adds a detailed message inidcating a failed operation 
		to a file.
		@param xtFile The file to add the message to
		@param message The message
	*/
	public void addErrorMessage(ExtensionsFile xtFile, String message)
	{
		addMessage(xtFile, message, true);
	}

	private void addMessage(ExtensionsFile xtFile, String message, boolean error)
	{
		FileResults m = fileResults.get(xtFile);
		if (m == null) 
		{
			m = new FileResults(xtFile, oldFileResults.get(xtFile));
			fileResults.put(xtFile, m);
			oldFileResults.put(xtFile, m);
		}
		m.messages.add(message);
		if (error) 
		{
			if (!m.hasError) numErrorFiles++;
			hasError = true;
			m.hasError = true;
			m.status = "Error";
		}
	}
	
	/**
		Scan results for a single XML or JAR file.
	
	*/
	public static class FileResults
	{
		
		private final ExtensionsFile xtFile;
		private final long time;
		List<String> messages;
		boolean hasError = false;
		String status = null;

		FileResults(ExtensionsFile xtFile, FileResults old)
		{
			this.xtFile = xtFile;
			this.time = System.currentTimeMillis();
			this.messages = new LinkedList<String>();
			if (old != null)
			{
				hasError = old.hasError;
				status = old.status;
				messages.addAll(old.messages);
			}
		}
		
		public ExtensionsFile getExtensionsFile()
		{
			return xtFile;
		}
		
		/**
			Get the time of the results as a millisecond value.
			@since 3.2.4
		*/
		public long getTime()
		{
			return time;
		}
		
		/**
			If there was an error in this file or not.
			@return TRUE if there was an error, FALSE if successful
		*/
		public boolean hasError()
		{
			return hasError;
		}
		
		/**
			A short status message indicating the completed
			results on this file.
			@see #getMessages()
		*/
		public String getStatus()
		{
			return status;
		}
		
		/**
			A list of detailed messages, indicating the results of
			individual operation on this file.
		*/
		public List<String> getMessages()
		{
			return messages;
		}
		
	}

}
