/**
	$Id: GtfValidationAction.java 5764 2011-09-27 12:35:28Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.gtf;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidRelationException;
import net.sf.basedb.util.fileset.SingleFileValidationAction;

/**
	Action for validation of GTF files in a file set. The
	GTF file must be connected with an {@link ArrayDesign}.
	<p>
	
	The validation uses the {@link GtfInputStream}. The number of lines
	to parse is configurable. If the entire file is parsed, it will
	set the {@link ArrayDesign#getNumFileFeatures()} value to the number
	of unique transcript_id values found in the file.
	
	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-09-27 14:35:28 +0200 (ti, 27 sep 2011) $
*/
public class GtfValidationAction
	extends SingleFileValidationAction
{

	private final DbControl dc;
	private final ArrayDesign design;
	private final int linesToParse;
	
	/**
		Create a new action. 
		@param dc An open DbControl
		@param design The array design the GTF file is connected with
		@param linesToParse The maximum number of lines to parse from the GTF file.
			Use {@link Integer#MAX_VALUE} to ensure that the entire file is parsed.
			The lowest value is 10
	*/
	public GtfValidationAction(DbControl dc, ArrayDesign design, int linesToParse)
	{
		super(DataFileType.REF_SEQ_GTF);
		this.dc = dc;
		this.design = design;
		this.linesToParse = linesToParse < 10 ? 10 : linesToParse;
	}

	/*
		From the ValidationAction interface
		-----------------------------------
	*/
	@Override
	public void validateAndExtractMetadata()
		throws InvalidDataException, InvalidRelationException
	{
		FileSetMember gtfMember = getAcceptedFile();
		File gtf = gtfMember.getFile();
		
		InputStream in = gtf.getDownloadStream(0);
		String charset = gtf.getCharacterSet();
		if (charset == null) charset = "ISO-8859-1";
		try
		{
			GtfInputStream gtfStream = new GtfInputStream(in, charset, false);
			BufferedReader reader = new BufferedReader(new InputStreamReader(gtfStream, charset));

			// Read 10 lines
			int lineNo = 0;
			while (reader.readLine() != null)
			{
				++lineNo;
				if (lineNo == linesToParse) break;
			}
			
			// If we reached the end, use the real number, 0 otherwise
			design.setNumFileFeatures(lineNo == linesToParse ? 0 : gtfStream.getNumUniqueTranscriptIds());
		}
		catch (IOException ex)
		{
			throw new InvalidDataException(ex);
		}
	}

	@Override
	public void resetMetadata()
	{
		design.setNumFileFeatures(0);
	}
	// -----------------------------------

}
