/**
	$Id: SecondPassSectionSpotsParser.java 5405 2010-09-10 11:09:05Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.io.IOException;
import java.util.List;

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.core.SpotExtraValueBatcher;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.basefile.BaseFileParser;
import net.sf.basedb.util.basefile.BaseFileSectionParser;
import net.sf.basedb.util.importer.spotdata.BaseFileInfo.SpotSectionInfo;
import net.sf.basedb.util.parser.FlatFileParser;

/**
	
	The second pass of the 'spots' section extracts spot and extra value
	data.

	@author Nicklas
	@version 2.14
	@base.modified $Date: 2010-09-10 13:09:05 +0200 (fr, 10 sep 2010) $
*/
public class SecondPassSectionSpotsParser
	implements BaseFileSectionParser
{

	private final DbControl dc;
	private final BaseFileInfo info;
	private final BioAssaySet child;
	private final int linesToParse;
	private final int progressReportInterval;
	private int nextProgressReport;
	private final int channels;
	
	private int sectionCount;
	
	/**
		Creates a new 'section spot' parser for the first pass. This
		parser will only extract header and reporter/position mapping
		information.
		
		@param dc A DbControl to use for database access
		@param info Information about the file we are parsing
		@param child The new child bioassay set that we are importing data to
	*/
	public SecondPassSectionSpotsParser(DbControl dc, BaseFileInfo info, 
		BioAssaySet child, int linesToParse)
	{
		this.dc = dc;
		this.info = info;
		this.child = child;
		this.linesToParse = linesToParse;
		this.progressReportInterval = Math.max(linesToParse / 50, 50);
		this.channels = child.getRawDataType().getChannels();
		this.nextProgressReport = progressReportInterval;
	}
	
	/*
		From the BaseFileSectionParser interface
		----------------------------------------
	*/
	@SuppressWarnings("unchecked")
	@Override
	public void parseSection(BaseFileParser parser, FlatFileParser ffp)
		throws IOException
	{
		++sectionCount;
		
		ffp.parseHeaders();

		// Get information about this spot section that was extracted
		// in the first pass
		SpotSectionInfo ssInfo = info.getSpotSectionInfo(sectionCount);
		int positionIndex = ssInfo.getPositionIndex();
		int numAssayFields = ssInfo.getAssayFields().size();

		// Get the data cube column numbers for each child assay
		List<Integer> assays = ssInfo.getAssays();
		short[] dataCubeColumns = new short[assays.size()];
		for (int i = 0; i < assays.size(); ++i)
		{
			BioAssay child = info.getChildAssay(assays.get(i)).getChild();
			dataCubeColumns[i] = child.getDataCubeColumnNo();
		}
		
		// Prepare the spot intensity parser and the spot batcher
		SpotIntensityParser spiParser = ssInfo.getSpotIntensityParser();
		spiParser.beginSection(parser, ffp, ssInfo.getAssayFields());
		SpotBatcher spotBatcher = child.getSpotBatcher();
		float[] intensities = new float[channels];
		
		// Prepare extra value parsers
		ExtraFloatParser[] extraParsers = null;
		SpotExtraValueBatcher<Float>[] extraBatchers = null;
		if (ssInfo.getExtraFloatParsers() != null)
		{
			List<ExtraFloatParser> efp = ssInfo.getExtraFloatParsers();
			extraParsers = new ExtraFloatParser[efp.size()];
			extraBatchers = new SpotExtraValueBatcher[efp.size()];
			Job job = child.getTransformation().getJob();
			for (int i = 0; i < efp.size(); ++i)
			{
				ExtraFloatParser p = efp.get(i);
				extraParsers[i] = p;
				ExtraValueType type = info.getExtraFloat(p.getId());
				extraBatchers[i] = child.getSpotExtraValueBatcher(Float.class, type, job);
			}
		}
		boolean hasExtraValues = extraParsers != null;
		
		// Parse data and check if each position has same reporter as before
		FlatFileParser.Data data = null;
		while ((data = ffp.nextData()) != null)
		{
			ThreadSignalHandler.checkInterrupted();
			int parsedLines = ffp.getParsedLines();
			if (parsedLines >= nextProgressReport)
			{
				nextProgressReport = parsedLines + progressReportInterval;
				parser.setProgress(parsedLines, "Importing spot data (second pass): " + 
						parsedLines + " of " + linesToParse + " lines");
			}

			int offset = ssInfo.getFirstAssayFieldIndex();
			int position = Values.getInt(data.get(positionIndex), 0);
			
			for (short dataCubeColumn : dataCubeColumns)
			{
				if (spiParser.setIntensities(data, intensities, offset))
				{
					spotBatcher.insert(dataCubeColumn, position, intensities);
					if (hasExtraValues)
					{
						for (int i = 0; i < extraParsers.length; ++i)
						{
							float value = extraParsers[i].getValue(data, offset);
							if (!Float.isNaN(value))
							{
								extraBatchers[i].insert(dataCubeColumn, position, value);
							}
						}
					}
				}
				offset += numAssayFields;
			}
		}
	}
	// ----------------------------------------------------
	

}
