/*
	$Id: RawChannelFunction.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.util.BioAssaySetUtil;

import java.sql.SQLException;
import java.util.Stack;

import org.nfunk.jep.Node;
import org.nfunk.jep.ParseException;

/**
	A JEP function class that adds a <code>rawCh(int)</code> function to a 
	JEP expression parser. The function will look up the (possible transformed) 
	intensity value for the given channel. For example: <code>rawCh(1)</code>
	<p>
	Note! This function always return the values as they are stored in 
	the database. Use {@link ChannelFunction} if you need the untransformed 
	intensity values.	
	<p>
	To be able to use this function it must be registered with the JEP
	parser and, before the expression is evaluated, a {@link SqlResult} object must be set.

	@author Nicklas
	@version 2.12
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
	@see Jep
	@see BioAssaySetUtil#createJepExpression(DbControl, String, DynamicQuery)
*/
public class RawChannelFunction
	implements JepExpressionFunction
{

	private int[] channelToIndex;
	private SqlResult result;
	private int numParameters;

	/**
		Create a new instance of this function. The new instance cannot be used
		to dynamically evaluate expressions. It should only be used for converting
		JEP formulas to {@link Expression}:s.
		@see Jep#formulaToExpression(String, JepFunction[])	
	*/
	public RawChannelFunction()
	{}
	
	/**
		Create a new instance of this function which can be used
		to dynamically evaluate expressions.
		@param channelToIndex An array that maps channel numbers to column indexes in
			the SqlResult, array position 0 maps the index of channel number 1, etc.
		@see #setSqlResult(SqlResult)
	*/
	public RawChannelFunction(int[] channelToIndex)
	{
		this.channelToIndex = channelToIndex;
	}
	
	/*
		From the JepFunction interface
		-------------------------------------------
	*/
	/**
		@return The string "rawCh"
	*/
	@Override
	public String getFunctionName()
	{
		return "rawCh";
	}
	// -------------------------------------------
	/*
		From the JepExpressionFunction interface
		-------------------------------------------
	*/
	/**
		Use the {@link Dynamic#column(VirtualColumn)} method to create an
		expression referencing a channel intensity.
	*/
	@Override
	public Expression toExpression(Node node)
	{
		int numChildren = node.jjtGetNumChildren();
		if (numChildren != 1)
		{
			throw new BaseException("Invalid number of expressions for 'ch' function: " + numChildren);
		}
		return Dynamic.column(VirtualColumn.channelRaw(Jep.nodeToInt(node.jjtGetChild(0))));
	}
	// -------------------------------------------
	/*
		From the PostfixMathCommandI interface
		-------------------------------------------
	*/
	/**
		@return Always 1
	*/
	@Override
	public int getNumberOfParameters()
	{
		return 1;
	}
	@Override
	public void setCurNumberOfParameters(int n)
	{
		this.numParameters = n;
	}
	@Override
	public boolean checkNumberOfParameters(int n)
	{
		return n == 1;
	}
	@Override
	@SuppressWarnings({"unchecked"})
	public void run(Stack stack)
		throws ParseException
	{
		if (stack == null || stack.empty()) 
		{
			throw new ParseException("Stack is empty");
		}
		Object channel = stack.pop();
		if (channel instanceof Number)
		{
			stack.push(channel(((Number)channel).intValue()));
		}
		else
		{
			throw new ParseException("Invalid parameter type: " + channel + "; expected integer");
		}
	}
	// -------------------------------------------

	/**
		Set a new {@link SqlResult} object that will be used the next time the
		JEP expression is evaluated.
		@param result The result object
	*/
	public void setSqlResult(SqlResult result)
	{
		this.result = result;
	}
	/**
		Get the value of the specified channel of the current sql result.
		@param channel The channel number
	 	@throws ParseException Either if it's an invalid channel number or 
	 		if no result object has been specified.
	*/
	public Object channel(int channel)
		throws ParseException
	{
		if (result == null)
		{
			throw new ParseException("No result object has been specified for function ch("+channel+")");
		}
		if (channelToIndex == null || channel < 1 || channel > channelToIndex.length)
		{
			throw new ParseException("Invalid channel number for function ch("+channel+")");
		}
		try
		{
			return result.getObject(channelToIndex[channel-1]);
		}
		catch (SQLException ex)
		{
			throw new ParseException(ex.getMessage());
		}
	}
}
