/**
	$Id: AbstractNodeLoader.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.Node;

/**
	Abstract node loader implementation that does nothing/throws an
	exception. It is recommended that all {@link NodeLoader} implementations 
	extend this class (or a subclass) and only override the methods
	they need to support. The node loader interface may be extended in the
	future with new methods. Default implementations will be added to this class
	which ensures that subclass implementation will continue to work.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (on, 09 apr 2014) $
*/
public abstract class AbstractNodeLoader<I>
	implements NodeLoader<I>
{
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(AbstractNodeLoader.class);
	
	private static final boolean debug = log.isDebugEnabled();

	protected AbstractNodeLoader()
	{}

	/*
		From the NodeLoader interface
		-----------------------------
	*/
	/**
		@throws UnsupportedOperationException
	*/
	@Override
	public Node createRootNode(DbControl dc, OverviewContext context, I rootItem)
	{
		throw new UnsupportedOperationException();
	}
	/**
		@throws UnsupportedOperationException
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		throw new UnsupportedOperationException();
	}

	/**
		@throws UnsupportedOperationException
	*/
	@Override
	public Node createPropertyNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		throw new UnsupportedOperationException();
	}

	/**
		@throws UnsupportedOperationException
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node childNode)
	{
		throw new UnsupportedOperationException();
	}
	
	/**
		Load childs nodes. This method simplifies the loading of child nodes.
		First, it will check if child nodes has already been loaded. If so it 
		simply returns. Then depending on the node's {@link ChildNodeDirection}
		it will call {@link #loadForwardChildNodes(DbControl, OverviewContext, Node)},
		{@link #loadReverseChildNodes(DbControl, OverviewContext, Node)} and 
		{@link #loadPropertyChildNodes(DbControl, OverviewContext, Node)} as needed.
		<p>
		NOTE! If the given node is a folder-node, this method will iterate
		over the children in the folder and do the above for each one of them.
	*/
	@Override
	public boolean loadChildNodes(DbControl dc, OverviewContext context, Node node)
	{
		if (debug) log.debug("Loading child nodes: " + node);
		if (node.getNodeType() == Node.Type.FOLDER)
		{
			return loadChildNodesOfFolderNode(dc, context, node);
		}
		else 
		{
			return loadChildNodesOfItemNode(dc, context, node);
		}
	}
	// ----------------------------------------------------

	/**
		Loads child nodes of an item-type node.
	*/
	protected boolean loadChildNodesOfItemNode(DbControl dc, OverviewContext context, Node node)
	{
		if (node.isChildrenLoaded()) return false;
		node.setChildrenLoaded();
		ChildNodeDirection direction = node.getChildNodeDirection();
		if (direction.loadReverse()) loadReverseChildNodes(dc, context, node);
		if (direction.loadForward()) loadForwardChildNodes(dc, context, node);
		if (direction.loadProperty()) loadPropertyChildNodes(dc, context, node);
		return node.numChildren() > 0;
	}
	
	/**
		Iterates the children of a folder-type node and loads the children for
		each one of them.
	*/
	protected boolean loadChildNodesOfFolderNode(DbControl dc, OverviewContext context, Node node)
	{
		boolean result = false;
		if (node.getChildNodeDirection() != ChildNodeDirection.NONE && node.numChildren() > 0)
		{
			for (Node child : node.getChildren())
			{
				result |= loadChildNodes(dc, context, child);
			}
		}
		return result;
	}
	
	/**
		Load forward-loading child nodes. This method is called from {@link 
		#loadChildNodesOfItemNode(DbControl, OverviewContext, Node)}
		if the current node has a {@link ChildNodeDirection#loadForward()} that returns true.
		The default implementation does nothing. This method is intended to be implemented
		by subclasses.
	*/
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node node)
	{}
	
	/**
		Load reverse-loading child nodes. This method is called from {@link 
		#loadChildNodesOfItemNode(DbControl, OverviewContext, Node)}
		if the current node has a {@link ChildNodeDirection#loadReverse()} that returns true.
		The default implementation does nothing. This method is intended to be implemented
		by subclasses.
	*/
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node node)
	{}
	
	/**
		Load property-type child nodes. This method is called from {@link 
		#loadChildNodesOfItemNode(DbControl, OverviewContext, Node)} if the current node
		has a {@link ChildNodeDirection#loadProperty()} that returns true.
		The default implementation does nothing. This method is intended to be implemented
		by subclasses.
	*/
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node node)
	{}

}
