/**
	$Id: PhysicalBioAssayLoader.java 6090 2012-08-22 11:36:15Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import java.util.List;

import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.OverviewUtil;
import net.sf.basedb.util.overview.cache.IndexedCacheKey;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for physical bioassays. There are two forward-loading
	directions:
	<ol>
	<li>{@link Extract}:s -&gt; {@link PhysicalBioAssay}:s -&gt; 
	{@link DerivedBioAssay}:s. 
	<li>{@link ArraySlide}:s -&gt; {@link PhysicalBioAssay}:s -&gt; 
	{@link DerivedBioAssay}:s
	</ol>
	
	The reverse-loading direction starts with a derived bioassay set and moves to both
	extracts and array slide.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2012-08-22 13:36:15 +0200 (on, 22 aug 2012) $
*/
public class PhysicalBioAssayLoader
	extends BasicItemNodeLoader<PhysicalBioAssay>
{
	public PhysicalBioAssayLoader()
	{
		super(Item.PHYSICALBIOASSAY, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<PhysicalBioAssay>("bioassay", "Physical bioassay"));
	}

	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading physical bioassay nodes from either an extract or
		an array slide node. The returned node is either folder-type node containing 
		nodes for each bioassay that has been created from the extract
		or an item node with the bioassay on the array slide.
		@return The folder node with the bioassays, or the item node (can be null)
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node returnNode = null;
		Item parentType = parentNode.getItemType();
		if (parentType == Item.EXTRACT)
		{
			returnNode = createForwardNode((Extract)parentNode.getItem(dc), dc, context, parentNode, false);
		}
		else if (parentType == Item.ARRAYSLIDE)
		{
			returnNode = createForwardNode((ArraySlide)parentNode.getItem(dc), dc, context, parentNode);
		}
		return returnNode;
	}
	
	/**
		Create a reverse-loading bioassay node from the derived bioassay node.
		@return A bioassay node
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node derivedBioAssayNode)
	{
		NodeFactory<PhysicalBioAssay> nf = getNodeFactory(dc, context);
		DerivedBioAssay dba = (DerivedBioAssay)derivedBioAssayNode.getItem(dc);
		
		ItemQuery<PhysicalBioAssay> query = context.initQuery(dba.getPhysicalBioAssays(), "name");
		List<PhysicalBioAssay> parents = query.list(dc);

		Node folderNode = null;
		Node bioAssayNode = null;
		if (parents.size() > 1)
		{
			folderNode = new Node("parents", "Physical bioassays", derivedBioAssayNode, ChildNodeDirection.REVERSE);
		}

		for (PhysicalBioAssay pba : parents)
		{
			bioAssayNode = createItemNode(nf, pba, null, false, folderNode == null ? derivedBioAssayNode : folderNode, ChildNodeDirection.REVERSE);
		}
		postValidateFolder(nf, folderNode, derivedBioAssayNode, bioAssayNode == null);
		return folderNode == null ? bioAssayNode : folderNode;
	}
	
	/**
		A property node is created when moving in the reverse direction so that we
		can load child physical bioassays with the {@link ItemSubtype#getPushAnnotations()} 
		flag set. The parent node should be an extract node.
		@since 3.1
	*/
	@Override
	public Node createPropertyNode(DbControl dc, OverviewContext context, Node parentNode) 
	{
		Node returnNode = null;
		Item parentType = parentNode.getItemType();
		if (parentType == Item.EXTRACT)
		{
			returnNode = createForwardNode((Extract)parentNode.getItem(dc), dc, context, parentNode, true);
		}
		return returnNode;
	}
	// --------------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads property nodes of a physical bioassay. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Hardware: {@link HardwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node bioAssayNode)
	{
		getNodeLoader(context, Item.ANNOTATIONTYPE).createPropertyNode(dc, context, bioAssayNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, bioAssayNode);
		getNodeLoader(context, Item.HARDWARE).createPropertyNode(dc, context, bioAssayNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, bioAssayNode);
	}
	
	/**
		Load forward-loading child nodes:
		<ul>
		<li>Derived bioassay sets (root): {@link DerivedBioAssayLoader#createForwardNode(DbControl, OverviewContext, Node)}
		<li>Array slide in reverse direction (if the parent node is an extract):
			{@link ArraySlideLoader#createReverseNode(DbControl, OverviewContext, Node)}
		<li>Extracts in reverse direction (if the parent node is the array slide):
			{@link ExtractLoader#createReverseNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node bioAssayNode)
	{
		getNodeLoader(context, Item.DERIVEDBIOASSAY).createForwardNode(dc, context, bioAssayNode);
		Node parent = bioAssayNode.getParent();
		if (parent != null && parent.getNodeType() == Node.Type.FOLDER) 
		{
			parent = parent.getParent();
		}
		Item parentItemType = parent == null ? null : parent.getItemType();
		if (parentItemType == Item.EXTRACT)
		{
			getNodeLoader(context, Item.ARRAYSLIDE).createReverseNode(dc, context, bioAssayNode);
		}
		if (parentItemType == Item.ARRAYSLIDE)
		{
			getNodeLoader(context, Item.EXTRACT).createReverseNode(dc, context, bioAssayNode);
		}
	}
	
	/**
		Load reverse child nodes:
		<ul>
		<li>Extracts: {@link ExtractLoader#createReverseNode(DbControl, OverviewContext, Node)}
		<li>Array slide: {@link ArraySlideLoader#createReverseNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node bioAssayNode)
	{
		getNodeLoader(context, Item.EXTRACT).createReverseNode(dc, context, bioAssayNode);
		getNodeLoader(context, Item.ARRAYSLIDE).createReverseNode(dc, context, bioAssayNode);
	}
	// --------------------------------------
	
	/**
		Create forward-loading bioassay nodes for the bioassays 
		that has a given extract as a source.
	*/
	private Node createForwardNode(Extract extract, DbControl dc, OverviewContext context, Node parentNode,
		boolean onlyPushAnnotationTypes)
	{
		NodeFactory<PhysicalBioAssay> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<PhysicalBioAssay> query = context.initQuery(extract.getPhysicalBioAssays(), "name");
		if (onlyPushAnnotationTypes) OverviewUtil.restrictToPushAnnotationSubtypes(query);
		ItemResultIterator<PhysicalBioAssay> it = query.iterate(dc);
		while (it.hasNext())
		{
			PhysicalBioAssay bioAssay = it.next();
			Object cacheKey = new IndexedCacheKey(bioAssay, extract);
			if (folderNode == null)
			{
				folderNode = new Node("bioassays", "Physical bioassays", parentNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, bioAssay, cacheKey, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, parentNode, false);
		return folderNode;
	}

	/**
		Create a forward-loading bioassay node from an array slide.
	*/
	private Node createForwardNode(ArraySlide slide, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<PhysicalBioAssay> nf = getNodeFactory(dc, context);
		PhysicalBioAssay bioAssay = null;
		boolean denied = false;
		try
		{
			bioAssay = slide.getPhysicalBioAssay();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Node hybNode = createItemNode(nf, bioAssay, bioAssay, denied, parentNode, ChildNodeDirection.FORWARD);
		return hybNode;
	}

}
