/*
$Id: HelpImporter.java 6127 2012-09-14 07:28:35Z nicklas $

Copyright (C) 2006 Nicklas Nordborg, Martin Svensson
Copyright (C) 2007 Nicklas Nordborg

This file is part of BASE - BioArray Software Environment.
Available at http://base.thep.lu.se/

BASE is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

BASE is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.Client;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.Help;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.plugin.AbstractPlugin;
import net.sf.basedb.core.plugin.AutoDetectingImporter;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.ImportInputStream;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.XMLUtil;

import org.jdom.Document;
import org.jdom.Element;

/**
	This plugin imports helptexts into database from a XML-file.
	The XML-file need to have a DOCLET definition in the first 15 lines
	and it should use the DTD-file: helpfile.dtd in /net/sf/basedb/core/dtd/.
	  
	@author Martin
	@version 2.0
	@base.modified $Date$
*/

public class HelpImporter	
	extends AbstractPlugin
	implements InteractivePlugin, AutoDetectingImporter, SignalTarget
{
	private int numImported = 0;
	private int numUpdated = 0;
	private int numTotal = 0;
	
	private boolean overwrite;
	private Client client;
	private File xmlFile;
	private String filename;
	
	private PluginParameter<File> fileParameter;
	
	private PluginParameter<Client> clientParameter;
	
	private PluginParameter<Boolean> overWriteParameter;
	
	private RequestInformation configureJob;
		
	private static final String dtdFileName = "helpfile.dtd";
	
	private static final GuiContext clientItem = new GuiContext(Item.CLIENT, GuiContext.Type.ITEM);
	private static final GuiContext helpList = new GuiContext(Item.HELP, GuiContext.Type.LIST);
	
	private static final Set<GuiContext> guiContexts =  
		new HashSet<GuiContext>(Arrays.asList(clientItem, helpList));

	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	private ThreadSignalHandler signalHandler;

	/*
	 	From the InteractivePlugin interface
	 	----------------------------------------------------
	*/	
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	
	@Override
	public String isInContext(GuiContext context, Object item)
	{		
		String message = null;
		if (context.equals(clientItem))
		{
			if (item == null)
			{
				message = "The item is null";
			}
			else if (!(item instanceof Client))
			{
				message = "The object is not a Client: " + item;
			}
		}
		return message;
	}
	
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestinformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestinformation = getConfiguredJobParameters();
			ItemContext fileContext = sc.getCurrentContext(Item.FILE);
			fileContext.removePropertyFilter("fileType"); // No existing filetype
		}		
		return requestinformation;
	}
	
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			List<Throwable> errors = validateRequestParameters (
					getConfiguredJobParameters().getParameters(), 
					request
			); 
			
			if (errors != null)
			{
				response.setError(errors.size() + " invalid parameters were found in the request",errors);
				return;
			}
			
			storeValue(job, request, fileParameter);
			storeValue(job, request, clientParameter);
			storeValue(job, request, overWriteParameter);
			
			File file = (File)job.getValue("file");
			Client client = (Client)job.getValue("client");
			response.setSuggestedJobName("Import help texts to client application '" + client.getName() +
					"' from file '" + file.getName() + "'");
			
			response.setDone("The job configuration is complete");			
		}
	}
	
	/*
	 	From the Plugin interface
	 	-----------------------------------
	*/
	@Override
	public MainType getMainType()
	{
		return Plugin.MainType.IMPORT;
	}
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request write access to Client:s and read access to
	 	File:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.CLIENT, null, EnumSet.of(Permission.WRITE)));
		}
		return permissions;
	}
	@Override
	public void run(Request request, Response response, ProgressReporter progress)
	{		
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_EXECUTE))
		{
			InputStream is = null;			
			try
			{
				xmlFile = (File)job.getValue("file");	
				overwrite = (Boolean)job.getValue("overwrite");			
				client = (Client)job.getValue("client");	
				is = xmlFile.getDownloadStream(0);
				boolean importable = isImportable(is);		
				is.close();
				if (!importable)
				{
					throw new InvalidDataException("File "+ xmlFile + " cannot be imported by this plugin");
				}
				is = xmlFile.getDownloadStream(0);
				doImport(is, progress);
				is.close();
				response.setDone(getSuccessMessage());
			}
			catch (Throwable ex)
			{
				if (is != null)					
				{
					try
					{
						is.close();
					}
					catch(IOException ioex)
					{}				
				}		
				response.setError(ex.getMessage(), Arrays.asList(ex));
			}			
		}
		else
		{
			response.setError("Unknown command: "+command, null);
		}
	}
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------
		
	/*
	 	From the AutoDetectingImporter interface
	 	-------------------------------------------
	*/
	@Override
	public boolean isImportable(InputStream in) 
		throws BaseException
	{
		boolean importable = false;		
		String charset = null;
		if (in instanceof ImportInputStream)
		{
			charset = ((ImportInputStream)in).getCharacterSet();
		}
		if (charset == null) charset = "UTF-8";
		BufferedReader reader = new BufferedReader(
				new InputStreamReader(in, Charset.forName(charset)));
		try
		{
			String headerLine = reader.readLine();			
			if (headerLine != null && headerLine.startsWith("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"))
			{
				//Look for the <!DOCTYPE definition in the first 30 lines 
				int numHeaderLines = 0;
				do
				{
					headerLine = reader.readLine();
					numHeaderLines++;
					if (headerLine != null && headerLine.contains("<!DOCTYPE") )
					{
						StringBuilder docTypeTag = new StringBuilder(headerLine);
						while(headerLine != null && !headerLine.contains(">") && !(numHeaderLines > 30) )
						{
							headerLine = reader.readLine();
							numHeaderLines++;
							docTypeTag.append(headerLine);
						}						
						importable = docTypeTag.toString().contains(dtdFileName);
					}					
				}
				while( !importable && !(numHeaderLines > 30 ) );
			}
		}
		catch (IOException ioex)
		{
			throw new BaseException(ioex);
		}		
		return importable;
	}

	@Override
	@SuppressWarnings("unchecked")
	public void doImport(InputStream in, ProgressReporter progress)
		throws BaseException
	{	
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();				 
			client = Client.getById(dc, client.getId());
			float progLastUpdated = 0;
			int updateInterval = 5;
		
			URL dtdURL = HelpImporter.class.getResource("/net/sf/basedb/core/dtd/"+dtdFileName);
			if (xmlFile != null)
			{
				File f = File.getById(dc, xmlFile.getId());
				filename = f.getPath().toString();
			}
			Document doc = XMLUtil.getValidatedXML(in, dtdURL, filename);			
			Element rootElement = doc.getRootElement();
			String clientExternalID = rootElement.getAttributeValue("clientExternalID");
			
			if(client.getExternalId().equals(clientExternalID))
			{
				List<Element> helpElements = (List<Element>)rootElement.getChildren();	
				float percentPerItem = 100.0f / helpElements.size();
				
				for (int i = 0; i < helpElements.size(); i++)
				{
					ThreadSignalHandler.checkInterrupted();
					Element helpElement = helpElements.get(i);
					String name = helpElement.getChildText("name");
					String desc = helpElement.getChildText("description");
					String externalHelpId = helpElement.getAttributeValue("externalID");
					Help newHelp = client.getHelpByExternalId(dc, externalHelpId, true);
					
					if(!newHelp.isInDatabase())
					{	
						dc.saveItem(newHelp);
						newHelp.setName(name);
						newHelp.setDescription(desc);
						numImported++;
					}	
					else if(newHelp.isInDatabase() && overwrite)
					{
						newHelp.setName(name);
						newHelp.setDescription(desc);
						numUpdated++;
					}	
					if ((progress != null) && (((percentPerItem*i)-progLastUpdated) > updateInterval))
					{						
						progress.display((int)(percentPerItem*i), i+" of "+helpElements.size()+" helptexts are imported");
						progLastUpdated = percentPerItem*i;
					}
					numTotal++;
				}
				dc.commit();				
			}			
			else
			{
				throw new InvalidDataException("The help texts are not for this client, " + client.getExternalId());
			}
		}
		catch (Throwable ex)
		{			
			throw new BaseException(ex);
		}
		finally
		{
			if(dc != null)dc.close();
		}
	}
	/*
	 ------------------------------------------------------------------
	*/
	
	private RequestInformation getConfiguredJobParameters()
	{
		if (configureJob == null)
		{
			fileParameter = new PluginParameter<File>
			(
				"file",
				"File",
				"The xml file to import helptexts from",
				new FileParameterType(null, true, 1)
			);
			
			clientParameter = new PluginParameter<Client>
			(
					"client",
					"Client", 
					"The client the help texts belongs to",
					new ItemParameterType<Client>(Client.class, null, true, 1, null)
			);
			
			overWriteParameter = new PluginParameter<Boolean>
			(
					"overwrite",
					"Overwrite texts in DB",
					"Overwrite existing texts in database?",
					new BooleanParameterType(false, true)
			);
			
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			
			parameters.add(fileParameter);
			parameters.add(clientParameter);
			parameters.add(overWriteParameter);
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Help import options",
				"Set the options to run this job",
				parameters
			);
		}
		return configureJob;
	}
	
	/**
	 	This method installs the helptexts from an inputstream.
	
		@param client The client that owns the helptexts
		@param overwrite Handles duplicate of externalID. 
				True - the existing texts in database will be overwritten 
						with the one from importfile that have same externalID.
				False - the existing texts in database are not overwritten 
						with the one from importfile that have the same externalID.
		@param is InputStream from the xml-file with texts to import
		@param filename The filename of the original file or null if not known
		@since 2.4
	*/
	public void installHelpTexts(Client client, boolean overwrite, InputStream is, String filename)
	{
		if (client == null)throw new InvalidUseOfNullException("Client");
		this.client = client;
		this.overwrite = overwrite;
		this.filename = filename;
		doImport(is, null);
	}
	
	private String getSuccessMessage()
	{
		StringBuilder message = new StringBuilder();
		if (numImported > 0) message.append(numImported).append(" help texts created; ");
		if (numUpdated > 0) message.append(numUpdated).append(" help texts updated; ");
		int numIgnored = numTotal - numUpdated - numImported;
		if (numIgnored > 0) message.append(numIgnored).append(" help texts ignored because the already existed in the database; ");
		return message.toString();
	}
}
