/**
	$Id: GtfReporterMapImporter.java 5771 2011-09-29 11:48:47Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.gtf;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.FeatureIdentificationMethod;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.plugins.ReporterMapFlatFileImporter;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.InputStreamTracker;
import net.sf.basedb.util.gtf.GtfInputStream;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.WrappedConfigureByExample;

/**
	Import features to an array design from a GTF file. The default
	settings use transcript_id+seqname as the feature id and reporter id. This
	can be changed by user configuration, but it is recommended that
	transcript_id+seqname is used as feature id since other combinations of
	values may not be unique in raw data files.
	
	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-09-29 13:48:47 +0200 (to, 29 sep 2011) $
*/
public class GtfReporterMapImporter
	extends ReporterMapFlatFileImporter
	implements InteractivePlugin, WrappedConfigureByExample
{

	
	/**
		Needed for progress reporting since reading from the source
		will produce less bytes when wrapped.
	*/
	private InputStreamTracker tracker;
	
	private RequestInformation configurePlugin;
	private RequestInformation configureJob;
	private List<PluginParameter<String>> allColumnMappings;

	protected static final PluginParameter<String> requiredFeatureIdColumnMapping = new PluginParameter<String>(
			"featureIdColumnMapping",
			"Feature ID",
			"Mapping that picks the feature's ID from the data columns. " +
			"For example: \\<transcript_id>\\@\\<seqname>\\",
			new StringParameterType(255, null, true)
			);

	public GtfReporterMapImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public void init(SessionControl sc, ParameterValues configuration,
			ParameterValues job) throws BaseException
	{
		// Wrapper that provides default values for a configuration
		configuration = new DefaultConfigurationValues(configuration);
		super.init(sc, configuration, job);
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = super.isInContext(context, item);
		if (message == null)
		{
			// We are ok so far, also check that the platform is the 
			ArrayDesign design = (ArrayDesign)item;
			if (!design.isPlatform(Platform.SEQUENCING))
			{
				message = "The array design is not a sequence reference design.";
			}
		}
		return message;
	}


	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigurePluginParameters(context);
		}
		else 
		{
			requestInformation = super.getRequestInformation(context, command);
		}
		return requestInformation;
	}

	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
			{
				RequestInformation ri = getConfigurePluginParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				
				// Parser settings
				storeValue(configuration, request, headerRegexpParameter);
				storeValue(configuration, request, dataHeaderRegexpParameter);
				storeValue(configuration, request, dataSplitterRegexpParameter);
				storeValue(configuration, request, trimQuotesParameter);
				storeValue(configuration, request, ignoreRegexpParameter);
				storeValue(configuration, request, dataFooterRegexpParameter);
				storeValue(configuration, request, minDataColumnsParameter);
				storeValue(configuration, request, maxDataColumnsParameter);
				storeValue(configuration, request, ri.getParameter(Parameters.CHARSET_PARAMETER));

				// Column mappings
				storeValue(configuration, featureIdentificationParameter, FeatureIdentificationMethod.FEATURE_ID.name());
				storeValue(configuration, request, complexMappings);
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter<String> pp : getAllColumnMappings())
				{
					String mapExpression = (String)request.getParameterValue(pp.getName());
					checkColumnMapping(mapExpression, allowComplex, pp.getLabel());
					storeValue(configuration, pp, mapExpression);
				}

				response.setDone("Plugin configuration complete");
			}
			else
			{
				super.configure(context, request, response);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
		
	}
	// ---------------------------------------

	/*
		From the WrappedConfigureByExample interface
		-------------------------------------------
	*/
	/**
		Wrap the input stream with a {@link GtfInputStream}.
	*/
	@Override
	public InputStream wrapInputStream(InputStream in)
		throws IOException
	{
		tracker = new InputStreamTracker(in);
		return new GtfInputStream(tracker, getCharset(), true);
	}
	// -------------------------------------------

	/*
		From the AbstractFlatFileImporter class
		---------------------------------------
	*/
	/**
		Get the number of bytes read from the original file.
	*/
	@Override
	protected long getNumBytes(FlatFileParser ffp)
	{
		return tracker.getNumRead();
	}
	// ----------------------------------------

	private List<PluginParameter<String>> getAllColumnMappings()
	{
		if (allColumnMappings == null)
		{
			allColumnMappings = new ArrayList<PluginParameter<String>>();
			allColumnMappings.add(cloneParameterWithDefaultValue(reporterIdColumnMapping));
			allColumnMappings.add(cloneParameterWithDefaultValue(requiredFeatureIdColumnMapping));
		}
		return allColumnMappings;
	}

	
	private RequestInformation getConfigurePluginParameters(GuiContext context)
	{
		if (configurePlugin == null)
		{
			// Parameters object for CONFIGURE_PLUGIN
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			// Parser regular expressions
			parameters.add(parserSection);
			parameters.add(cloneParameterWithDefaultValue(headerRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataHeaderRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataSplitterRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(trimQuotesParameter));
			parameters.add(cloneParameterWithDefaultValue(ignoreRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(dataFooterRegexpParameter));
			parameters.add(cloneParameterWithDefaultValue(minDataColumnsParameter));
			parameters.add(cloneParameterWithDefaultValue(maxDataColumnsParameter));
			parameters.add(Parameters.charsetParameter(null, null, 
					(String)getJobOrConfigurationValue(Parameters.CHARSET_PARAMETER)));
			
			// Column mappings
			parameters.add(mappingSection);
			parameters.add(cloneParameterWithDefaultValue(complexMappings));
			parameters.addAll(getAllColumnMappings());

			configurePlugin = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_PLUGIN,
				"Parser settings",
				"Enter the regular expressions used to parse the text file " +
				"and the column mappings used to find matching properties for the features.",
				parameters
			);
		}
		return configurePlugin;
	}

}
