/* $Id$
	------------------------------------------------------------------
	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	JavaScript functions for working with subtypable items

	@author Nicklas
*/

var ItemSubtype = function()
{
	var subtype = {};
	var cache = [];
	
	/**
		Get the subtype of the given item. Uses an AJAX request
		and return the subtype as an object with 'id' and 'name'
		properties.
		
		@param The type of the item 
		@param The id of the item
	*/
	subtype.getSubtype = function(itemType, itemId)
	{
		var cacheKey = 'subtype.'+itemType+'.'+itemId;
		var jsonSubtype = cache[cacheKey];
		if (!jsonSubtype)
		{
			var request = Ajax.getXmlHttpRequest();
			var url = App.getRoot() + 'admin/itemsubtypes/ajax.jsp?ID=' + App.getSessionId();
			url += '&cmd=GetSubtypeForItem&item_id=' + itemId;
			url += '&itemType='+itemType;
			request.open("GET", url, false); 
			request.send(null);
			jsonSubtype = JSON.parse(request.responseText);
			cache[cacheKey] = jsonSubtype;
		}
		return jsonSubtype && jsonSubtype.id ? jsonSubtype : null;
	}
	
	
	/**
		Get the subtype that is related to the given one.
		@param subtypeId The main subtype
		@param relatedItemType The item type of the related subtype we are
			looking for
		@param defaultSubtype The id of the default subtype to return if
			no related subtype is found (optional)
		@return An object with the 'id' and 'name' of the related subtype or null
			if no related subtype is found
	*/
	subtype.getRelatedSubtype = function(subtypeId, relatedItemType, defaultSubtypeId)
	{
		if (!subtypeId && !defaultSubtypeId) 
		{
			return null;
		}
		
		var cacheKey = 'relatedsubtype.'+relatedItemType+'.'+subtypeId;
		var jsonSubtype = cache[cacheKey];
		if (!jsonSubtype)
		{
			var request = Ajax.getXmlHttpRequest();
			var url = App.getRoot() + 'admin/itemsubtypes/ajax.jsp?ID=' + App.getSessionId();
			url += "&cmd=GetRelatedSubtype&relatedType=" + relatedItemType;
			if (subtypeId) 
			{
				url += '&item_id='+subtypeId;
			}
			if (defaultSubtypeId) 
			{
				url += '&defaultRelatedId=' + defaultSubtypeId;
			}
			request.open("GET", url, false); 
			request.send(null);
			jsonSubtype = JSON.parse(request.responseText);
			cache[cacheKey] = jsonSubtype;
		}
		return jsonSubtype && jsonSubtype.id ? jsonSubtype : null;
	}

	/**
		Get information about project default items and recently used items 
		of the given subtype or item type. If a subtype is given, only project default 
		and recent items of the same subtype are returned, otherwise the item type 
		parameter is used and all project default items WITHOUT any subtype are returned.
		
		The function returns an object with the 'id' and 'name' of the subtype and
		two arrays ('default' and 'recent') which both contain elements that are 
		objects with the 'id' and 'name' of the found items.
		
		@param itemType The main type of items to find
		@param subtypeId The ID of the subtype we want to find information for
	*/
	subtype.getProjectDefaultAndRecentItems = function(itemType, subtypeId)
	{
		var cacheKey = 'defaultitems.'+itemType+'.'+subtypeId;
		var jsonItems = cache[cacheKey];
		
		if (!jsonItems)
		{
			var request = Ajax.getXmlHttpRequest();
			var url = App.getRoot() + 'admin/itemsubtypes/ajax.jsp?ID=' + App.getSessionId();
			url += '&cmd=GetProjectDefaultAndRecentItems';
			url += '&item_id='+subtypeId;
			url += '&itemType='+itemType;
			request.open("GET", url, false); 
			request.send(null);
			jsonItems = JSON.parse(request.responseText);
			cache[cacheKey] = jsonItems;
		}

		return jsonItems;
	}

	/**
		Utility function for finding the project default and recently used items for
		several types of items at the same time. 
		
	*/
	subtype.getRelatedProjectDefaultAndRecentItems = function(itemType, subtypeId, relatedItemTypes)
	{
		var cacheKey = 'relateditems.'+itemType+'.'+subtypeId;
		
		var info = cache[cacheKey];
		if (info)
		{
			// We have some info already, but...
			// ...check if we are missing info about some related item types
			var missingInfo = [];
			for (var i = 0; i < relatedItemTypes.length; i++)
			{
				if (!info[relatedItemTypes[i]]) missingInfo[missingInfo.length] = relatedItemTypes[i];
			}
			relatedItemTypes = missingInfo;
		}
		
		// Get more info...
		if (relatedItemTypes.length > 0)
		{
			var request = Ajax.getXmlHttpRequest();
			var url = App.getRoot() + 'admin/itemsubtypes/ajax.jsp?ID=' + App.getSessionId();
			url += '&cmd=GetRecentAndRelated&itemType='+itemType;
			if (subtypeId) 
			{
				url += '&item_id='+subtypeId;
			}
			for (var i = 0; i < relatedItemTypes.length; i++)
			{
				url += '&relatedType=' + relatedItemTypes[i];
			}
			request.open("GET", url, false); 
			request.send(null);
			var moreInfo = JSON.parse(request.responseText);
			if (!info)
			{
				info = moreInfo;
			}
			else
			{
				// Merge the new info with the existing info
				for (var i = 0; i < relatedItemTypes.length; i++)
				{
					info[relatedItemTypes[i]]= moreInfo[relatedItemTypes[i]];
				}
			}
			cache[cacheKey] = info;
		}
		return info;
	}
	
	/**
		Utility function for getting the ID of the currently selected 
		subtype in a form.
		@param list The ID or element of a selection list with subtype items 
	*/
	subtype.getSubtypeId = function(list)
	{
		list = Doc.element(list);
		if (!list || list.disabled || list.length == 0) return 0;
		var id = Math.abs(parseInt(list.value));
		return id;
	}

	/**
		Utility function for creating a (temporary) filter string that can be added
		to a popup URL for selecting item with a specified subtype. From the
		currently selected subtype we find the related subtype for the associated
		item and generate a filter based on that. For example, when selecting a FILE
		for a PROTOCOL, we use the current subtype of the protocol to find the
		related subtype for FILE items. 
		
		If there is no related subtype, an empty string is returned.
		@param list The ID or element of a selection list with subtype items
		@param relatedItemType The item type of the related subtype we are
			looking for
		@param defaultSubtype The id of the default subtype to use if
			no related subtype is found (optional)
	*/
	subtype.createRelatedFilter = function(list, relatedItemType, defaultSubtypeId)
	{
		var subtypeId = subtype.getSubtypeId(list);
		var relatedSubtype = subtype.getRelatedSubtype(subtypeId, relatedItemType, defaultSubtypeId);
		return relatedSubtype ? '&tmpfilter:INT:itemSubtype=' + relatedSubtype.id : '';
	}

	
	/**
		Update the selection list with 'recently used' and 'project default' items.
		Currently existing options under those headers will be cleared. 
	*/
	subtype.updateSelectionList = function(list, recentItems, projectDefaults, noNoneOption)
	{
		list = Doc.element(list);
		
		var oldSelectedValue = list.selectedIndex >= 0 ? list[list.selectedIndex].value : 0;
		var reselectValue = 0;
		
		// Find the first header option
		for (var i = noNoneOption ? 0 : 1; i < list.length; i++)
		{
			if (list[i].value == 0)
			{
				// Found the header entry
				// Remember the currently selected value if it is after header
				if (list.selectedIndex > i)
				{
					reselectValue = list[list.selectedIndex].value;
				}
				// Clear everything from the header and onwards
				list.length = i;
			}
		}
		
		var firstNewIndex = list.length + 1;
		if (recentItems && recentItems.length)
		{
			// Add recently used items
			var recentlyUsedHeader = new Option('- recently used -', 0);
			recentlyUsedHeader.className = 'recentheader';
			recentlyUsedHeader.disabled = true;
			list[list.length] = recentlyUsedHeader;
			
			for (var i = 0; i < recentItems.length; i++)
			{
				var selected = reselectValue == recentItems[i].id;
				list[list.length] = new Option((i+1) + '. ' + recentItems[i].name, recentItems[i].id, false, selected);
			}
		}
		
		if (projectDefaults && projectDefaults.length)
		{
			// Add project default items
			var projectDefaultHeader = new Option('- project default -', 0);
			projectDefaultHeader.className = 'defaultheader';
			projectDefaultHeader.disabled = true;
			list[list.length] = projectDefaultHeader;
			
			for (var i = 0; i < projectDefaults.length; i++)
			{
				var selected = reselectValue == projectDefaults[i].id;
				list[list.length] = new Option( projectDefaults[i].name, projectDefaults[i].id, false, selected);
			}
		}
		
		// If no item is selected, select the first new item
		if (list.selectedIndex == 0 && list.length > firstNewIndex) 
		{
			list.selectedIndex = firstNewIndex;
		}
		
		var currentSelectedValue = list.selectedIndex >= 0 ? list[list.selectedIndex].value : 0;
		return oldSelectedValue != currentSelectedValue;
	}

	return subtype;
}();

