/* $Id$
	------------------------------------------------------------------
	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	JavaScript functions for the TabControl2 taglib.

	@author Nicklas
	@since 3.3
*/

var TabControl = function()
{
	var tc = {};
	var internal = {};
	
	/**
		Register a function that is responsible for switching 
		between tabs. Typically used when a 'virtual' tab is used as
		a placeholder for loading a different page in the browser.
		Note that validators that have been registered for the current tab
		are not automatically called when a custom switch function is used.
		The switch function must either call TabControl.setActiveTab() (which also
		switches to the new tab) or TabControl.validateActiveTab() if validation
		is wanted.
	*/
	tc.setSwitchFunction = function(tabControl, switchFunction)
	{
		if (!switchFunction || !switchFunction.call) return; // 'switchFunction' is not a function
		tabControl = Doc.element(tabControl);
		if (!tabControl) return;
		tabControl.switchFunction = switchFunction;
	}
	
	/**
		Adds a validator callback to the given tab. All registered validators
		are called whenever the current tab is about to lose focus due to switching
		to another tab or due to form submission. A validator function should
		take a single 'event' parameter. The event object is a custom event object
		with the 'target' property set to the tab that should be validated.
		The validator should return 'true' to indicate successful validation, or
		'false' to indicate failure. If at least one validator signals a failure
		the entire validation is considered a failure and the default action (tab 
		switch, form submission, etc) is cancelled.
	*/
	tc.addTabValidator = function(tab, validator, attributes)
	{
		if (!validator || !validator.call) return; // 'validator' is not a function
		tab = Doc.element(tab);
		if (!tab) return;
		if (attributes) Data.define(tab, attributes);
		tab.addEventListener('base-validate', 
				function(event)
				{
					if (!validator.call(tab, event)) event.preventDefault();
				}
			);
	}
	
	/**
		Adds a listener to a tab that is called whenever that tab is made the active
		tab. The listener function should take a single 'event' parameter. The event 
		object is a custom event object with the 'target' property set to the tab that 
		has been activated.
	*/
	tc.addTabActivateListener = function(tab, listener, attributes)
	{
		if (!listener || !listener.call) return; // 'listener' is not a function
		tab = Doc.element(tab);
		if (!tab) return;
		if (attributes) Data.define(tab, attributes);
		tab.addEventListener('base-activate', listener, false);
	}
	
	
	/**
		Switch to the given tab. Validators attached to the currently active tab
		are called first which may abort the switch. This function returns true of
		the switch was successful, false if not or if the current tab is the same
		as the given tab. If the switch is sucessful, the 'activate' event is sent
		to all listeners registered with TabControl.addTabActivateListener().
		@param tab The tab element or id of a tab element to switch to
	*/
	tc.setActiveTab = function(tab)
	{
		tab = Doc.element(tab);
		if (!tab) return;
		var tabControl = internal.tabControl(tab);
		
		if (tabControl.activeTab == tab) return false;
		if (!tc.validateActiveTab(tabControl)) return false;
	
		if (tabControl.activeTab != null)
		{
			Doc.removeClass(tabControl.activeTab, 'active');
		}
		Doc.addClass(tab, 'active');

		tabControl.activeTab = tab;
		if (tabControl.activeContent != null)
		{
			tabControl.activeContent.style.display = 'none';
		}
		var content = Doc.element(tab.id+'.content');
		if (content != null)
		{
			content.style.display = 'block';
			tabControl.activeContent = content;
		}

		Data.setPageValue('last-tab.'+tabControl.id, tab.id);

		var evt = document.createEvent('Event');
		evt.initEvent('base-activate', false, true);
		tab.dispatchEvent(evt);
		
		return true;
	}
	
	/**
		Validate the currently active tab. If the validation is sucessful or if there
		is no active tab 'true' is returned, 'false' if not.
		@param tabControl A tab control element or the id of a tab control
	*/
	tc.validateActiveTab = function(tabControl)
	{
		tabControl = Doc.element(tabControl);
		if (!tabControl.activeTab) return true;
		return tc.validateTab(tabControl.activeTab);
	}
	
	/**
		Validate the given tab. If the validation is sucessful 'true'
		is returned, 'false' if not.
		@param tab A tab element or the id of a tab
	*/
	tc.validateTab = function(tab)
	{
		var evt = document.createEvent('Event');
		evt.initEvent('base-validate', false, true);
		return tab.dispatchEvent(evt);
	}
	
	/**
		Get the id for the help section that applies to the currently active tab.
		@param tabControl A tab control element or the id of a tab control
	*/
	tc.getActiveHelpId = function(tabControl)
	{
		tabControl = Doc.element(tabControl);
		tab = tabControl.activeTab;
		return tc.getHelpId(tab);
	}
	
	/**
		Get the id for the help section that applies to the given tab.
		@param tabControl A tab element or the id of a tab
	*/
	tc.getHelpId = function(tab)
	{
		return tab ? Data.get(tab,  'help-id') : null;
	}

	/**
		Get a reference to the tab-control object that contains the given tab.
	*/
	internal.tabControl = function(tab)
	{
		return Doc.element(tab.id.substr(0, tab.id.indexOf('.')));
	}
	
	/**
		'click' event handler for tabs. If a switch function has been
		assigned to the tab control, the switch function is called with the
		event as parameter. Otherwise the TabControl.setActiveTab() function
		is used to switch to the clicked tab.
	*/
	internal.tabOnClick = function(event)
	{
		var tab = event.currentTarget;
		var tabControl = internal.tabControl(tab);
		if (tabControl.switchFunction)
		{
			tabControl.switchFunction.call(tabControl, event);
		}
		else
		{
			tc.setActiveTab(tab);
		}
	}
	
	/**
		Initializer that add 'onclick' event handlers to all 'tab' elements.
	*/
	internal.addOnClickHandler = function(tab, autoInit)
	{
		if (autoInit != 'tab') return;
		Events.addEventHandler(tab, 'click', internal.tabOnClick);
		if (tab.tabIndex >= 0)
		{
			Events.enableClickOnEnter(tab);
		}
	}
	Doc.addElementInitializer(internal.addOnClickHandler);

	/**
		Initializer that set the initially active tab on the tab control.
	*/
	internal.setInitialTab = function(tabControl, autoInit)
	{
		if (autoInit != 'tabcontrol') return;
		var initialTab = Data.get(tabControl, 'initial-tab');
		var checkRemembered = Data.int(tabControl, 'remember-last');
		if (initialTab)
		{
			// Prepend the tabcontrol id
			initialTab = tabControl.id + '.' + initialTab;
		}
		else if (checkRemembered)
		{
			var remembered = Data.getPageValue('last-tab.'+tabControl.id);
			if (remembered) initialTab = remembered;
		}
		if (!initialTab)
		{
			// Locate the ID of the first defined tab and make that the initial tab
			var tabs = tabControl.getElementsByClassName('tab');
			if (tabs.length > 0) initialTab = tabs[0].id;
		}
		Doc.addFinalizer(function()
			{
				TabControl.setActiveTab(initialTab);
			});
		
	}
	Doc.addElementInitializer(internal.setInitialTab);
	
	return tc;
}();


